@@ -108,8 +108,7 @@ noprofile(Spec)		 :- '$set_pattr'(Spec, (noprofile)).
 	      call_cleanup/2,
 	      call_cleanup/3,
 	      setup_call_cleanup/3,
-	      setup_call_catcher_cleanup/4,
-	      (^)/2)).
+	      setup_call_catcher_cleanup/4)).
 
 :- meta_predicate
 	';'(0,0),
@@ -123,14 +122,12 @@ noprofile(Spec)		 :- '$set_pattr'(Spec, (noprofile)).
 	call(5,?,?,?,?,?),
 	call(6,?,?,?,?,?,?),
 	call(7,?,?,?,?,?,?,?),
-	^(+,0),
 	not(0),
 	\+(0),
 	'->'(0,0),
 	'*->'(0,0),
 	once(0),
 	ignore(0),
-	block(+,0,-),
 	catch(0,?,0),
 	setup_call_cleanup(0,0,0),
 	setup_call_catcher_cleanup(0,0,?,0),
@@ -145,7 +142,7 @@ noprofile(Spec)		 :- '$set_pattr'(Spec, (noprofile)).
 	call/10,
 	call/11.
 
-:- '$iso'((call/1, (\+)/1, once/1, (;)/2, (,)/2, (->)/2, catch/3)).
+:- '$iso'((call/1, (\+)/1, once/1, (;)/2, (',')/2, (->)/2, catch/3)).
 
 %   ->/2, ;/2, |/2 and \+/1 are normally compiled. These predicate catch them
 %   in case they are called via the meta-call predicates.
@@ -210,30 +207,9 @@ ignore(Goal) :-
 	Goal, !.
 ignore(_Goal).
 
-_Var^Goal :-					% setof/3, bagof/3
-	Goal.
-
 false :-					% SICStus compatibility
 	fail.
 
-%	block/3, !/1, exit/2, fail/1
-%	`longjmp' like control-structures.  See manual.  The predicate
-%	system:block/3 is used by the VMI's I_CUT_BLOCK and B_EXIT.
-%	'$exit' and '$cut' are interpreted by the compiler/decompiler.
-
-block(_Label, Goal, _RVal) :-
-	Goal.
-
-!(Label) :-
-	'$cut'(Label).				% handled by compiler
-
-exit(Label, RVal) :-
-	'$exit'(Label, RVal).			% handled by compiler
-
-fail(Label) :-
-	'$cut'(Label),				% handled by compiler
-	fail.
-
 %%	catch(:Goal, +Catcher, :Recover)
 %
 %	ISO compliant exception handling.
@@ -241,6 +217,18 @@ fail(Label) :-
 catch(_Goal, _Catcher, _Recover) :-
 	'$catch'.
 
+%%	'$recover_and_rethrow'(:Goal, +Term)
+%
+%	This goal is used to wrap  the   catch/3  recover handler if the
+%	exception is not supposed to be   `catchable'.  An example of an
+%	uncachable exception is '$aborted', used   by abort/0. Note that
+%	we cut to ensure  that  the   exception  is  not delayed forever
+%	because the recover handler leaves a choicepoint.
+
+'$recover_and_rethrow'(Goal, Exception) :-
+	call_cleanup(Goal, throw(Exception)), !.
+
+
 %%	setup_call_cleanup(:Setup, :Goal, :Cleanup).
 %%	setup_call_catcher_cleanup(:Setup, :Goal, +Catcher, :Cleanup).
 %%	call_cleanup(:Goal, :Cleanup)
@@ -375,8 +363,6 @@ default_module(Me, Super) :-
 %	to give a DWIM warning. Otherwise fail. C will print an error
 %	message.
 
-:- set_prolog_flag(autoload, true).
-:- set_prolog_flag(verbose_autoload, false).
 :- flag('$autoloading', _, 0).
 
 '$undefined_procedure'(Module, Name, Arity, Action) :-
@@ -419,25 +405,27 @@ default_module(Me, Super) :-
 %	 handle debugger 'w', 'p' and <N> depth options.
 
 '$set_debugger_print_options'(write) :- !,
-	set_prolog_flag(debugger_print_options,
-			[ quoted(true),
-			  attributes(write)
-			]).
+	create_prolog_flag(debugger_print_options,
+			   [ quoted(true),
+			     attributes(write),
+			     spacing(next_argument)
+			   ], []).
 '$set_debugger_print_options'(print) :- !,
-	set_prolog_flag(debugger_print_options,
-			[ quoted(true),
-			  portray(true),
-			  max_depth(10),
-			  attributes(portray)
-			]).
+	create_prolog_flag(debugger_print_options,
+			   [ quoted(true),
+			     portray(true),
+			     max_depth(10),
+			     attributes(portray),
+			     spacing(next_argument)
+			   ], []).
 '$set_debugger_print_options'(Depth) :-
 	current_prolog_flag(debugger_print_options, Options0),
 	(   '$select'(max_depth(_), Options0, Options)
 	->  true
 	;   Options = Options0
 	),
-	set_prolog_flag(debugger_print_options,
-			[max_depth(Depth)|Options]).
+	create_prolog_flag(debugger_print_options,
+			   [max_depth(Depth)|Options], []).
 
 
 		/********************************
@@ -499,7 +487,7 @@ user:file_search_path(app_preferences, UserHome) :-
 	catch(expand_file_name(~, [UserHome]), _, fail).
 
 
-%	expand_file_search_path(+Spec, -Expanded)
+%	expand_file_search_path(+Spec, -Expanded) is nondet.
 %
 %	Expand a search path.  The system uses depth-first search upto a
 %	specified depth.  If this depth is exceeded an exception is raised.
@@ -511,7 +499,7 @@ expand_file_search_path(Spec, Expanded) :-
 	      throw(error(loop_error(Spec), file_search(Used)))).
 
 '$expand_file_search_path'(Spec, Expanded, N, Used) :-
-	functor(Spec, Alias, 1),
+	functor(Spec, Alias, 1), !,
 	user:file_search_path(Alias, Exp0),
 	NN is N + 1,
 	(   NN > 16
@@ -519,32 +507,17 @@ expand_file_search_path(Spec, Expanded) :-
 	;   true
 	),
 	'$expand_file_search_path'(Exp0, Exp1, NN, [Alias=Exp0|Used]),
-	arg(1, Spec, Base),
-	'$make_path'(Exp1, Base, Expanded).
-'$expand_file_search_path'(A/B, Expanded, _, _) :- !,
-	'$make_path_from_slash'(A/B, Parts, []),
-	atomic_list_concat(Parts, Expanded).
-'$expand_file_search_path'(Spec, Spec, _, _) :-
-	atomic(Spec).
+	arg(1, Spec, Segments),
+	'$segments_to_atom'(Segments, File),
+	'$make_path'(Exp1, File, Expanded).
+'$expand_file_search_path'(Spec, Path, _, _) :-
+	'$segments_to_atom'(Spec, Path).
 
 '$make_path'(Dir, File, Path) :-
 	atom_concat(_, /, Dir), !,
 	atom_concat(Dir, File, Path).
 '$make_path'(Dir, File, Path) :-
-	'$make_path_from_slash'(File, Parts, []),
-	'$append'([Dir, /], Parts, AllParts),
-	atomic_list_concat(AllParts, Path).
-
-
-%%	'$make_path_from_slash'(+SlashPath, -Parts, ?Tail) is det.
-%
-%	Translate a/b/c into [a,/,b,/,c]. We cannot   use DCG as the DCG
-%	compiler is not yet defined.
-
-'$make_path_from_slash'(A/B, P0, P) :- !,
-	'$make_path_from_slash'(A, P0, [/|P1]),
-	'$make_path_from_slash'(B, P1, P).
-'$make_path_from_slash'(A, [A|T], T).
+	atomic_list_concat([Dir, /, File], Path).
 
 
 		/********************************
@@ -579,6 +552,7 @@ absolute_file_name(Spec, Args, Path) :-
 	;   Conditions = Args,
 	    Exts = ['']
 	),
+	'$canonise_extensions'(Exts, Extensions),
 	(   '$select'(solutions(Sols), Conditions, C1)
 	->  true
 	;   Sols = first,
@@ -596,15 +570,15 @@ absolute_file_name(Spec, Args, Path) :-
 	;   Spec1 = Spec,
 	    C3 = C2
 	),
-	(   '$chk_file'(Spec1, Exts, C3, Path)
-	*-> (   Sols == first
-	    ->  !
-	    ;   true
-	    )
-	;   (   FileErrors == fail
-	    ->  fail
-	    ;   throw(error(existence_error(source_sink, Spec), _))
+	(   Sols == first
+	->  (	'$chk_file'(Spec1, Extensions, C3, true, Path)
+	    ->	true
+	    ;	(   FileErrors == fail
+		->  fail
+		;   throw(error(existence_error(source_sink, Spec), _))
+		)
 	    )
+	;   '$chk_file'(Spec1, Extensions, C3, false, Path)
 	).
 
 '$file_type_extensions'(source, Exts) :- !, 	% SICStus 3.9 compatibility
@@ -621,8 +595,8 @@ absolute_file_name(Spec, Args, Path) :-
 %	absolute_file_name/3 and may be used to extend the list of
 %	extensions used for some type.
 
-:- multifile(user:prolog_file_type/2),
-   dynamic(user:prolog_file_type/2).
+:- multifile(user:prolog_file_type/2).
+:- dynamic(user:prolog_file_type/2).
 
 user:prolog_file_type(pl,	prolog).
 user:prolog_file_type(Ext,	prolog) :-
@@ -636,39 +610,48 @@ user:prolog_file_type(Ext,	executable) :-
 %	File is a specification of a Prolog source file. Return the full
 %	path of the file.
 
-'$chk_file'(Spec, Extensions, Cond, FullName) :-
-	'$canonise_extensions'(Extensions, Exts),
-	'$dochk_file'(Spec, Exts, Cond, FullName).
-
-'$dochk_file'(Spec, Extensions, Cond, FullName) :-
+'$chk_file'(Spec, Extensions, Cond, Cache, FullName) :-
 	compound(Spec),
-	functor(Spec, Alias, 1),
-	user:file_search_path(Alias, _), !,
+	functor(Spec, _, 1), !,
 	'$relative_to'(Cond, cwd, CWD),
-	'$chk_alias_file'(Spec, Extensions, Cond, CWD, FullName).
-'$dochk_file'(Term, Ext, Cond, FullName) :-	% allow a/b, a-b, etc.
-	\+ atomic(Term), !,
-	term_to_atom(Term, Raw),
-	atom_chars(Raw, S0),
-	'$delete'(S0, ' ', S1),
-	atom_chars(Atom, S1),
-	'$dochk_file'(Atom, Ext, Cond, FullName).
-'$dochk_file'(File, Exts, Cond, FullName) :-
+	'$chk_alias_file'(Spec, Extensions, Cond, Cache, CWD, FullName).
+'$chk_file'(Segments, Ext, Cond, Cache, FullName) :-	% allow a/b/...
+	\+ atomic(Segments), !,
+	'$segments_to_atom'(Segments, Atom),
+	'$chk_file'(Atom, Ext, Cond, Cache, FullName).
+'$chk_file'(File, Exts, Cond, _, FullName) :-
 	is_absolute_file_name(File), !,
 	'$extend_file'(File, Exts, Extended),
 	'$file_condition'(Cond, Extended),
 	'$absolute_file_name'(Extended, FullName).
-'$dochk_file'(File, Exts, Cond, FullName) :-
+'$chk_file'(File, Exts, Cond, _, FullName) :-
 	'$relative_to'(Cond, source, Dir),
 	atomic_list_concat([Dir, /, File], AbsFile),
 	'$extend_file'(AbsFile, Exts, Extended),
 	'$file_condition'(Cond, Extended), !,
 	'$absolute_file_name'(Extended, FullName).
-'$dochk_file'(File, Exts, Cond, FullName) :-
+'$chk_file'(File, Exts, Cond, _, FullName) :-
 	'$extend_file'(File, Exts, Extended),
 	'$file_condition'(Cond, Extended),
 	'$absolute_file_name'(Extended, FullName).
 
+'$segments_to_atom'(Atom, Atom) :-
+	atomic(Atom), !.
+'$segments_to_atom'(Segments, Atom) :-
+	'$segments_to_list'(Segments, List, []), !,
+	atomic_list_concat(List, /, Atom).
+'$segments_to_atom'(Segments, _) :-
+	throw(error(type_error(file_path, Segments), _)).
+
+'$segments_to_list'(Var, _, _) :-
+	var(Var), !, fail.
+'$segments_to_list'(A/B, H, T) :-
+	'$segments_to_list'(A, H, T0),
+	'$segments_to_list'(B, T0, T).
+'$segments_to_list'(A, [A|T], T) :-
+	atomic(A).
+
+
 %	'$relative_to'(+Condition, +Default, -Dir)
 %
 %	Determine the directory to work from.  This can be specified
@@ -691,28 +674,38 @@ user:prolog_file_type(Ext,	executable) :-
 	    file_directory_name(ContextFile, Dir)
 	).
 
+%%	'$chk_alias_file'(+Spec, +Exts, +Cond, +Cache, +CWD,
+%%			  -FullFile) is nondet.
+
 :- dynamic
-	'$search_path_file_cache'/5.
+	'$search_path_file_cache'/4.	% Spec, Hash, Cache, Path
 :- volatile
-	'$search_path_file_cache'/5.
+	'$search_path_file_cache'/4.
 
-:- set_prolog_flag(verbose_file_search, false).
-
-'$chk_alias_file'(Spec, Exts, Cond, CWD, FullFile) :-
-	'$search_path_file_cache'(Spec, Cond, CWD, FullFile, Exts),
-	'$file_condition'(Cond, FullFile),
-	'$search_message'(file_search(cache(Spec, Cond), FullFile)).
-'$chk_alias_file'(Spec, Exts, Cond, CWD, FullFile) :-
+'$chk_alias_file'(Spec, Exts, Cond, true, CWD, FullFile) :- !,
+	findall(Exp, expand_file_search_path(Spec, Exp), Expansions),
+	Cache = cache(Exts, Cond, CWD, Expansions),
+	term_hash(Cache, Hash),
+	(   '$search_path_file_cache'(Spec, Hash, Cache, FullFile),
+	    '$file_condition'(Cond, FullFile)
+	->  '$search_message'(file_search(cache(Spec, Cond), FullFile))
+	;   '$member'(Expanded, Expansions),
+	    '$extend_file'(Expanded, Exts, LibFile),
+	    (   '$file_condition'(Cond, LibFile),
+		'$absolute_file_name'(LibFile, FullFile),
+		\+ '$search_path_file_cache'(Spec, Hash, Cache, FullFile),
+		assert('$search_path_file_cache'(Spec, Hash, Cache, FullFile))
+	    ->  '$search_message'(file_search(found(Spec, Cond), FullFile))
+	    ;   '$search_message'(file_search(tried(Spec, Cond), LibFile)),
+		fail
+	    )
+	).
+'$chk_alias_file'(Spec, Exts, Cond, false, _CWD, FullFile) :-
 	expand_file_search_path(Spec, Expanded),
 	'$extend_file'(Expanded, Exts, LibFile),
-	(   '$file_condition'(Cond, LibFile),
-	    '$absolute_file_name'(LibFile, FullFile),
-	    \+ '$search_path_file_cache'(Spec, Cond, CWD, FullFile, Exts),
-	    assert('$search_path_file_cache'(Spec, Cond, CWD, FullFile, Exts))
-	->  '$search_message'(file_search(found(Spec, Cond), FullFile))
-	;   '$search_message'(file_search(tried(Spec, Cond), LibFile)),
-	    fail
-	).
+	'$file_condition'(Cond, LibFile),
+	'$absolute_file_name'(LibFile, FullFile).
+
 
 '$search_message'(Term) :-
 	current_prolog_flag(verbose_file_search, true), !,
@@ -1019,28 +1012,45 @@ consult(List) :-
 load_files(Files) :-
 	load_files(Files, []).
 load_files(Module:Files, Options) :-
+	(   is_list(Options)
+	->  true
+	;   throw(error(type_error(list, Options), _))
+	),
         '$load_files'(Files, Module, Options).
 
+'$load_files'(X, _, _) :-
+	var(X), !,
+	throw(error(instantiation_error, context(load_files/2,_))).
+'$load_files'([], _, _) :- !.
 '$load_files'(Id, Module, Options) :-	% load_files(foo, [stream(In)])
 	memberchk(stream(_), Options), !,
 	(   atom(Id)
 	->  '$load_file'(Id, Module, Options)
 	;   throw(error(type_error(atom, Id), _))
 	).
-'$load_files'(X, _, _) :-
-	var(X), !,
-	throw(error(instantiation_error, context(load_files/2,_))).
-'$load_files'([], _, _) :- !.
-'$load_files'([H|T], Module, Options) :- !,
-	'$load_files'(H, Module, Options),
-	'$load_files'(T, Module, Options).
-'$load_files'(Spec, Module, Options) :-
+'$load_files'(List, Module, Options) :-
+	List = [_|_], !,
+	(   is_list(List)
+	->  '$load_file_list'(List, Module, Options)
+	;   throw(error(type_error(list, List), context(load_files/2,_)))
+	).
+'$load_files'(File, Module, Options) :-
+	'$load_one_file'(File, Module, Options).
+
+'$load_file_list'([], _, _).
+'$load_file_list'([File|Rest], Module, Options) :-
+	catch('$load_one_file'(File, Module, Options), E,
+	      print_message(error, E)),
+	'$load_file_list'(Rest, Module, Options).
+
+
+'$load_one_file'(Spec, Module, Options) :-
 	atom(Spec),
 	'$get_option'(expand(Expand), Options, true),
 	Expand == true, !,
 	expand_file_name(Spec, Files),
 	'$load_files'(Files, Module, [expand(false)|Options]).
-'$load_files'(File, Module, Options) :-
+'$load_one_file'(File, Module, Options) :-
 	strip_module(Module:File, Into, PlainFile),
 	'$load_file'(PlainFile, Into, Options).
 
@@ -1069,7 +1079,7 @@ load_files(Module:Files, Options) :-
         time_file(FullFile, Modified),
         Modified @=< LoadTime, !.
 
-%	'$qlf_file'(+Spec, +PlFile, -LoadFile)
+%	'$qlf_file'(+Spec, +PlFile, -LoadFile, -Mode, +Options)
 %
 %	Return the QLF file if it exists.  Might check for modification
 %	time, version, etc.
@@ -1077,19 +1087,70 @@ load_files(Module:Files, Options) :-
 %	If the user-specification specified a prolog file, do not
 %	replace this with a .qlf file.
 
-'$qlf_file'(Spec, FullFile, FullFile) :-
+'$qlf_file'(Spec, _, Spec, stream, Options) :-
+	memberchk(stream(_), Options), !.
+'$qlf_file'(Spec, FullFile, FullFile, compile, _) :-
 	'$spec_extension'(Spec, Ext),
 	user:prolog_file_type(Ext, prolog), !.
-'$qlf_file'(_, FullFile, QlfFile) :-
+'$qlf_file'(_, FullFile, QlfFile, Mode, Options) :-
 	flag('$compiling', database, database),
 	file_name_extension(Base, PlExt, FullFile),
 	user:prolog_file_type(PlExt, prolog),
 	user:prolog_file_type(QlfExt, qlf),
 	file_name_extension(Base, QlfExt, QlfFile),
-	access_file(QlfFile, read).
-'$qlf_file'(_, FullFile, FullFile).
+	(   access_file(QlfFile, read),
+	    (	'$qlf_up_to_date'(FullFile, QlfFile)
+	    ->	Mode = qload
+	    ;	access_file(QlfFile, write)
+	    ->	Mode = qcompile
+	    )
+	->  !
+	;   '$qlf_auto'(FullFile, QlfFile, Options)
+	->  !, Mode = qcompile
+	).
+'$qlf_file'(_, FullFile, FullFile, compile, _).
 
 
+%%	'$qlf_up_to_date'(+PlFile, +QlfFile) is semidet.
+%
+%	True if the QlfFile file is  considered up-to-date. This implies
+%	that either the PlFile does not exist or that the QlfFile is not
+%	older than the PlFile.
+
+'$qlf_up_to_date'(PlFile, QlfFile) :-
+	(   exists_file(PlFile)
+	->  time_file(PlFile, PlTime),
+	    time_file(QlfFile, QlfTime),
+	    QlfTime >= PlTime
+	;   true
+	).
+
+%%	'$qlf_auto'(+PlFile, +QlfFile, +Options) is semidet.
+%
+%	True if we create QlfFile using   qcompile/2. This is determined
+%	by the option qcompile(QlfMode) or, if   this is not present, by
+%	the prolog_flag qcompile.
+
+:- create_prolog_flag(qcompile, false, [type(atom)]).
+
+'$qlf_auto'(PlFile, QlfFile, Options) :-
+	\+ '$in_system_dir'(PlFile),
+	(   memberchk(qcompile(QlfMode), Options)
+	->  true
+	;   current_prolog_flag(qcompile, QlfMode)
+	),
+	(   QlfMode == auto
+	->  true
+	;   QlfMode == large,
+	    size_file(PlFile, Size),
+	    Size > 100000
+	),
+	access_file(QlfFile, write).
+
+'$in_system_dir'(PlFile) :-
+	current_prolog_flag(home, Home),
+	sub_atom(PlFile, 0, _, _, Home).
+
 '$spec_extension'(File, Ext) :-
 	atom(File),
 	file_name_extension(_, Ext, File).
@@ -1146,6 +1207,13 @@ load_files(Module:Files, Options) :-
 %	=|$load|=
 
 '$do_load_file'(File, FullFile, Module, Options) :-
+	'$qlf_file'(File, FullFile, Absolute, Mode, Options),
+	(   Mode == qcompile
+	->  qcompile(Module:File, Options)
+	;   '$do_load_file_2'(File, Absolute, Module, Options)
+	).
+
+'$do_load_file_2'(File, Absolute, Module, Options) :-
 	statistics(heapused, OldHeap),
 	statistics(cputime, OldTime),
 
@@ -1155,18 +1223,19 @@ load_files(Module:Files, Options) :-
 
 	current_prolog_flag(generate_debug_info, DebugInfo),
 
-	(   memberchk(stream(FromStream), Options)
-	->  Absolute = File
-	;   '$qlf_file'(File, FullFile, Absolute)
-	),
-
 	flag('$compilation_level', Level, Level),
 	'$load_message_level'(MessageLevel),
 
 	'$print_message'(silent /*MessageLevel*/,
 			 load_file(start(Level,
 					 file(File, Absolute)))),
-	(   nonvar(FromStream),
+
+	(   memberchk(stream(FromStream), Options)
+	->  Input = stream
+	;   Input = source
+	),
+
+	(   Input == stream,
 	    (   '$get_option'(format(qlf), Options, source)
 	    ->  set_stream(FromStream, file_name(Absolute)),
 		'$qload_stream'(FromStream, Module, Action, LM, Options)
@@ -1174,7 +1243,7 @@ load_files(Module:Files, Options) :-
 				Module, Action, LM, Options)
 	    )
 	->  true
-	;   var(FromStream),
+	;   Input == source,
 	    '$consult_goal'(Absolute, Goal),
 	    call(Goal, Absolute, Module, Action, LM, Options)
 	->  true
@@ -1213,6 +1282,7 @@ load_files(Module:Files, Options) :-
 % 	Import public predicates from LoadedModule into Module
 
 '$import_from_loaded_module'(LoadedModule, Module, Options) :-
+	LoadedModule \== Module,
 	atom(LoadedModule), !,
 	'$get_option'(imports(Import), Options, all),
 	'$get_option'(reexport(Reexport), Options, false),
@@ -1308,7 +1378,7 @@ load_files(Module:Files, Options) :-
 	'$restore_lex_state'(LexState),
 	'$set_source_module'(_, OldModule).	% Restore old module
 
-:- set_prolog_flag(emulated_dialect, swi).
+:- create_prolog_flag(emulated_dialect, swi, [type(atom)]).
 
 '$save_lex_state'(lexstate(Style, Dialect)) :-
 	'$style_check'(Style, Style),
@@ -1435,7 +1505,8 @@ load_files(Module:Files, Options) :-
 	source_location(_File, Line),
 	'$get_option'(redefine_module(Action), Options, false),
 	'$super_module'(File, Super),
-	'$declare_module'(Module, Super, File, Line, Action),
+	'$redefine_module'(Module, File, Action),
+	'$declare_module'(Module, Super, File, Line, false),
 	'$export_list'(Public, Module, Ops),
 	'$ifcompiling'('$qlf_start_module'(Module)),
 	'$export_ops'(Ops, Module, File),
@@ -1444,6 +1515,46 @@ load_files(Module:Files, Options) :-
 	'$check_export'(Module),
 	'$ifcompiling'('$qlf_end_part').
 
+%%	'$redefine_module'(+Module, +File, -Redefine)
+
+'$redefine_module'(_Module, _, false) :- !.
+'$redefine_module'(Module, File, true) :- !,
+	(   module_property(Module, file(OldFile)),
+	    File \== OldFile
+	->  unload_file(OldFile)
+	;   true
+	).
+'$redefine_module'(Module, File, ask) :-
+	(   stream_property(user_input, tty(true)),
+	    module_property(Module, file(OldFile)),
+	    File \== OldFile,
+	    '$rdef_response'(Module, OldFile, File, true)
+	->  '$redefine_module'(Module, File, true)
+	;   true
+	).
+
+'$rdef_response'(Module, OldFile, File, Ok) :-
+	repeat,
+	print_message(query, redefine_module(Module, OldFile, File)),
+	get_single_char(Char),
+	'$rdef_response'(Char, Ok0), !,
+	Ok = Ok0.
+
+'$rdef_response'(Char, true) :-
+	memberchk(Char, "yY"),
+	format(user_error, 'yes~n', []).
+'$rdef_response'(Char, false) :-
+	memberchk(Char, "nN"),
+	format(user_error, 'no~n', []).
+'$rdef_response'(Char, _) :-
+	memberchk(Char, "a"),
+	format(user_error, 'abort~n', []),
+	abort.
+'$rdef_response'(_, _) :-
+	print_message(help, redefine_module_reply),
+	fail.
+
+
 %%	'$super_module'(+File, -Super) is det.
 %
 %	Determine the initial module from which   I  inherit. All system
@@ -1477,7 +1588,7 @@ load_files(Module:Files, Options) :-
 	throw(error(instantitation_error, _)).
 '$import_list'(Target, Source, all, Reexport) :- !,
 	'$module_property'(Source, exports(Import)),
-	'$import_ops'(Target, Source),
+	'$import_ops'(Target, Source, _All),
 	'$import_list'(Target, Source, Import, Reexport).
 '$import_list'(Target, Source, except(Spec), Reexport) :- !,
 	'$module_property'(Source, exports(Export)),
@@ -1524,10 +1635,7 @@ load_files(Module:Files, Options) :-
 	(   Reexport == true,
 	    '$list_to_conj'(Imported, Conj)
 	->  export(Context:Conj),
-	    (	flag('$compiling', wic, wic)
-	    ->	'$add_directive_wic'(export(Context:Conj))
-	    ;	true
-	    )
+	    '$ifcompiling'('$add_directive_wic'(export(Context:Conj)))
 	;   true
 	).
 
@@ -1545,9 +1653,12 @@ load_files(Module:Files, Options) :-
 	),
 	'$store_clause'((NewHead :- Source:Head), File),
 	'$import_all2'(Rest, Context, Source, Imported).
+'$import_all2'([op(P,A,N)|Rest], Context, Source, Imported) :- !,
+	'$import_ops'(Context, Source, op(P,A,N)),
+	'$import_all2'(Rest, Context, Source, Imported).
 '$import_all2'([Pred|Rest], Context, Source, [Pred|Imported]) :-
 	Context:import(Source:Pred),
-	'$import_wic'(Source, Pred),
+	'$ifcompiling'('$import_wic'(Source, Pred)),
 	'$import_all2'(Rest, Context, Source, Imported).
 
 
@@ -1555,14 +1666,15 @@ load_files(Module:Files, Options) :-
 '$list_to_conj'([H|T], (H,Rest)) :-
 	'$list_to_conj'(T, Rest).
 
-%%	'$import_ops'(+Target, +Source)
+%%	'$import_ops'(+Target, +Source, +Pattern)
 %
 %	Import the operators export from Source into the module table of
-%	Target.
+%	Target.  We only import operators that unify with Pattern.
 
-'$import_ops'(To, From) :-
+'$import_ops'(To, From, Pattern) :-
 	(   '$c_current_predicate'(_, From:'$exported_op'(_, _, _)),
 	    From:'$exported_op'(Pri, Assoc, Name),
+	    Pattern = op(Pri, Assoc, To:Name),
 	    op(Pri, Assoc, To:Name),
 	    fail
 	;   true
@@ -1734,11 +1846,16 @@ load_files(Module:Files, Options) :-
 '$load_goal'(consult(_)).
 '$load_goal'(load_files(_)).
 '$load_goal'(load_files(_,Options)) :-
-	memberchk(qcompile(true), Options).
+	memberchk(qcompile(QlfMode), Options),
+	'$qlf_part_mode'(QlfMode).
 '$load_goal'(ensure_loaded(_)) :- flag('$compiling', wic, wic).
 '$load_goal'(use_module(_))    :- flag('$compiling', wic, wic).
 '$load_goal'(use_module(_, _)) :- flag('$compiling', wic, wic).
 
+'$qlf_part_mode'(part).
+'$qlf_part_mode'(true).			% compatibility
+
+
 		/********************************
 		*        TERM EXPANSION         *
 		*********************************/
@@ -1758,17 +1875,16 @@ load_files(Module:Files, Options) :-
 '$store_clause'((_, _), _) :- !,
 	print_message(error, cannot_redefine_comma),
 	fail.
-'$store_clause'('$source_location'(File, Line):Term, _) :- !,
-	'$record_clause'(Term, File:Line, Ref),
-	'$qlf_assert_clause'(Ref).
 '$store_clause'(Term, File) :-
-	'$record_clause'(Term, File, Ref),
-        '$qlf_assert_clause'(Ref).
+	'$clause_source'(Term, Clause, File, SrcLoc),
+	(   flag('$compiling', database, database)
+	->  '$record_clause'(Clause, SrcLoc)
+	;   '$record_clause'(Clause, SrcLoc, Ref),
+	    '$qlf_assert_clause'(Ref, development)
+	).
 
-'$qlf_assert_clause'(_) :-
-	flag('$compiling', database, database), !.
-'$qlf_assert_clause'(Ref) :-
-	'$qlf_assert_clause'(Ref, development).
+'$clause_source'('$source_location'(File,Line):Clause, Clause, _, File:Line) :- !.
+'$clause_source'(Clause, Clause, File, File).
 
 
 %%	compile_aux_clauses(+Clauses) is det.
