@@ -3,9 +3,9 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -39,10 +39,10 @@
 
 This module defines goal_expansion/2 rules to   deal with commonly used,
 but fundamentally slow meta-predicates. Notable   maplist/2... defines a
-useful set of predicates, but its exection is considerable slower than a
-traditional Prolog loop. Using this library   calls  to maplist/2... are
+useful set of predicates, but its  execution is considerable slower than
+a traditional Prolog loop. Using this  library calls to maplist/2... are
 translated into an call  to  a   generated  auxilary  predicate  that is
-compiled using compile_aux_clauses/1.  Currently this module supports:
+compiled using compile_aux_clauses/1. Currently this module supports:
 
 	* maplist/2..
 	* forall/2
@@ -51,6 +51,9 @@ compiled using compile_aux_clauses/1.  Currently this module supports:
 	* phrase/2
 	* phrase/3
 
+The idea for this library originates from ECLiPSe and came to SWI-Prolog
+through YAP.
+
 @tbd	Support more predicates
 @author	Jan Wielemaker
 */
@@ -99,7 +102,7 @@ expand_maplist(Callable0, Lists, Goal) :-
 	    append(Tails, Argv, IttArgs),
 	    NextIterate =.. [AuxName|IttArgs],
 	    NextClause = (NextHead :- NextGoal, NextIterate),
-	    
+
 	    (	predicate_property(NextGoal, transparent)
 	    ->	compile_aux_clauses([ (:- module_transparent(Module:AuxName/AuxArity)),
 				      BaseClause,
@@ -129,19 +132,23 @@ expand_apply(Maplist, Goal) :-
 	functor(Maplist, maplist, N),
 	N >= 2,
 	Maplist =.. [maplist, Callable|Lists],
-	callable(Callable), !,
+	qcall_instantiated(Callable), !,
 	expand_maplist(Callable, Lists, Goal).
 expand_apply(forall(Cond, Action), \+((Cond, \+(Action)))).
-expand_apply(once(Goal), (Goal->true)).
+expand_apply(once(Goal), (Goal->true;fail)).
 expand_apply(ignore(Goal), (Goal->true;true)).
-expand_apply(phrase(NT,Xs), NTXsNil) :-
+expand_apply(Phrase, Expanded) :-
+	expand_phrase(Phrase, Expanded), !.
+
+
+expand_phrase(phrase(NT,Xs), NTXsNil) :- !,
 	expand_apply(phrase(NT,Xs,[]), NTXsNil).
-expand_apply(phrase(NT,Xs0,Xs), NewGoal) :-
+expand_phrase(Goal, NewGoal) :-
 	Goal = phrase(NT,Xs0,Xs),
 	nonvar(NT),
-	catch('$translate_rule'((pseudo_nt --> NT), Rule),
+	catch(dcg_translate_rule((pseudo_nt --> NT), Rule),
 	      error(Pat,ImplDep),
-	      ( \+ harmless_dcgexception(Pat), 
+	      ( \+ harmless_dcgexception(Pat),
 		throw(error(Pat,ImplDep))
 	      )),
 	Rule = (pseudo_nt(Xs0c,Xsc) :- NewGoal0),
@@ -157,6 +164,22 @@ expand_apply(phrase(NT,Xs0,Xs), NewGoal) :-
 	;  ( Xs0 = Xs0c, NewGoal1 ) = NewGoal
 	).
 
+%%	qcall_instantiated(@Term) is semidet.
+%
+%	True if Term is instantiated sufficiently to call it.
+%
+%	@tbd	Shouldn't this be callable straight away?
+
+qcall_instantiated(Var) :-
+	var(Var), !,
+	fail.
+qcall_instantiated(M:C) :- !,
+	atom(M),
+	callable(C).
+qcall_instantiated(C) :-
+	callable(C).
+
+
 harmless_dcgexception(instantiation_error).	% ex: phrase(([1],x:X,[3]),L)
 harmless_dcgexception(type_error(callable,_)).	% ex: phrase(27,L)
 
@@ -177,9 +200,12 @@ contains_illegal_dcgnt(NT) :-
 		 *	     ACTIVATE		*
 		 *******************************/
 
+:- multifile
+	system:goal_expansion/2.
+
 %	@tbd	Should we only apply if optimization is enabled (-O)?
 
-user:goal_expansion(GoalIn, GoalOut) :-
+system:goal_expansion(GoalIn, GoalOut) :-
 	\+ current_prolog_flag(xref, true),
 	expand_apply(GoalIn, GoalOut).
-	
+
