@@ -1,4 +1,4 @@
-/*  $Id: dde.pl,v 1.8 2004/09/30 10:41:35 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -71,11 +71,16 @@ dde_poke(Handle, Item, Data) :-
 :- module_transparent
 	dde_register_service/2.
 
-%	dde_register_service(+Service(Topic, Item, Value), +Goal).
-%	dde_register_service(+Service(Topic, Command), +Goal).
+%%	dde_register_service(Template, +Goal) is det.
 %
-%	Register a DDE service satisfied by this SWI-Prolog instance.  The
-%	first form is for dde_request/3, the second for dde_execute/2.
+%	Register a DDE service  in   this  SWI-Prolog instance. Template
+%	takes one of the formats
+%
+%		$ Service(Topic, Item, Value) :
+%		Register a service for dde_request/3
+%
+%		$ Service(Topic, Command) :
+%		Register a service for dde_execute/2
 %
 %	Topic is either an atom or a variable.  In the latter case, any
 %	topic is confirmed on the Service.  Item, Value and Command are
@@ -107,18 +112,18 @@ dde_unregister_all_services :-
 	dde_current_service(Service, _),
 	dde_unregister_service(Service),
 	fail ; true.
-	
+
 :- at_halt(dde_unregister_all_services). % required by Windows!
 
-%	dde_current_service(?Service, ?Topic)
+%%	dde_current_service(?Service, ?Topic)
 %
 %	Unifies Service and Topic with currently supported servers/topics
 
 dde_current_service(Service, Topic) :-
 	dde_service(Service, Topic, _, _, _, _).
 
-%	dde_current_connection(?Service, ?Topic)
-%	
+%%	dde_current_connection(?Service, ?Topic)
+%
 %	Unifies Service and Topic with the currently open server connections.
 
 dde_current_connection(Service, Topic) :-
@@ -128,7 +133,7 @@ dde_current_connection(Service, Topic) :-
 		 *	    CALL-BACKS		*
 		 *******************************/
 
-%	$dde_connect(+Service, +Topic, +IsSelf)
+%	'$dde_connect'(+Service, +Topic, +IsSelf)
 %
 %	Called by the DDEML XTYP_CONNECT request.  IsSelf is 0 if the
 %	connection is requested by another Prolog, 1 otherwise.
@@ -136,7 +141,7 @@ dde_current_connection(Service, Topic) :-
 '$dde_connect'(Service, Topic, _Self) :-
 	dde_service(Service, Topic, _, _, _, _).
 
-%	$dde_connect_confirm(+Service, +Topic, +Handle)
+%	'$dde_connect_confirm'(+Service, +Topic, +Handle)
 %
 %	Called by the DDEML XTYP_CONNECT_CONFIRM request.  Used to update
 %	our list of current conversations.
@@ -144,7 +149,7 @@ dde_current_connection(Service, Topic) :-
 '$dde_connect_confirm'(Service, Topic, Handle) :-
 	asserta(dde_current_connection(Handle, Service, Topic)).
 
-%	$dde_disconnect(+Handle)
+%	'$dde_disconnect'(+Handle)
 %
 %	Called by the DDEML XTYP_DISCONNECT request.  Used to update our
 %	list of current conversations.
@@ -152,7 +157,7 @@ dde_current_connection(Service, Topic) :-
 '$dde_disconnect'(Handle) :-
 	retractall(dde_current_connection(Handle, _, _)).
 
-%	$dde_request(+Handle, +Topic, +Item, -Answer)
+%	'$dde_request'(+Handle, +Topic, +Item, -Answer)
 %
 %	Called by the DDEML XTYP_REQUEST request.  Answer should be unified
 %	with a Prolog object that can be translated into a textual value
@@ -166,7 +171,7 @@ dde_current_connection(Service, Topic) :-
 '$dde_request'(_Handle, Topic, _Item, _Answer) :-
 	throw(error(existence_error(dde_topic, Topic), _)).
 
-%	$dde_execute(+Handle, +Topic, +Command)
+%	'$dde_execute'(+Handle, +Topic, +Command)
 %
 %	Called by the DDEML XTYP_EXECUTE request.  Command is a SWI-Prolog
 %	string object.  Use string_to_atom or string_to_list to convert it
