@@ -244,13 +244,15 @@ do_portray_clause(Out, (Head :- Body)) :- !,
 	pprint(Out, Head, LeftPri),
 	write(Out, ' :-'),
 	(   nonvar(Body),
-	    Body = Module:LocalBody
+	    Body = Module:LocalBody,
+	    \+ primitive(LocalBody)
 	->  nlindent(Out, Indent),
 	    format(Out, '~q', [Module]),
 	    '$put_token'(Out, :),
 	    nlindent(Out, Indent),
 	    write(Out, '(   '),
-	    portray_body(LocalBody, 2, noindent, 1200, Out),
+	    inc_indent(Indent, 1, BodyIndent),
+	    portray_body(LocalBody, BodyIndent, noindent, 1200, Out),
 	    nlindent(Out, Indent),
 	    write(Out, ')')
 	;   inc_indent(0, 2, BodyIndent),
@@ -310,13 +312,13 @@ portray_body(Term, Indent, _, Pri, Out) :-
 	nlindent(Out, Indent),
 	write(Out, ')').
 portray_body((A,B), Indent, _, _Pri, Out) :- !,
-	infix_op((,), LeftPri, RightPri),
+	infix_op(',', LeftPri, RightPri),
 	portray_body(A, Indent, noindent, LeftPri, Out),
 	write(Out, ','),
 	portray_body(B, Indent, indent, RightPri, Out).
 portray_body(\+(Goal), Indent, _, _Pri, Out) :- !,
-	write(Out, (\+)), write(Out, ' '),
-	prefix_op((\+), ArgPri),
+	write(Out, \+), write(Out, ' '),
+	prefix_op(\+, ArgPri),
 	ArgIndent is Indent+3,
 	portray_body(Goal, ArgIndent, noindent, ArgPri, Out).
 portray_body(Meta, Indent, _, Pri, Out) :-
@@ -386,14 +388,14 @@ portray_or((If *-> Then), Indent, Out) :- !,
 	portray_or(Then, Indent, RightPri, Out).
 portray_or((A;B), Indent, Out) :- !,
 	inc_indent(Indent, 1, NestIndent),
-	infix_op((;), LeftPri, RightPri),
+	infix_op(;, LeftPri, RightPri),
 	portray_body(A, NestIndent, noindent, LeftPri, Out),
 	nlindent(Out, Indent),
 	write(Out, ';   '),
 	portray_or(B, Indent, RightPri, Out).
 portray_or((A|B), Indent, Out) :- !,
 	inc_indent(Indent, 1, NestIndent),
-	infix_op((|), LeftPri, RightPri),
+	infix_op('|', LeftPri, RightPri),
 	portray_body(A, NestIndent, noindent, LeftPri, Out),
 	nlindent(Out, Indent),
 	write(Out, '|   '),
@@ -437,6 +439,11 @@ or_layout((_;_)).
 or_layout((_->_)).
 or_layout((_*->_)).
 
+primitive(G) :-
+	or_layout(G), !, fail.
+primitive((_,_)) :- !, fail.
+primitive(_).
+
 %%	meta_call(+Goal, -Arg) is semidet.
 %
 %	True if Goal is a meta-predicate for which we wish to format the
@@ -529,5 +536,6 @@ pprint(Out, Term, Pri) :-
 	write_term(Out, Term,
 		   [ quoted(true),
 		     numbervars(true),
-		     priority(Pri)
+		     priority(Pri),
+		     spacing(next_argument)
 		   ]).
