@@ -35,9 +35,10 @@
 	    predicate_name/2,		% +Head, -Name
 	    clause_name/2		% +ClauseRef, -Name
 	  ]).
-:- use_module(library(debug), [debug/3]).
 :- use_module(library(lists), [append/3]).
 :- use_module(library(occurs), [sub_term/2]).
+:- use_module(library(debug)).
+:- use_module(library(listing)).
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 This module started life as part of the   GUI tracer. As it is generally
@@ -64,7 +65,7 @@ clause_info(ClauseRef, File, TermPos, NameOffset) :-
 	),
 	File \== user,			% loaded using ?- [user].
 	clause_property(ClauseRef, line_count(LineNo)),
-	(   current_module(Module, File)
+	(   module_property(Module, file(File))
 	->  true
 	;   strip_module(user:Head, Module, _)
 	),
@@ -77,13 +78,13 @@ clause_info(ClauseRef, File, TermPos, NameOffset) :-
 	debug(clause_info, 'got names~n', []), !.
 
 %%	unify_term(+T1, +T2)
-%	
+%
 %	Unify the two terms, where T2 is created by writing the term and
 %	reading it back in, but  be   aware  that  rounding problems may
 %	cause floating point numbers not to  unify. Also, if the initial
 %	term has a string object, it is written   as "..." and read as a
 %	code-list. We compensate for that.
-%	
+%
 %	NOTE: Called directly from  library(trace/clause)   for  the GUI
 %	tracer.
 
@@ -105,7 +106,10 @@ unify_term(_:X, Y) :-
 unify_term(X, _:Y) :-
 	unify_term(X, Y), !.
 unify_term(X, Y) :-
-	format('[INTERNAL ERROR: Diff: ~q <-> ~q]~n', [X, Y]),
+	format('[INTERNAL ERROR: Diff:~n'),
+	portray_clause(X),
+	format('~N*** <->~n'),
+	portray_clause(Y),
 	break.
 
 unify_args(N, N, _, _) :- !.
@@ -141,11 +145,11 @@ read(Line, Handle, Module, Clause, TermPos, VarNames) :-
 	read(Handle, Module, Clause, TermPos, VarNames).
 
 %%	read(+Stream, +Module, -Clause, -TermPos, -VarNames)
-%	
+%
 %	Read clause from Stream at current position with unknown syntax.
 %	It returns the term read  at  that   position,  as  well  as the
 %	subterm position info and variable names.
-%	
+%
 %	NOTE: Called directly from  library(trace/clause)   for  the GUI
 %	tracer.
 
@@ -172,14 +176,14 @@ read(Handle, Module, Clause, TermPos, VarNames) :-
 	;   set_stream_position(Handle, Here),
 	    fail
 	).
-	
+
 
 %%	make_varnames(+ReadClause, +Offsets, +Names, -Term) is det.
-%	
+%
 %	Create a Term varnames(...) where each argument contains the name
 %	of the variable at that offset.  If the read Clause is a DCG rule,
 %	name the two last arguments <DCG_list> and <DCG_tail>
-%	
+%
 %	@param Offsets	List of Offset=Var
 %	@param Names	List of Name=Var
 %
@@ -215,17 +219,17 @@ find_varname(Var, [_|T], Name) :-
 	find_varname(Var, T, Name).
 
 %%	unify_clause(+Read, +Decompiled, +ReadTermPos, -RecompiledTermPos).
-%	
+%
 %	What you read isn't always what goes into the database. The task
-%	of this predicate is to establish the relation between the term
+%	of this predicate is to establish  the relation between the term
 %	read from the file and the result from decompiling the clause.
-%	
-%	This really must be more flexible, dealing with much more
-%	complex source-translations, falling back to a heristic method
+%
+%	This really must be  more  flexible,   dealing  with  much  more
+%	complex source-translations, falling back to  a heuristic method
 %	locating as much as possible.
 
 unify_clause(Read, Read, TermPos, TermPos) :- !.
-					% XPCE send-methods 
+					% XPCE send-methods
 unify_clause(:->(Head, Body), (PlHead :- PlBody), TermPos0, TermPos) :- !,
 	pce_method_clause(Head, Body, PlHead, PlBody, TermPos0, TermPos).
 					% XPCE get-methods
@@ -293,19 +297,20 @@ expand_failed(E, Read) :-
 	fail.
 
 %%	unify_body(+Read, +Decompiled, +Pos0, -Pos)
-%	
+%
 %	Deal with translations implied by the compiler.  For example,
 %	compiling (a,b),c yields the same code as compiling a,b,c.
-%	
+%
 %	Pos0 and Pos still include the term-position of the head.
 
 unify_body(B, B, Pos, Pos) :-
+	acyclic_term(B),		% X = call(X)
 	\+ sub_term(brace_term_position(_,_,_), Pos), !.
 unify_body(R, D,
 	   term_position(F,T,FF,FT,[HP,BP0]),
 	   term_position(F,T,FF,FT,[HP,BP])) :-
 	ubody(R, D, BP0, BP).
-	   
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Some remarks.
 
@@ -321,7 +326,11 @@ a --> { x, y, z }.
 %	@param TermPosRead	Sub-term positions of source
 
 ubody(B, B, P, P) :-
+	acyclic_term(B),		% X = call(X)
 	\+ sub_term(brace_term_position(_,_,_), P), !.
+ubody(X, call(X),			% X = call(X)
+      From-To,
+      term_position(From, To, From, To, [From-To])) :- !.
 ubody(B0, B,
       brace_term_position(F,T,A0),
       Pos) :-
@@ -348,7 +357,27 @@ ubody(X0, X,
 	X0 =.. [_|A0],
 	X  =.. [_|A],
 	ubody_list(A0, A, PA0, PA).
-	
+					% 5.7.X optimizations
+ubody(_=_, true,			% singleton = Any
+      term_position(F,T,_FF,_TT,_PA),
+      F-T) :- !.
+ubody(_==_, fail,			% singleton/firstvar == Any
+      term_position(F,T,_FF,_TT,_PA),
+      F-T) :- !.
+ubody(A1=B1, B2=A2,			% Term = Var --> Var = Term
+      term_position(F,T,FF,TT,[PA1,PA2]),
+      term_position(F,T,FF,TT,[PA2,PA1])) :-
+	(A1==B1) =@= (B2==A2), !,
+	A1 = A2, B1=B2.
+ubody(A1==B1, B2==A2,			% const == Var --> Var == const
+      term_position(F,T,FF,TT,[PA1,PA2]),
+      term_position(F,T,FF,TT,[PA2,PA1])) :-
+	(A1==B1) =@= (B2==A2), !,
+	A1 = A2, B1=B2.
+ubody(A is B - C, A is B + C2, Pos, Pos) :-
+	integer(C),
+	C2 =:= -C, !.
+
 ubody_list([], [], [], []).
 ubody_list([G0|T0], [G|T], [PA0|PAT0], [PA|PAT]) :-
 	ubody(G0, G, PA0, PA),
@@ -445,7 +474,7 @@ pce_unify_head_arg(A:_, A).
 %	on an isolated method.
 %
 %	TermPos0 is the term-position term of the whole clause!
-%	
+%
 %	Further, please note that the body of the method-clauses reside
 %	in another module than pce_principal, and therefore the body
 %	starts with an I_CONTEXT call. This implies we need a
@@ -484,7 +513,7 @@ pce_method_body2(A0, A, TermPos0, TermPos) :-
 pce_method_body2(A0, A, TermPos0, TermPos) :-
 	expand_goal(A0, A, TermPos0, TermPos).
 
-control_op((,)).
+control_op(',').
 control_op((;)).
 control_op((->)).
 control_op((*->)).
@@ -588,8 +617,8 @@ predicate_name(Predicate, PName) :-
 	->  true
 	;   functor(Head, Name, Arity),
 	    (   hidden_module(Module)
-	    ->  sformat(PName, '~q/~d', [Name, Arity])
-	    ;   sformat(PName, '~q:~q/~d', [Module, Name, Arity])
+	    ->  format(string(PName), '~q/~d', [Name, Arity])
+	    ;   format(string(PName), '~q:~q/~d', [Module, Name, Arity])
 	    )
 	).
 
@@ -603,14 +632,14 @@ clause_name(Ref, Name) :-
 	nth_clause(Head, N, Ref), !,
 	predicate_name(Head, PredName),
 	thaffix(N, Th),
-	sformat(Name, '~d-~w clause of ~w', [N, Th, PredName]).
+	format(string(Name), '~d-~w clause of ~w', [N, Th, PredName]).
 clause_name(_, '<meta-call>').
 
 
 		 /*******************************
 		 *        LOW-LEVEL STUFF	*
 		 *******************************/
-	
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 These predicates communicate about lines.  We   should  consider using a
 line-cache for this for speed.
