@@ -118,9 +118,9 @@ update_state([]) :- !.
 update_state([H|T]) :- !,
 	update_state(H),
 	update_state(T).
-update_state(:- Directive) :- !,
+update_state((:- Directive)) :- !,
 	update_directive(Directive).
-update_state(?- Directive) :- !,
+update_state((?- Directive)) :- !,
 	update_directive(Directive).
 update_state(_).
 
@@ -132,6 +132,8 @@ update_directive(op(P,T,N)) :- !,
 	push_op(P,T,SM:N).
 update_directive(style_check(Style)) :-
 	style_check(Style), !.
+update_directive(expects_dialect(sicstus)) :-
+	style_check(-atom), !.
 update_directive(_).
 
 public_operators([]).
@@ -173,8 +175,8 @@ prolog_open_source(Src, Fd) :-
 	),
 	push_operators([]),
 	'$set_source_module'(SM, SM),
-	'$style_check'(Style, Style),
-	asserta(open_source(Fd, state(Style, SM))).
+	'$save_lex_state'(LexState),
+	asserta(open_source(Fd, state(LexState, SM))).
 
 
 %%	prolog_close_source(+In:stream) is det.
@@ -184,8 +186,8 @@ prolog_open_source(Src, Fd) :-
 
 prolog_close_source(In) :-
 	pop_operators,
-	(   retract(open_source(In, state(Style, SM)))
-	->  '$style_check'(_, Style),
+	(   retract(open_source(In, state(LexState, SM)))
+	->  '$restore_lex_state'(LexState),
 	    '$set_source_module'(_, SM)
 	;   assertion(fail)
 	),
