@@ -1,10 +1,3 @@
-/*  Part of SWI-Prolog
-
-    Author:        Nicos Angelopoulos
-    WWW:           http://www.swi-prolog.org
-    Copyright (C): 2008, Nicos Angelopoulos
-*/
-
 :- module( r_session,
           [
                r_open/0, r_open/1,
@@ -23,52 +16,120 @@
                r_session_data/3, r_streams_data/3,
                r_history/0, r_history/1, r_history/2,
                r_session_version/1,
-	   op( 950, xfx, (<-) )
+               r_bin/1,
+               r_bin_version/1, r_bin_version/2,
+               r_verbosity/1,
+               op( 950, xfx, (<-) )
           ] ).
 
 :- use_module( library(lists) ).
 :- use_module( library(readutil) ). % read_line_to_codes/2.
 
+:- ( current_predicate(r_verbosity_level/1) -> true;
+          assert(r_verbosity_level(0)) ).
+
+:- dynamic( r_bin_location/1 ).
+:- dynamic( r_session/3 ).
+:- dynamic( r_session_history/2 ).
+:- dynamic( r_old_bin_warning_issued/1 ).
+
 % Swi declaration:
-:- use_module( library(process) ).   % process_create/3.
+:- ensure_loaded( library(process) ).   % process_create/3.
 :- at_halt( r_close(all) ).
 % end of Swi declaration.
 
-
-:- dynamic( r_session/3 ).
-:- dynamic( r_session_history/2 ).
-
-
 /** <module> R session
 
-This library facilitates interaction with an R session. On the Yap
-system it depends on library(System) and on SWI on library(process)-
-part of the clib package. Currently it only works on Linux systems. It
-assumes an R executable in $PATH or can be given a location to a
-functioning R executable. R is run as a slave with Prolog writing and
-reading on/off the associated streams.
-
-Multiple session can be managed simultaneously. Each has 3 main
-components: a name or alias, a term structure holding the communicating
-streams and a number of associated data items.
+This library facilitates interaction with the R system for statistical
+computing.  It assumes an R executable in $PATH or can be given a location
+to a functioning R executable (see r_bin/1 and r_open/1 for details on how
+R is located). R is ran as a slave with Prolog writing on and reading from
+the associated streams. Multiple sessions can be managed simultaneously.
+Each has 3 main components: a name or alias, a term structure holding the
+communicating streams and a number of associated data items.
 
 The library attempts to ease the translation between prolog terms and R
 inputs. Thus, Prolog term =|x <- c(1,2,3)|= is translated to atomic =|'x
 <- c(1,2,3)'|= which is then passed on to R. That is, =|<-|= is a
 defined/recognised operator. =|X <- c(1,2,3)|=, where X is a variable,
-instantiates X to the list =|[1,2,3]|=. Currently only vectors can be
+instantiates X to the list =|[1,2,3]|=. Also 'Atom' <- [x1,...,xn]
+translates to R code: Atom <- c(x1,...,xn). Currently only vectors can be
 translated in this fashion.
 
+Although the library is primarily meant to be used as a research tool,
+it still provides access to many functions of the R system that may render it 
+useful to a wider audience. The library provides access to R's plethora of vector and scalar
+functions. We adicipate that of particular interest to Prolog programmers might be the fact
+that the library can be used to create plots from Prolog objects.
+Notably creating plots from lists of numbers.
+
+These capabilities are illustrated in the following example :
+
+==
+rtest :-
+     r_open,
+     r_in( y <- rnorm(50) ),
+     r_print( y ),
+     r_in( x <- rnorm(y) ),
+     r_in( x11(width=5,height=3.5) ),
+     r_in( plot(x,y)),
+     write( 'Press Return to continue...' ), nl,
+     read_line_to_codes( user, _ ),
+     r_print( 'dev.off()' ),
+     r_in( Y <- y ),
+     write( y(Y) ), nl,
+     Z = [1,2,3,4,5,6,7,8,9],
+     r_in( z <- Z ),
+     r_print( z ),
+     r_close.
+==
+
 @author 	Nicos Angelopoulos
-@version	0:0:1
+@version	0:0:3
 @copyright	Nicos Angelopoulos
 @license	YAP: Artistic
-@see		examples/R/r_demo.pl, http://www.r-project.org/
-@tbd		Fix starting the R process on Windows.
+@see		ensure_loaded(library('../doc/packages/examples/R/r_demo.pl'))
+@see		http://www.r-project.org/
+@author		Windows-compatibility is based on work by `JAB'
 */
 
 %%% Section: Interface predicates
 
+
+%% r_bin( ?Rbin )
+%
+%   Register the default R location, +Rbin, or interrogate the current location: -Rbin.
+%   When interrogating Rbin is bound to the R binary that would be used by an r_open/0.
+%   The order of search is: registered location, environment variable 'R_BIN' and path defined.
+%   On unix systems path defined is the first R executable in $PATH. On MS wins it is the latest
+%   Rterm.exe found by expand_file_name( 'C:/Program Files/R/R-*/bin/Rterm.exe', Candidates ).
+%   The value Rbin == =retract= retracts the current registered location. 
+%   Rbin == =test=, succeeds if an R location has been registered.
+%
+r_bin( Rbin ) :-
+     var( Rbin ),
+     !,
+     ( r_bin_location(Rbin) ->
+          true
+          ;
+          ( locate_rbin_file(Rbin) ->
+               M = 'There is no registered R executable. Using the one found by searching.',
+               r_verbose( M, 1 )
+               ;
+               M = 'There is no registered or default R executatble. Use, r_bin(+Rbin).',
+               fail_term( M )
+          )
+     ).
+r_bin( retract ) :-
+     !,
+     retractall( r_bin_location(_) ).
+r_bin( test ) :-
+     !,
+     r_bin_location(_).
+r_bin( Rbin ) :-
+     retractall( r_bin_location(_) ),
+     assert( r_bin_location(Rbin) ).
+
 %%	r_open
 %
 %	Open a new R session. Same as r_open( [] ).
@@ -76,8 +137,6 @@ translated in this fashion.
 r_open :-
      r_open( [] ).
 
-/* Opts = [alias/1,assert/1,at_r_halt/1,copy/2,rbin/1,with/1] */
-
 %% r_open( +Opts )
 %
 %   Open a new R session with optional list of arguments. Opts should be
@@ -93,9 +152,11 @@ r_open :-
 %	session to the bottom of the pile.
 %
 %	* at_r_halt(RHAction)
-%	R slaves often halt when they   encounter  an error. This option
-%	provides a handle to changing the  behaviour of the session when
-%	this happens. RHAction should be one of =abort=, =fail=, call/1,
+%	R slaves used to halt when they encounter  an error. 
+%    This is no longer the case but this option is still present in case
+%    it is useful in the future. This option provides a handle to changing
+%    the  behaviour of the session when a halt of the R-slave occurs.
+%    RHAction should be one of =abort=, =fail=, call/1,
 %	call_ground/1, =reinstate= or =restart=. Default is =fail=. When
 %	RHAction is =reinstate=, the history of   the session is used to
 %	roll-back all the commands sent so far. At `restart' the session
@@ -107,22 +168,26 @@ r_open :-
 %	one   of   =null=,   stream(Stream),   OpenStream,   AtomicFile,
 %	once(File) or many(File). In the  case   of  many(File), file is
 %	opened and closed at each write   operation.  CopyWhat should be
-%	one of =both=, =in=, =out= or =none=.   In  all cases apart from
-%	when CopyTo is =null=, error stream is copied to CopyTo. Default
-%	is no recording (CopyTo = =null=).
+%	one of =both=, =in=, =out= or =none=. Default is no recording
+%    (CopyTo = =null=).
 %
 %       * ssh(Host)
 %       * ssh(Host,Dir)
 %       Run R on Host with start directory Dir. Dir defaults to /tmp.
+%       Not supported on MS Windows.
+%
 %
 %       * rbin(Rbin)
-%       R executable location. Default is 'R'.
+%       R executable location to use for this open operation.
+%       If the option is not present binary registered with r_bin/1 and
+%       environment variable R_BIN are examined for the full location of
+%       the R binary. In MS windows Rbin should point to Rterm.exe. Also see r_bin/1.
 %
 %       * with(With)
-%	With is in [environ,restore,save]. The   default behaviour is to
-%	start the R executable  is   started  with  flags =|--no-environ
-%	--no-restore --no-save|=. For each With value  found in Opts the
-%	corresponding =|--no-|= flag is removed.
+%       With is in [environ,restore,save]. The   default behaviour is to
+%       start the R executable  with  flags =|--no-environ --no-restore
+%       --no-save|=. For each With value  found in Opts the corresponding
+%       =|--no-|= flag is removed.
 %
 r_open( Opts ) :-
      r_open_1( Opts, _R, false ).
@@ -146,7 +211,7 @@ r_close( All ) :-
      All == all,
      !,
      findall( Alias, ( retract( r_session(Alias,Streams,Data) ),
-                       r_close_session( Alias, Streams, Data ) ), _All ).
+                       r_close_session( Alias, Streams, Data ) ), _AllAls ).
      % write( closed_all(All) ), nl.
 r_close( Alias ) :-
      ( retract( r_session(Alias,Streams,Data) ) ->
@@ -324,12 +389,12 @@ r_flush_onto( RinStreamS, OntoS ) :-
 %
 %         As r_flush_onto/2 for specified session R.
 %
-r_flush_onto( R, RinStreamS, OntoS ) :-
-     ( is_list(RinStreamS) -> RinStreams = RinStreamS; RinStreams=[RinStreams] ),
+r_flush_onto( R, RinStreams, Ontos ) :-
+     ( is_list(RinStreams) -> RStreams = RinStreams; RStreams=[RinStreams] ),
      % to_list( RinStreamS, RinStreams ),
-     r_input_streams_list( RinStreams ),
-     r_flush_onto_1( RinStreams, R, Ontos ),
-     ( is_list(RinStreamS) -> OntoS = Ontos; Ontos=[OntoS] ).
+     r_input_streams_list( RStreams ),
+     r_flush_onto_1( RStreams, R, ROntos ),
+     ( is_list(RinStreams) -> Ontos = ROntos; Ontos=[ROntos] ).
 
 %%   current_r_session(?R)
 %         True if R is the name of current R session.
@@ -428,38 +493,110 @@ r_history( R, History ) :-
 %         Installed version. Version is of the form Major:Minor:Fix,
 %         where all three are integers.
 %
-r_session_version( 0:0:1 ).
+r_session_version( 0:0:3 ).
+
+%% r_verbose( What, CutOff )
+%
+r_verbose( What, CutOff ) :-
+     r_verbosity_level( Level ),
+     ( CutOff > Level ->
+          true
+          ;
+          write( What ), nl
+     ).
+
+%% r_verbosity( ?Level )
+%
+%    Set, +Level, or interrogate, -Level, the verbosity level. +Level could be
+%    =false= (=0), =true= (=3) or an integer in {0,1,2,3}. 3 being the most verbose.
+%    The default is 0. -Level will instantiate to the current verbosity level,
+%    an integer in {0,1,2,3}.
+%
+r_verbosity( Level ) :-
+     var( Level ),
+     !,
+     r_verbosity_level( Level ).
+r_verbosity( Level ) :-
+     ( Level == true ->
+          Numeric is 3
+          ;
+          ( Level == false ->
+               Numeric is 0
+               ;
+               ( integer(Level) ->
+                    ( Level < 0 ->
+                         write( 'Adjusting verbosity level to = 0. ' ), nl,
+                         Numeric is 0
+                         ;
+                         ( Level > 3 ->
+                              write( 'Adjusting verbosity level to = 3. ' ), nl,
+                              Numeric is 3
+                              ;
+                              Numeric is Level
+                         )
+                    )
+                    ;
+                    fail_term( 'Unknown verbosity level. Use : true, false, 0-3' )
+               )
+          )
+     ),
+     retractall( r_verbosity_level(_) ),
+     assert( r_verbosity_level(Numeric) ).
+
+%% r_bin_version( -Version)
+%
+%    Get the version of R binary identified by r_bin/1. Version will have the
+%    same structure as in r_session_version/1 ie M:N:F.
+%
+r_bin_version( Version ) :-
+     r_bin( R ), 
+     r_bin_version( R, Version ).
+
+%% r_bin_version( +Rbin, -Version )
+%
+%    Get the version of R binary identified by +Rbin. Version will have the
+%    same structure as in r_session_version/1 ie M:N:F.
+%
+r_bin_version( R, Version ) :-
+     r_bin_version_pl( R, Version ).
 
 %%% Section: Auxiliary predicates
 
 % Rcv == true iff r_open_1/3 is called from recovery.
 %
 r_open_1( Opts, Alias, Rcv ) :-
-     ( memberchk(rbin(RBin),Opts)->
-          ( options_have_ssh(Opts,Host,Dir) ->
-               atoms_concat( ['ssh ', Host,' "cd ',Dir,'; ',RBin], RPfx ),
-               RPsf = '"'
+     ( options_have_ssh(Opts,Host,Dir) ->
+          ( current_prolog_flag(windows,true) ->
+               fail_term( ssh_option_not_supported_on_ms_windows )
                ;
-               RPfx = RBin,
-               RPsf = ''
+               which( ssh, Ssh )
           )
           ;
-          ( options_have_ssh(Opts,Host,Dir) ->
-               atoms_concat( ['ssh ', Host,' "cd ',Dir,'; R '], RPfx ),
-               RPsf = '"'
-               ;
-               ( shell('sh -c "which R &> /dev/null"  ') ->
-                    RPfx = 'R',
-                    RPsf = ''
-                    ;
-                    fail_term( cannot_locate_r_executable )
-               )
-          )
+          true
      ),
-     r_opt_with_exec( RPfx, Opts, PrvExec ),
-     atom_concat( PrvExec, RPsf, Exec ),
-     r_process( Exec, Ri, Ro, Re ),
-     R = r(Ri,Ro,Re),
+     ( (memberchk(rbin(Rbin),Opts);locate_rbin(Ssh,Rbin)) ->
+          true
+          ;
+          fail_term( 'Use rbin/1 in r_open/n, or r_bin(\'Rbin\') or set R_BIN.' )
+     ),
+     r_bin_arguments( Opts, Rbin, OptRArgs ),
+     % ( var(Harg) -> RArgs = OptRArgs; RArgs = [Host,Harg|OptRArgs] ),
+     ( var(Ssh) ->
+          Exec = Rbin,
+          Args = OptRArgs
+          ;
+          Exec = Ssh,
+          % atoms_concat( [' "cd ',Dir,'; ',Rbin,'"'], Harg ),
+          atoms_concat( ['cd ',Dir,'; '], Cd ),
+          PreArgs = [Cd,Rbin|OptRArgs],
+          double_quote_on_yap( PreArgs, TailArgs ),
+          Args = [Host|TailArgs]
+          % atoms_concat( ['ssh ', Host,' "cd ',Dir,'; ',RBin,'"'], R )
+     ),
+     % atom_concat( PrvExec, RPsf, Exec ),
+     r_verbose( r_process( Exec, Args, Ri, Ro, Re ), 3 ),
+     r_process( Exec, Args, Ri, Ro, Re ),
+     RStreams = r(Ri,Ro,Re),
      r_streams_set( Ri, Ro, Re ),
      r_process_was_successful( Ri, Ro, Re ),
      r_open_opt_copy( Opts, CpOn, CpWh, Rcv ),
@@ -480,16 +617,16 @@ r_open_1( Opts, Alias, Rcv ) :-
      RData = rsdata(CpOn,CpWh,RHalt,Opts),
      ( memberchk(assert(Assert),Opts) ->
           ( Assert == a ->
-               asserta( r_session(Alias,R,RData) )
+               asserta( r_session(Alias,RStreams,RData) )
                ;
                ( Assert == z ->
-                    assertz( r_session(Alias,R,RData) )
+                    assertz( r_session(Alias,RStreams,RData) )
                     ;
                     fail_term( 'Cannot decipher argument to assert/1 option':Assert )
                )
           )
           ;
-          asserta( r_session(Alias,R,RData) )
+          asserta( r_session(Alias,RStreams,RData) )
      ),
      AtRH = at_r_halt(reinstate),
      ( (memberchk(history(false),Opts),\+memberchk(AtRH,Opts)) ->
@@ -543,7 +680,8 @@ r_push( R, RCmd, Rplc, RoLns, ReLns, Halt, HCall ) :-
           r_lines( Streams, output, RoLns )
      ),
      r_record_lines( RoLns, output, CopyTo ),
-     r_record_lines( ReLns, error, CopyTo ).
+     r_record_lines( ReLns, error, CopyTo ),
+     ( (Halt==true,CopyTo=stream(Cl)) -> close(Cl); true ).
 
 r_out_halted_record( true, _Alias, [] ).
 r_out_halted_record( false, _Alias, Lines ) :-
@@ -597,6 +735,14 @@ r_input_normative( (A;B), R, I, This, Rplc, OutI ) :-
      atoms_concat( [ThisA,'; ',ThisB], This ),
      append( RplcA, RplcB, Rplc ).
 
+% r_input_normative( Obj<-List, _R, I, This, Rplc, NxI ) :-
+     % % atomic( Obj ),
+     % is_list( List ),
+     % !,
+     % Rplc = [],
+     % NxI is I,
+     % pl_list_to_r_combine( List,
+
 r_input_normative( Obj<-Call, R, I, This, Rplc, NxI ) :-
      !,
      ( var(Obj) ->
@@ -625,6 +771,10 @@ r_input_normative( Opt=Val, This ) :-
      r_input_normative( Val, ThisVal ),
      atoms_concat( [ThisOpt,'=',ThisVal], This ).
 % 2008ac06, careful! we are changing behaviour here
+r_input_normative( List, This ) :-
+     is_list( List ),
+     pl_list_to_r_combine( List, This ),
+     !.
 r_input_normative( PrvThis, This ) :-
      ( (\+ var(PrvThis),(PrvThis = [_|_];PrvThis=[])) ->
           append( PrvThis, [0'"], ThisRight ),
@@ -660,6 +810,32 @@ r_input_normative_tuple( [H|T], Tuple ) :-
      ( Psf == '' -> Tuple = HNorm
         ; atoms_concat([HNorm,',',Psf], Tuple) ).
 
+pl_list_to_r_combine( [H|T], This ) :-
+     number_atom_to_atom( H, Hatm ),
+     atom_concat( 'c(', Hatm, Pfx ),
+     pl_list_to_r_combine( T, Pfx, This ).
+
+pl_list_to_r_combine( [], Pfx, This ) :-
+     atom_concat( Pfx, ')', This ).
+pl_list_to_r_combine( [H|T], Pfx, This ) :-
+     number_atom_to_atom( H, Hatm ),
+     atom_concat( Pfx, ',', PfxComma ),
+     atom_concat( PfxComma, Hatm, Nxt ),
+     pl_list_to_r_combine( T, Nxt, This ).
+
+number_atom_to_atom( NorA, Atom ) :-
+     number_atom_to_codes( NorA, Codes ),
+     atom_codes( Atom, Codes ).
+
+number_atom_to_codes( NorA, Codes ) :-
+     number( NorA ),
+     !,
+     number_codes( NorA, Codes ).
+number_atom_to_codes( NorA, Codes ) :-
+     atom( NorA ),
+     !,
+     atom_codes( NorA, Codes ).
+
 r_read_lines( Ro, TermLine, Lines ) :-
      read_line_to_codes( Ro, Line ),
      r_read_lines_1( Line, TermLine, Ro, Lines ).
@@ -688,14 +864,19 @@ r_halted_recovery( [rs(AliasH,StreamsH,DataH)|T], R, Which ) :-
      ),
      r_halted_recovery( T, R, Which ).
 
-r_halted_recovery_action( restart, Alias, _Streams, Data, true ) :-
-     write( user_error, 'at_r_halt(restart): restarting r_session ':Alias ), nl( user_error ),
+r_halted_recovery_action( restart, Alias, _Streams, Data, RecCall ) :-
+     Mess = 'at_r_halt(restart): restarting r_session ':Alias,
+     RecCall = (write( user_error, Mess ),nl( user_error )),
      r_session_data( opts, Data, Opts ),
-     ( memberchk(copy(CopyTo,_),Opts) -> r_halted_restart_copy(CopyTo); true ),
+     ( memberchk(copy(CopyTo,_),Opts) ->
+          r_halted_restart_copy(CopyTo)
+          ;
+          true
+     ),
      r_open_1( Opts, Alias, true ),
      current_r_session( Alias, Streams, _ ),
      r_lines( Streams, output, _ReLines ).
-r_halted_recovery_action( reinstate, Alias, _Streams, Data, true ) :-
+r_halted_recovery_action( reinstate, Alias, _Streams, Data, RecCall ) :-
      ( r_session_history(Alias,History) ->
           r_session_data( opts, Data, Opts ),
           r_open_1( Opts, Alias, true ),
@@ -703,12 +884,24 @@ r_halted_recovery_action( reinstate, Alias, _Streams, Data, true ) :-
           r_halted_recovery_rollback( Hicory, Alias )
           ;
           fail_term( 'at_r_halt(reinstate): cannnot locate history for':Alias )
-     ).
-r_halted_recovery_action( abort, _Alias, _Streams, _Data, abort ) :-
-     write( user_error, 'at_r_halt(abort): R session halted by slave' ), nl( user_error ).
+     ),
+     Mess = 'at_r_halt(reinstate): reinstating r_session ':Alias,
+     RecCall = (write( user_error, Mess ), nl( user_error ) ).
+r_halted_recovery_action( abort, _Alias, _Streams, _Data, RecCall ) :-
+     Mess = 'at_r_halt(abort): R session halted by slave',
+     RecCall = (write( user_error, Mess ),nl( user_error ),abort).
 r_halted_recovery_action( fail, Alias, _Streams, _Data, Call ) :-
      retractall( r_session_history(Alias,_) ),
-     Call = fail_term( 'at_r_halt(fail): failure due to execution halted by slave on r_session':Alias ).
+     % % r_session_data( copy_to, Data, CopyTo ),
+     % write( copy_to(CopyTo) ), nl,
+     % ( CopyTo = stream(Stream) ->
+          % close(Stream)
+          % ;
+          % true
+     % ),
+     % trace,
+     L='at_r_halt(fail): failure due to execution halted by slave on r_session',
+     Call = fail_term( L:Alias ).
 r_halted_recovery_action( call(Call), _Alias, Streams, _Data, Call ) :-
      Call = call( Call, Streams ).
 r_halted_recovery_action( call_ground(Call), _Alias, _Streams, _Data, Call) :-
@@ -722,7 +915,7 @@ r_halted_restart_copy( CopyTo ) :-
      stream_property( Dummy, file_name(Full) ),
      close( Dummy ),
      ( stream_property(OpenStream,file_name(Full)) ->
-          write( close( OpenStream ) ), nl,
+          write( close(OpenStream) ), nl,
           close( OpenStream )
           ;
           true
@@ -860,7 +1053,8 @@ r_open_opt_copy( Opts, CpTerm, What, Rcv ) :-
 
 r_open_opt_at_r_halt( Opts, RHalt ) :-
      ( memberchk(at_r_halt(RHalt),Opts) ->
-          ( memberchk(RHalt,[restart,reinstate,fail,abort,call(_),call_ground(_)]) ->
+          Poss = [restart,reinstate,fail,abort,call(_),call_ground(_)],
+          ( memberchk(RHalt,Poss) ->
                true
                ;
                fail_term( 'Cannot decipher argument to at_r_halt option':RHalt )
@@ -869,19 +1063,30 @@ r_open_opt_at_r_halt( Opts, RHalt ) :-
           RHalt = fail
      ).
 
-r_opt_with_exec( _RBin, Opts, _Exec ) :-
+r_bin_arguments( Opts, _Rbin, _RArgs ) :-
      member( with(With), Opts ),
      \+ memberchk(With, [environ,restore,save] ),
      !,
      fail_term( 'Cannot decipher argument to option with/1': With ).
-r_opt_with_exec( RBin, Opts, Exec ) :-
+r_bin_arguments( Opts, Rbin, Args ) :-
+     ( current_prolog_flag(windows,true) ->
+          Args = ['--ess','--slave'|RArgs]
+          ; % assuming unix here, --interactive is only supported on these
+          r_bin_version( Rbin, Version ),
+          % Issue is present on 2:9:2
+          ( 2:9:10 @< Version ->
+               Args = ['--interactive','--slave'|RArgs]
+               ;
+               r_old_bin_warning( Version ),
+               Args = ['--slave'|RArgs]
+          )
+     ),
      findall( W, member(with(W),Opts), Ws ),
      sort( Ws, Sr ),
      length( Ws, WsL ),
      length( Sr, SrL ),
      ( WsL =:= SrL ->
-          atom_concat( RBin, ' --slave ', Pfx ),
-          r_opt_with_concat( [environ,restore,save], Ws, Pfx, Exec )
+          r_bin_arguments_complement( [environ,restore,save], Ws, RArgs )
           ;
           fail_term( 'Multiple identical args in with/1 option': Ws )
      ).
@@ -892,20 +1097,21 @@ r_opt_exec_no( [H|T], Ws, Exec ) :-
      ( memberchk(H,Ws) ->
           TExec=Exec
           ;
-          atom_concat( ' --no-', H, NoH ),
+          atom_concat( '--no-', H, NoH ),
           Exec=[NoH|TExec]
      ),
      r_opt_exec_no( T, Ws, TExec ).
 
-r_opt_with_concat( [], _Ws, Exec, Exec ).
-r_opt_with_concat( [H|T], Ws, Pfx, Exec ) :-
+r_bin_arguments_complement( [], _Ws, [] ).
+r_bin_arguments_complement( [H|T], Ws, Args ) :-
      ( memberchk(H,Ws) ->
-          Nxt = Pfx
+          % we could add explicit --with- here ?
+          Args = TArgs
           ;
-          atom_concat( ' --no-', H, NoH ),
-          atom_concat( Pfx, NoH, Nxt )
+          atom_concat( '--no-', H, NoH ),
+          Args = [NoH|TArgs]
      ),
-     r_opt_with_concat( T, Ws, Nxt, Exec ).
+     r_bin_arguments_complement( T, Ws, TArgs ).
 
 r_record_lines( [], _Type, _CopyTo ) :- !.
 r_record_lines( Lines, Type, CopyTo ) :-
@@ -930,8 +1136,21 @@ copy_stream_open( stream(CopyStream), CopyStream ).
 copy_stream_open( file(File), CopyStream ) :-
      open( File, append, CopyStream ).
 
-copy_stream_close( file(CopyTo) ) :- close( CopyTo ).
-copy_stream_close( stream(_) ).
+copy_stream_close( Atom ) :-
+     atomic( Atom ),
+     !,
+     ( Atom == user ->
+          true
+          ;
+          close( Atom )
+     ).
+copy_stream_close( CopyTo ) :-
+     copy_stream_close_non_atomic( CopyTo ).
+
+copy_stream_close_non_atomic( file(CopyTo) ) :- close( CopyTo ).
+copy_stream_close_non_atomic( once(CopyTo) ) :- close( CopyTo ).
+copy_stream_close_non_atomic( many(CopyTo) ) :- close( CopyTo ).
+copy_stream_close_non_atomic( stream(_) ).
 
 /*
 write_list_to_comma_separated( [], _Sep, _Out ).
@@ -1028,6 +1247,72 @@ options_have_ssh( Opts, Host, Dir ) :-
           memberchk( ssh(Host,Dir), Opts )
      ).
 
+locate_rbin( Ssh, RBin ) :-
+     locate_rbin_file( File ),
+     ( var(Ssh) ->
+          ( current_prolog_flag(windows,true) -> Exe='exe'; Exe='' ),
+          file_name_extension( File, Exe, RBin ),
+          exists_file( RBin )
+          ;
+          % currently when we using ssh, there is no check for existance
+          % of the binary on the remote host
+          File = RBin
+     ),
+     r_verbose( using_R_bin(RBin), 1 ).
+
+% order of clauses matters. only first existing one to succeed is considered.
+locate_rbin_file( RBin ) :-
+     % current_predicate( r_bin/1 ),
+     r_bin_location( RBin ).
+locate_rbin_file( RBin ) :-
+     environ( 'R_BIN', RBin ).
+locate_rbin_file( RBin ) :-
+     current_prolog_flag( unix, true ),
+     which( 'R', RBin ).
+locate_rbin_file( RBin ) :-
+     current_prolog_flag( windows, true ),
+     r_bin_wins( RBin ).
+
+r_bin_wins( Rbin ) :-
+     r_expand_wins_rterm( Stem, Candidates ),
+     r_verbose( wins_candidates(Candidates), 3 ),
+     Candidates \== [],
+     ( Candidates = [Rbin] -> 
+          true
+          ;
+          maplist( atom_concat(Stem), Tails, Candidates ),
+          maplist( atom_codes, Tails, TailsCs ),
+          cur_tail_candidates_with_pair( TailsCs, Candidates, Pairs ),
+          keysort( Pairs, Sorted ),
+          reverse( Sorted, [_-Rbin|_] )
+     ),
+     !.
+
+cur_tail_candidates_with_pair( [], [], [] ).
+cur_tail_candidates_with_pair( [H|T], [F|R], [Hnum-F|TPairs] ) :-
+     ( break_list_on( H, 0'/, Hlft, _ ) -> true; break_list_on( H, 0'\\, Hlft, _) ),
+     break_list_on( Hlft, 0'., MjCs, NonMjCs ),
+     break_list_on( NonMjCs, 0'., MnCs, FxCs ),
+     maplist( number_codes, Nums, [MjCs,MnCs,FxCs] ),
+     integers_list_to_integer( Nums, 2, 1000, 0, Hnum ),
+     cur_tail_candidates_with_pair( T, R, TPairs ).
+
+integers_list_to_integer( [], _Pow, _Spc, Int, Int ).
+integers_list_to_integer( [H|T], Pow, Spc, Acc, Int ) :-
+     Nxt is Acc + ( H * (Spc ** Pow) ),
+     Red is Pow - 1,
+     integers_list_to_integer( T, Red, Spc, Nxt, Int ).
+
+r_old_bin_warning( Version ) :-
+     ( r_old_bin_warning_issued( true ) ->
+          true
+          ;
+          assert( r_old_bin_warning_issued(true) ),
+          write( 'Flag --interactive which is used by R.pl when starting R, is in your R version, ' ), write( Version ), 
+          write( ', behaving differently than in later versions.' ), nl,
+          write( 'There will be limited functionality of interactive graphics.'),
+          nl, nl
+     ).
 % Section: Swi Specifics.
 
 /*
@@ -1049,12 +1334,50 @@ r_lines( Streams, ROstream, Lines ) :-
 atoms_concat( Atoms, Concat ) :-
      atomic_list_concat( Atoms, Concat ).
 
+which( Which, This ) :-
+     absolute_file_name( path(Which), This,
+			 [ extensions(['',exe]),
+			   access(exist)
+			 ]),
+     r_verbose( which(Which,This), 2 ).
+
 r_streams_set( Ri, Ro, Re ) :-
      set_stream( Ri, buffer(false) ), set_stream( Ri, close_on_abort(true) ),
      set_stream( Ro, buffer(false) ), set_stream( Ro, close_on_abort(true) ),
      set_stream( Re, buffer(false) ), set_stream( Re, close_on_abort(true) ).
 
-r_process( Exec, Ri, Ro, Re ) :-
+r_process( R, Args, Ri, Ro, Re ) :-
      Streams = [stdin(pipe(Ri)),stdout(pipe(Ro)),stderr(pipe(Re))],
-     process_create( '/bin/sh', ['-c',Exec], Streams ).
+     process_create( R, Args, Streams ),
+     r_verbose( created(R,Args,Streams), 3 ).
 
+r_bin_version_pl( R, Mj:Mn:Fx ) :-
+     Streams = [stdout(pipe(Ro))],
+     process_create( R, ['--version'], Streams ),
+     % read_line_to_codes( Ro, _ ),
+     read_line_to_codes( Ro, Codes ),
+     break_list_on( Codes, 0' , _R, Psf1 ),
+     break_list_on( Psf1, 0' , _V, Psf2 ),
+     break_list_on( Psf2, 0' , VersionCs, _ ),
+     break_list_on( VersionCs, 0'., MjCs, VPsf1Cs ),
+     break_list_on( VPsf1Cs, 0'., MnCs, FxCs ),
+     number_codes( Mj, MjCs ),
+     number_codes( Mn, MnCs ),
+     number_codes( Fx, FxCs ).
+     
+r_expand_wins_rterm( Stem, Candidates ) :-
+     Stem = 'C:/Program Files/R/R-',
+     Psfx = '*/bin/Rterm.exe',
+     atom_concat( Stem, Psfx, Search ),
+     expand_file_name( Search, Candidates ).
+
+
+environ( Var, Val ) :-
+     \+ var(Var),
+     ( var(Val) ->
+          getenv(Var,Val)
+          ;
+          setenv(Var,Val)
+     ).
+
+double_quote_on_yap( A, A ).
