@@ -123,6 +123,13 @@ print_chr_error(syntax(Term),Message,Params) :- !,
 	format(user_error,Message,Params),
 	long_line_with_equality_signs.
 
+print_chr_error(type_error,Message,Params) :- !,
+	long_line_with_equality_signs,
+	format(user_error,'CHR compiler TYPE ERROR:\n',[]),	
+	format(user_error,'    `--> ',[]),
+	format(user_error,Message,Params),
+	long_line_with_equality_signs.
+
 print_chr_error(internal,Message,Params) :- !,
 	long_line_with_equality_signs,
 	format(user_error,'CHR compiler ERROR: something unexpected happened in the CHR compiler.\n',[]),	
