@@ -230,10 +230,10 @@ typedef struct _plsocket
     struct
     { int bytes;			/* byte count */
       void *buffer;			/* the buffer */
-      size_t size;			/* buffer size */
+      int size;				/* buffer size */
       int flags;
       const struct sockaddr *to;
-      socklen_t tolen;
+      int tolen;
     } sendto;
   } rdata;
 #endif
@@ -256,7 +256,7 @@ need_retry(int error)
 #ifdef O_DEBUG
 static int debugging;
 
-int
+NBIO_EXPORT(int)
 nbio_debug(int level)
 { int old = debugging;
 
@@ -710,7 +710,9 @@ doRequest(plsocket *s)
       break;
     case REQ_RECVFROM:
       if ( s->w32_flags & (FD_READ|FD_CLOSE) )
-      { s->w32_flags &= ~FD_READ;
+      { int iflen = (int)*s->rdata.recvfrom.fromlen;
+
+	s->w32_flags &= ~FD_READ;
 
 	if ( true(s, PLSOCK_WAITING) )
 	{ doneRequest(s);
@@ -723,7 +725,7 @@ doRequest(plsocket *s)
 			 (int)s->rdata.recvfrom.size,
 			 s->rdata.recvfrom.flags,
 			 s->rdata.recvfrom.from,
-			 s->rdata.recvfrom.fromlen);
+			 &iflen);
 
 	if ( s->rdata.recvfrom.bytes < 0 )
 	{ s->error = WSAGetLastError();
@@ -734,7 +736,8 @@ doRequest(plsocket *s)
 	    doneRequest(s);
 	  }
 	} else
-	{ doneRequest(s);
+	{ *s->rdata.recvfrom.fromlen = iflen;
+	  doneRequest(s);
 	}
       }
       break;
@@ -1201,7 +1204,7 @@ nbio_to_plsocket(nbio_sock_t socket)
 }
 
 
-SOCKET
+NBIO_EXPORT(SOCKET)
 nbio_fd(nbio_sock_t socket)
 { plsocket *p;
 
@@ -1462,6 +1465,9 @@ nbio_error(int code, nbio_error_map mapid)
   term_t except = PL_new_term_ref();
   error_codes *map;
 
+  if ( code == EPLEXCEPTION )
+    return FALSE;
+
   switch( mapid )
   { case TCP_HERRNO:
       map = h_errno_codes;
@@ -1492,11 +1498,12 @@ nbio_error(int code, nbio_error_map mapid)
     msg = strerror(code);
 #endif /*__WINDOWS__*/
 
-  PL_unify_term(except,
-		CompoundArg("error", 2),
-		  CompoundArg("socket_error", 1),
-		    AtomArg(msg),
-		  PL_VARIABLE);
+  if ( !PL_unify_term(except,
+		      CompoundArg("error", 2),
+		        CompoundArg("socket_error", 1),
+		          AtomArg(msg),
+		        PL_VARIABLE) )
+    return FALSE;
   }
 
   return PL_raise_exception(except);
@@ -1524,7 +1531,7 @@ nbio_last_error(nbio_sock_t socket)
 		 *	  INITIALISATION	*
 		 *******************************/
 
-int
+NBIO_EXPORT(int)
 nbio_init(const char *module)
 { INITLOCK();				/* is this ok? */
 
@@ -1567,7 +1574,7 @@ nbio_init(const char *module)
 }
 
 
-int
+NBIO_EXPORT(int)
 nbio_cleanup(void)
 { if ( initialised )
   {
@@ -1586,7 +1593,7 @@ socket(-Socket)
     the format $socket(Id).
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-nbio_sock_t
+NBIO_EXPORT(nbio_sock_t)
 nbio_socket(int domain, int type, int protocol)
 { SOCKET sock;
   plsocket *s;
@@ -1597,7 +1604,6 @@ nbio_socket(int domain, int type, int protocol)
   { nbio_error(errno, TCP_ERRNO);
     return -1;
   }
-
   if ( !(s=allocSocket(sock)) )		/* register it */
   { closesocket(sock);
     return -1;
@@ -1612,7 +1618,7 @@ nbio_socket(int domain, int type, int protocol)
 }
 
 
-int
+NBIO_EXPORT(int)
 nbio_closesocket(nbio_sock_t socket)
 { plsocket *s;
 
@@ -1647,7 +1653,7 @@ nbio_closesocket(nbio_sock_t socket)
   return 0;
 }
 
-int
+NBIO_EXPORT(int)
 nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
 { plsocket *s;
   va_list args;
@@ -1802,7 +1808,7 @@ nbio_get_sockaddr(term_t Address, struct sockaddr_in *addr)
   { char *hostName;
     term_t arg = PL_new_term_ref();
 
-    PL_get_arg(1, Address, arg);
+    _PL_get_arg(1, Address, arg);
     if ( PL_get_atom_chars(arg, &hostName) )
     { struct hostent *host;
 
@@ -1815,7 +1821,7 @@ nbio_get_sockaddr(term_t Address, struct sockaddr_in *addr)
     { return pl_error(NULL, 0, NULL, ERR_ARGTYPE, 1, arg, "atom|ip/4");
     }
 
-    PL_get_arg(2, Address, arg);
+    _PL_get_arg(2, Address, arg);
     if ( !nbio_get_port(arg, &port) )
       return FALSE;
   } else if ( PL_is_variable(Address) )
@@ -1838,7 +1844,7 @@ nbio_get_ip(term_t ip4, struct in_addr *ip)
     term_t a = PL_new_term_ref();
 
     for(i=1; i<=4; i++)
-    { PL_get_arg(i, ip4, a);
+    { _PL_get_arg(i, ip4, a);
       if ( PL_get_integer(a, &ia) )
 	hip |= ia << ((4-i)*8);
       else
@@ -1852,7 +1858,7 @@ nbio_get_ip(term_t ip4, struct in_addr *ip)
   { term_t a = PL_new_term_ref();
     atom_t id;
 
-    PL_get_arg(1, ip4, a);
+    _PL_get_arg(1, ip4, a);
     if ( PL_get_atom(a, &id) )
     { if ( id == ATOM_any )
 	ip->s_addr = INADDR_ANY;
@@ -2240,17 +2246,11 @@ nbio_close_output(nbio_sock_t socket)
       rc = -1;
     } else if ( shutdown(sock, SD_SEND) == SOCKET_ERROR )
     { const char *msg;
-#ifdef __WINDOWS__
+
       msg = WinSockError(WSAGetLastError());
-#else
-      msg = strerror(errno);
-#endif
       Sseterr(s->output, SIO_FERR, msg);
-#ifdef O_DEBUG
-      if ( debugging )
-	Sdprintf("shutdown(%d=%d, SD_SEND) failed: %s\n",
-		 socket, s->socket, msg);
-#endif
+      DEBUG(1, Sdprintf("shutdown(%d=%d, SD_SEND) failed: %s\n",
+			socket, s->socket, msg));
       rc = -1;
     }
 #endif
@@ -2279,34 +2279,37 @@ nbio_recvfrom(int socket, void *buf, size_t bufSize, int flags,
     return -1;
 
 #ifdef __WINDOWS__
-
   DEBUG(3, Sdprintf("[%d] recvfrom from socket %d\n",
 		    PL_thread_self(), socket));
 
-  n = recvfrom(s->socket, buf, (int)bufSize, flags, from, fromlen);
-  if ( n < 0 )
-  { int wsaerrno;
-
-    if ( (wsaerrno=WSAGetLastError()) == WSAEWOULDBLOCK )
-    { s->rdata.recvfrom.buffer  = buf;
-      s->rdata.recvfrom.size    = bufSize;
-      s->rdata.recvfrom.flags   = flags;
-      s->rdata.recvfrom.from    = from;
-      s->rdata.recvfrom.fromlen = fromlen;
-      placeRequest(s, REQ_RECVFROM);
-      if ( !waitRequest(s) )
-      { errno = EPLEXCEPTION;
-	return -1;
-      }
-      n = s->rdata.recvfrom.bytes;
-    }
+  { int iflen = (int)*fromlen;		/* Windows recvfrom uses int */
 
+    n = recvfrom(s->socket, buf, (int)bufSize, flags, from, &iflen);
     if ( n < 0 )
-    { s->error = wsaerrno;
-      errno = EIO;
+    { int wsaerrno;
+
+      if ( (wsaerrno=WSAGetLastError()) == WSAEWOULDBLOCK )
+      { s->rdata.recvfrom.buffer  = buf;
+	s->rdata.recvfrom.size    = bufSize;
+	s->rdata.recvfrom.flags   = flags;
+	s->rdata.recvfrom.from    = from;
+	s->rdata.recvfrom.fromlen = fromlen;
+	placeRequest(s, REQ_RECVFROM);
+	if ( !waitRequest(s) )
+	{ errno = EPLEXCEPTION;
+	  return -1;
+	}
+	n = s->rdata.recvfrom.bytes;
+      }
+
+      if ( n < 0 )
+      { s->error = wsaerrno;
+	errno = EIO;
+      }
+    } else
+    { *fromlen = iflen;
+      s->error = 0;
     }
-  } else
-  { s->error = 0;
   }
 
 #else /*__WINDOWS__*/
@@ -2355,7 +2358,7 @@ nbio_sendto(nbio_sock_t socket, void *buf, size_t bufSize, int flags,
   DEBUG(3, Sdprintf("[%d] sending %d bytes using socket %d\n",
 		    PL_thread_self(), (int)bufSize, socket));
 
-  n = sendto(s->socket, buf, (int)bufSize, flags, to, tolen);
+  n = sendto(s->socket, buf, (int)bufSize, flags, to, (int)tolen);
   if ( n < 0 )
   { int error = WSAGetLastError();
 
@@ -2373,11 +2376,11 @@ nbio_sendto(nbio_sock_t socket, void *buf, size_t bufSize, int flags,
 
 wouldblock:
   s->rdata.sendto.buffer  = buf;
-  s->rdata.sendto.size    = bufSize;
+  s->rdata.sendto.size    = (int)bufSize;
   s->rdata.sendto.bytes   = 0;
   s->rdata.sendto.flags   = flags;
   s->rdata.sendto.to      = to;
-  s->rdata.sendto.tolen   = tolen;
+  s->rdata.sendto.tolen   = (int)tolen;
   placeRequest(s, REQ_SENDTO);
   if ( !waitRequest(s) )
   { errno = EPLEXCEPTION;		/* handled Prolog signal */
