@@ -106,7 +106,7 @@ dump(Target,NewVars,Constraints) :-
 
 projecting_assert(QClause) :-
 	strip_module(QClause, Module, Clause),  % JW: SWI-Prolog not always qualifies the term!
-	copy_term(Clause,Copy,Constraints),
+	copy_term_clpq(Clause,Copy,Constraints),
 	l2c(Constraints,Conj),			% fails for []
 	(   Sm = clpq
 	;   Sm = clpr
@@ -121,7 +121,7 @@ projecting_assert(QClause) :-
 projecting_assert(Clause) :-	% not our business
 	assert(Clause).
 
-copy_term(Term,Copy,Constraints) :-
+copy_term_clpq(Term,Copy,Constraints) :-
 	(   term_variables(Term,Target),	% get all variables in Term
 	    related_linear_vars(Target,All),	% get all variables of the classes of the variables in Term
 	    nonlin_crux(All,Nonlin),		% get a list of all the nonlinear goals of these variables
@@ -300,4 +300,27 @@ copy(N,T,C,D0,D2) :-
 	arg(N,C,Ac),
 	copy(At,Ac,D0,D1),
 	N1 is N-1,
-	copy(N1,T,C,D1,D2).
\ No newline at end of file
+	copy(N1,T,C,D1,D2).
+
+%%	attribute_goals(@V)// is det.
+%
+%	Translate  attributes  back  into  goals.    This   is  used  by
+%	copy_term/3, which also determines  the   toplevel  printing  of
+%	residual constraints.
+
+itf:attribute_goals(V) -->
+	(   { term_attvars(V, Vs),
+	      dump(Vs, Vs, List),
+	      list_to_conj(List, Conj) }
+	->  [ {}(Conj) ]
+	;   []
+	).
+
+class:attribute_goals(_) --> [].
+
+geler:attribute_goals(V) --> itf:attribute_goals(V).
+
+list_to_conj([], true) :- !.
+list_to_conj([X], X) :- !.
+list_to_conj([H|T0], (H,T)) :-
+	list_to_conj(T0, T).
