@@ -1,4 +1,4 @@
-/*  $Id: bv_r.pl,v 1.4 2006/05/22 14:43:56 leslie Exp $
+/*  $Id$
 
     Part of CLP(R) (Constraint Logic Programming over Reals)
 
@@ -978,7 +978,7 @@ sd([l(X*K,_)|Xs],Class0,ClassN,Preference0,PreferenceN,NV0,NVt) :-
 	    )
   	;   % new: has no class
 	    Class1 = Class0,
-	    'C'(NV0,X,NV1), % X has no class yet, add to list of new variables
+	    NV0 = [X|NV1], % X has no class yet, add to list of new variables
 	    (   arg(2,Att,type(t_none))
 	    ->  preference(Preference0,1-X-K,Preference1)
 		    % no class, no bounds => category 1
@@ -1783,4 +1783,4 @@ dump_nz(_,H,I) -->
 	    mult_hom(H,Kr,H1),
 	    nf2sum(H1,0.0,Sum)
 	},
-	[Sum =\= I1].
\ No newline at end of file
+	[Sum =\= I1].
