@@ -35,13 +35,16 @@
 	    http_redirect/3,		% +How, +Path, +Request
 	    http_current_handler/2,	% ?Path, ?Pred
 	    http_current_handler/3,	% ?Path, ?Pred
-	    http_location_by_id/2	% +ID, -Location
+	    http_location_by_id/2,	% +ID, -Location
+	    http_link_to_id/3,		% +ID, +Parameters, -HREF
+	    http_safe_file/2		% +Spec, +Options
 	  ]).
 :- use_module(library(option)).
 :- use_module(library(lists)).
 :- use_module(library(time)).
 :- use_module(library(error)).
 :- use_module(library(settings)).
+:- use_module(library(uri)).
 :- use_module(library(http/mimetype)).
 :- use_module(library(http/http_path)).
 :- use_module(library(http/http_header)).
@@ -89,10 +92,26 @@ write_index(Request) :-
 %		Use =|Transfer-encoding: chunked|= if the client
 %		allows for it.
 %
+%		* content_type(+Term)
+%		Specifies the content-type of the reply.  This value is
+%		currently not used by this library.  It enhances the
+%		reflexive capabilities of this library through
+%		http_current_handler/3.
+%
 %		* id(+Term)
 %		Identifier of the handler.  The default identifier is
 %		the predicate name.  Used by http_location_by_id/2.
 %
+%		* hide_children(+Bool)
+%		If =true= on a prefix-handler (see prefix), possible
+%		children are masked.  This can be used to (temporary)
+%		overrule part of the tree.
+%
+%		* prefix
+%		Call Pred on any location that is a specialisation of
+%		Path.  If multiple handlers match, the one with the
+%		longest path is used.
+%
 %		* priority(+Integer)
 %		If two handlers handle the same path, the one with the
 %		highest priority is used.  If equal, the last registered
@@ -100,11 +119,6 @@ write_index(Request) :-
 %		multifile predicates can change due to reloading files.
 %		The default priority is 0 (zero).
 %
-%		* prefix
-%		Call Pred on any location that is a specialisation of
-%		Path.  If multiple handlers match, the one with the
-%		longest path is used.
-%
 %		* spawn(+SpawnOptions)
 %		Run the handler in a seperate thread.  If SpawnOptions
 %		is an atom, it is interpreted as a thread pool name
@@ -117,12 +131,6 @@ write_index(Request) :-
 %		One of =infinite=, =default= or a positive number
 %		(seconds)
 %
-%		* content_type(+Term)
-%		Specifies the content-type of the reply.  This value is
-%		currently not used by this library.  It enhances the
-%		reflexive capabilities of this library through
-%		http_current_handler/3.
-%
 %	Note that http_handler/3 is normally invoked  as a directive and
 %	processed using term-expansion.  Using   term-expansion  ensures
 %	proper update through make/0 when the specification is modified.
@@ -158,14 +166,30 @@ system:term_expansion((:- http_handler(Path, Pred, Options)), Clause) :-
 	next_generation.
 
 
-%%	http_delete_handler(+Path) is det.
+%%	http_delete_handler(+Spec) is det.
 %
-%	Delete handler for Path. Typically, this should only be used for
-%	handlers that are registered dynamically.
+%	Delete handler for Spec. Typically, this should only be used for
+%	handlers that are registered dynamically. Spec is one of:
+%
+%	    * id(Id)
+%	    Delete a handler with the given id.  The default id is the
+%	    handler-predicate-name.
+%
+%	    * path(Path)
+%	    Delete handler that serves the given path.
 
-http_delete_handler(Path) :-
+http_delete_handler(id(Id)) :- !,
+	clause(handler(_Path, _:Pred, _, Options), true, Ref),
+	functor(Pred, _, DefID),
+	option(id(Id0), Options, DefID),
+	Id == Id0,
+	erase(Ref),
+	next_generation.
+http_delete_handler(path(Path)) :- !,
 	retractall(handler(Path, _Pred, _, _Options)),
 	next_generation.
+http_delete_handler(Path) :-
+	http_delete_handler(path(Path)).
 
 
 %%	next_generation is det.
@@ -379,6 +403,20 @@ location_by_id_raw(ID, Location, Priority) :-
 	;   ID = PN
 	), !.
 
+
+%%	http_link_to_id(+HandleID, +Parameters, -HREF)
+%
+%	HREF is a link on the local server to a handler with given ID,
+%	passing the given Parameters.
+
+http_link_to_id(HandleID, Parameters, HREF) :-
+	http_location_by_id(HandleID, Location),
+	uri_data(path, Components, Location),
+	uri_query_components(String, Parameters),
+	uri_data(search, Components, String),
+	uri_components(HREF, Components).
+
+
 %	hook into html_write:attribute_value//1.
 
 :- multifile
@@ -397,7 +435,7 @@ html_write:expand_attribute_value(location_by_id(ID)) -->
 %	The  library  http_authenticate.pl  provides  an  implementation
 %	thereof.
 %
-%	@error	permission_error(http_location, access, Location)
+%	@error	permission_error(access, http_location, Location)
 
 :- multifile
 	http:authenticate/3.
@@ -408,7 +446,7 @@ authentication([authentication(Type)|Options], Request, Fields) :- !,
 	->  append(XFields, More, Fields),
 	    authentication(Options, Request, More)
 	;   memberchk(path(Path), Request),
-	    throw(error(permission_error(http_location, access, Path), _))
+	    throw(error(permission_error(access, http_location, Path), _))
 	).
 authentication([_|Options], Request, Fields) :-
 	authentication(Options, Request, Fields).
@@ -445,7 +483,8 @@ find_handler(Path, Action, Options) :-
 find_handler([node(prefix(Prefix), PAction, POptions, Children)|_],
 	     Path, Action, Options) :-
 	sub_atom(Path, 0, _, After, Prefix), !,
-	(   find_handler(Children, Path, Action, Options)
+	(   option(hide_children(false), POptions, false),
+	    find_handler(Children, Path, Action, Options)
 	->  true
 	;   Action = PAction,
 	    path_info(After, Path, POptions, Options)
@@ -554,7 +593,7 @@ extend(G0, Extra, G) :-
 %	@throws http_reply(file(MimeType, Path))
 
 http_reply_file(File, Options, Request) :-
-	check_file_safeness(File, Options),
+	http_safe_file(File, Options),
 	absolute_file_name(File, Path,
 			   [ access(read)
 			   ]),
@@ -579,7 +618,7 @@ http_reply_file(File, Options, Request) :-
 	),
 	throw(http_reply(Reply)).
 
-%%	check_file_safeness(+FileSpec, +Options) is det.
+%%	http_safe_file(+FileSpec, +Options) is det.
 %
 %	True if FileSpec is considered _safe_.  If   it  is  an atom, it
 %	cannot  be  absolute  and  cannot   have  references  to  parent
@@ -589,20 +628,20 @@ http_reply_file(File, Options, Request) :-
 %	@error instantiation_error
 %	@error permission_error(read, file, FileSpec)
 
-check_file_safeness(File, _) :-
+http_safe_file(File, _) :-
 	var(File), !,
 	instantiation_error(File).
-check_file_safeness(_, Options) :-
+http_safe_file(_, Options) :-
 	option(unsafe(true), Options, false), !.
-check_file_safeness(File, _) :-
-	check_file_safeness(File).
+http_safe_file(File, _) :-
+	http_safe_file(File).
 
-check_file_safeness(File) :-
+http_safe_file(File) :-
 	compound(File),
 	functor(File, _, 1), !,
 	arg(1, File, Name),
 	safe_name(Name, File).
-check_file_safeness(Name) :-
+http_safe_file(Name) :-
 	(   is_absolute_file_name(Name)
 	->  permission_error(read, file, Name)
 	;   true
@@ -805,8 +844,9 @@ prolog:meta_goal(http_current_handler(_, G), [G+1]).
 
 prolog_edit:locate(Path, Spec, Location) :-
 	atom(Path),
+	sub_atom(Path, 0, _, _, /),
 	Pred = _M:_H,
-	http_current_handler(Path, Pred),
+	catch(http_current_handler(Path, Pred), _, fail),
 	closure_name_arity(Pred, 1, PI),
 	prolog_edit:locate(PI, Spec, Location).
 
