@@ -87,8 +87,8 @@ http_read_request(In, Request) :-
 	    Request =  [input(In)|Request1],
 	    phrase(request(In, Request1), Codes),
 	    (	Request1 = [unknown(Text)|_]
-	    ->	string_to_list(S, Text),
-		domain_error(http_request, S)
+	    ->	atom_codes(S, Text),
+		syntax_error(http_request(S))
 	    ;	true
 	    )
 	).
@@ -107,7 +107,8 @@ http_read_reply_header(In, [input(In)|Reply]) :-
 	;   debug(http(header), 'First line: ~s~n', [Codes]),
 	    (   phrase(reply(In, Reply), Codes)
 	    ->  true
-	    ;   throw(error(syntax(http_reply_header, Codes), _))
+	    ;   atom_codes(Header, Codes),
+		syntax_error(http_reply_header(Header))
 	    )
 	).
 
@@ -240,6 +241,23 @@ http_status_reply(Status, Out, HdrExtra, Code) :-
 			   set_stream(Out, encoding(octet))), !.
 
 
+status_reply(no_content, Out, HrdExtra, Code) :- !,
+	phrase(reply_header(status(no_content), HrdExtra, Code), Header),
+	format(Out, '~s', [Header]),
+	flush_output(Out).
+status_reply(created(Location), Out, HrdExtra, Code) :- !,
+	phrase(page([ title('201 Created')
+		    ],
+		    [ h1('Created'),
+		      p(['The document was created ',
+			 a(href(Location), ' Here')
+			]),
+		      \address
+		    ]),
+	       HTML),
+	phrase(reply_header(created(Location, HTML), HrdExtra, Code), Header),
+	format(Out, '~s', [Header]),
+	print_html(Out, HTML).
 status_reply(moved(To), Out, HrdExtra, Code) :- !,
 	phrase(page([ title('301 Moved Permanently')
 		    ],
@@ -354,6 +372,17 @@ status_reply(server_error(ErrorTerm), Out, HrdExtra, Code) :-
 			    HrdExtra, Code), Header),
 	format(Out, '~s', [Header]),
 	print_html(Out, HTML).
+status_reply(not_acceptable(WhyHTML), Out, HdrExtra, Code) :- !,
+	phrase(page([ title('406 Not Acceptable')
+		    ],
+		    [ h1('Not Acceptable'),
+		      WhyHTML,
+		      \address
+		    ]),
+	       HTML),
+	phrase(reply_header(status(not_acceptable, HTML), HdrExtra, Code), Header),
+	format(Out, '~s', [Header]),
+	print_html(Out, HTML).
 status_reply(unavailable(WhyHTML), Out, HdrExtra, Code) :- !,
 	phrase(page([ title('503 Service Unavailable')
 		    ],
@@ -785,6 +814,14 @@ reply_header(moved(To, Tokens), HdrExtra, Code) -->
 	content_length(html(Tokens), CLen),
 	content_type(text/html, utf8),
 	"\r\n".
+reply_header(created(Location, Tokens), HdrExtra, Code) -->
+	vstatus(moved, Code),
+	date(now),
+	header_field('Location', Location),
+	header_fields(HdrExtra, CLen),
+	content_length(html(Tokens), CLen),
+	content_type(text/html, utf8),
+	"\r\n".
 reply_header(moved_temporary(To, Tokens), HdrExtra, Code) -->
 	vstatus(moved_temporary, Code),
 	date(now),
@@ -844,15 +881,19 @@ status_number(Status, Code) -->
 
 status_number(continue,		   100).
 status_number(ok,		   200).
+status_number(created,		   201).
+status_number(accepted,		   202).
+status_number(no_content,	   204).
 status_number(partial_content,	   206).
 status_number(moved,		   301).
 status_number(moved_temporary,	   302).
 status_number(see_other,	   303).
 status_number(not_modified,	   304).
 status_number(bad_request,	   400).
-status_number(not_found,	   404).
-status_number(forbidden,	   403).
 status_number(authorise,	   401).
+status_number(forbidden,	   403).
+status_number(not_found,	   404).
+status_number(not_acceptable,      406).
 status_number(server_error,	   500).
 status_number(service_unavailable, 503).
 
@@ -865,6 +906,14 @@ status_comment(continue) -->
 	"Continue".
 status_comment(ok) -->
 	"OK".
+status_comment(created) -->
+	"Created".
+status_comment(accepted) -->
+	"Accepted".
+status_comment(no_content) -->
+	"No Content".
+status_comment(created) -->
+	"Created".
 status_comment(partial_content) -->
 	"Partial content".
 status_comment(moved) -->
@@ -887,6 +936,8 @@ status_comment(server_error) -->
 	"Internal Server Error".
 status_comment(service_unavailable) -->
 	"Service Unavailable".
+status_comment(not_acceptable) -->
+	"Not Acceptable".
 
 authenticate(Method, '') --> !,
 	"WWW-Authenticate: ",
@@ -1189,11 +1240,12 @@ http_timestamp(Time, Atom) :-
 		 *	   REQUEST DCG		*
 		 *******************************/
 
-request(Fd, [method(Method)|Header]) -->
+request(Fd, [method(Method),request_uri(ReqURI)|Header]) -->
 	method(Method),
 	blanks,
 	nonblanks(Query),
-	{ http_location(Parts, Query),
+	{ atom_codes(ReqURI, Query),
+	  http_location(Parts, Query),
 	  append(Parts, Header0, Header)
 	},
 	request_header(Fd, Header0), !.
@@ -1402,6 +1454,13 @@ read_header_data(_, Fd, Tail) :-
 	read_line_to_codes(Fd, Tail, NewTail),
 	read_header_data(Tail, Fd, NewTail).
 
+%%	http_parse_header(+Text:codes, -Header:list) is det.
+%
+%	Header is a list of Name(Value)-terms representing the structure
+%	of the HTTP header in Text.
+%
+%	@error domain_error(http_request_line, Line)
+
 http_parse_header(Text, Header) :-
 	phrase(header(Header), Text),
 	debug(http(header), 'Fields: ~w~n', [Header]).
@@ -1414,7 +1473,12 @@ header(List) -->
 	header(Tail).
 header([]) -->
 	blanks,
-	[].
+	eos, !.
+header(_) -->
+	string(S), blanks_to_nl, !,
+	{ atom_codes(Line, S),
+	  syntax_error(http_request_line(Line))
+	}.
 
 %%	address//
 %
@@ -1435,7 +1499,9 @@ address -->
 	http:http_address, !.
 address -->
 	{ gethostname(Host) },
-	html(address(['SWI-Prolog httpd at ', Host])).
+	html(address([ a(href('http://www.swi-prolog.org'), 'SWI-Prolog'),
+		       ' httpd at ', Host
+		     ])).
 
 mkfield(host, Host:Port, [host(Host),port(Port)|Tail], Tail) :- !.
 mkfield(Name, Value, [Att|Tail], Tail) :-
