@@ -53,6 +53,19 @@ additional argument with options. Currently only one option is defined:
     priorities is needed because we want to be able to overrule
     paths, but we do not want to become dependent on clause ordering.
 
+    The default priority is 0. Note however that notably libraries may
+    decide to provide a fall-back using a negative priority.  We suggest
+    -100 for such cases.
+
+This library predefines three locations at   priority  -100: The =icons=
+and =css= aliases are intended for images   and css files and are backed
+up by file a file-search-path that  allows   finding  the  icons and css
+files that belong to the server infrastructure (e.g., http_dirindex/2).
+
+    * root
+    The root of the server.  Default is /, but this may be overruled
+    the the setting (see setting/2) =|http:prefix|=
+
 Here is an example that binds =|/login|=  to login/1. The user can reuse
 this application while moving all locations  using   a  new rule for the
 admin location with the option =|[priority(10)]|=.
@@ -78,13 +91,14 @@ login(Request) :-
 :- dynamic
 	http:location/3.		% Alias, Expansion, Options
 
-http:location(root, Root, []) :-
+http:location(root, Root, [priority(-100)]) :-
 	(   catch(setting(http:prefix, Prefix), _, fail),
 	    Prefix \== ''
 	->  Root = Prefix
 	;   Root = (/)
 	).
 
+
 %%	http_absolute_location(+Spec, -Path, +Options) is det.
 %
 %	Path is the HTTP location for the abstract specification Spec.
@@ -144,13 +158,13 @@ http_location_path(Alias, Path) :-
 	(   Result = [_-One]
 	->  Path = One
 	;   Result == []
-	->  existence_error(http_location, Spec)
+	->  existence_error(http_alias, Alias)
 	;   Result = [P-Best,P2-_|_],
 	    P \== P2
 	->  Path = Best
 	;   Result = [_-First|_],
 	    pairs_values(Result, Paths),
-	    print_message(warning, http(ambiguous_location(Spec, Paths))),
+	    print_message(warning, http(ambiguous_location(Alias, Paths))),
 	    Path = First
 	).
 
