@@ -3,9 +3,9 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2010, University of Amsterdam, VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -344,10 +344,12 @@ odbc_report(HENV henv, HDBC hdbc, HSTMT hstmt, RETCODE rc)
     case SQL_SUCCESS:
       if ( msglen > SQL_MAX_MESSAGE_LENGTH )
 	msglen = SQL_MAX_MESSAGE_LENGTH; /* TBD: get the rest? */
-      PL_unify_term(msg, PL_FUNCTOR, FUNCTOR_odbc3,
-			   PL_CHARS,   state,
-		           PL_INTEGER, (long)native,
-		           PL_NCHARS,  (size_t)msglen, message);
+      if ( !PL_unify_term(msg,
+			  PL_FUNCTOR, FUNCTOR_odbc3,
+			    PL_CHARS,   state,
+			    PL_INTEGER, (long)native,
+			    PL_NCHARS,  (size_t)msglen, message) )
+	return FALSE;
       break;
     case SQL_INVALID_HANDLE:
       return PL_warning("ODBC INTERNAL ERROR: Invalid handle in error");
@@ -371,13 +373,16 @@ odbc_report(HENV henv, HDBC hdbc, HSTMT hstmt, RETCODE rc)
       return TRUE;
     }
     case SQL_ERROR:
-    { term_t ex = PL_new_term_ref();
+    { term_t ex;
 
-      PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		    PL_TERM, msg,
-		    PL_VARIABLE);
+      if ( (ex=PL_new_term_ref()) &&
+	   PL_unify_term(ex,
+			 PL_FUNCTOR, FUNCTOR_error2,
+			   PL_TERM, msg,
+			 PL_VARIABLE) )
+	return PL_raise_exception(ex);
 
-      return PL_raise_exception(ex);
+      return FALSE;
     }
     default:
       return PL_warning("Statement returned %d\n", rc);
@@ -415,96 +420,117 @@ report_status(context *ctxt)
 
 static int
 type_error(term_t actual, const char *expected)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_type_error2,
-		        PL_CHARS, expected,
-		        PL_TERM, actual,
-		      PL_VARIABLE);
+  if ( (ex=PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR, FUNCTOR_type_error2,
+		         PL_CHARS, expected,
+		         PL_TERM, actual,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 static int
 domain_error(term_t actual, const char *expected)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_domain_error2,
-		        PL_CHARS, expected,
-		        PL_TERM, actual,
-		      PL_VARIABLE);
+  if ( (ex=PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR, FUNCTOR_domain_error2,
+		         PL_CHARS, expected,
+		         PL_TERM, actual,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 static int
 existence_error(term_t actual, const char *expected)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_existence_error2,
-		        PL_CHARS, expected,
-		        PL_TERM, actual,
-		      PL_VARIABLE);
+  if ( (ex=PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR, FUNCTOR_existence_error2,
+		         PL_CHARS, expected,
+		         PL_TERM, actual,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 static int
 resource_error(const char *error)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_resource_error1,
-		        PL_CHARS, error,
-		      PL_VARIABLE);
+  if ( (ex=PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR, FUNCTOR_resource_error1,
+		         PL_CHARS, error,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
 static int
 representation_error(term_t t, const char *error)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_representation_error1,
-		        PL_CHARS, error,
-		      PL_TERM, t);
+  if ( (ex=PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR, FUNCTOR_representation_error1,
+		         PL_CHARS, error,
+		       PL_TERM, t) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
 static int
 context_error(term_t term, const char *error, const char *what)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_context_error3,
-			PL_TERM, term,
-		        PL_CHARS, error,
-			PL_CHARS, what,
-		      PL_VARIABLE);
+  if ( (ex=PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR, FUNCTOR_context_error3,
+			 PL_TERM, term,
+		         PL_CHARS, error,
+			 PL_CHARS, what,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
 static int
 permission_error(const char *op, const char *type, term_t obj)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_permission_error3,
-		        PL_CHARS, op,
-			PL_CHARS, type,
-			PL_TERM, obj,
-		      PL_VARIABLE);
+  if ( (ex=PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR, FUNCTOR_permission_error3,
+		         PL_CHARS, op,
+			 PL_CHARS, type,
+			 PL_TERM, obj,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
@@ -541,7 +567,7 @@ odbc_malloc(size_t bytes)
 
 static int
 get_text(term_t t, char **s)
-{ return PL_get_chars(t, s, CVT_ATOM|CVT_STRING|CVT_LIST|BUF_RING);
+{ return PL_get_chars(t, s, CVT_ATOM|CVT_STRING|CVT_LIST|REP_MB|BUF_RING);
 }
 
 static int
@@ -571,19 +597,13 @@ PL_get_typed_arg(int i, term_t t, int (*func)(), void *ap)
 
 static int
 list_length(term_t list)
-{ term_t tail = PL_copy_term_ref(list);
-  term_t head = PL_new_term_ref();
-  int n = 0;
+{ size_t len;
 
-  while(PL_get_list(tail, head, tail))
-    n++;
+  if ( PL_skip_list(list, 0, &len) == PL_LIST )
+    return (int)len;
 
-  if ( !PL_get_nil(tail) )
-  { type_error(tail, "list");
-    return -1;
-  }
-
-  return n;
+  type_error(list, "list");
+  return -1;
 }
 
 
@@ -601,11 +621,11 @@ formatted_string(term_t in, size_t *len, char **out)
 
   if ( !format )
     format = PL_predicate("format", 3, "user");
-  PL_unify_stream(av+0, fd);
-  PL_get_arg(1, in, av+1);
-  PL_get_arg(2, in, av+2);
-					/* TBD: call natively */
-  if ( !PL_call_predicate(NULL, PL_Q_PASS_EXCEPTION, format, av) )
+
+  if ( !PL_unify_stream(av+0, fd) ||
+       !PL_get_arg(1, in, av+1) ||
+       !PL_get_arg(2, in, av+2) ||
+       !PL_call_predicate(NULL, PL_Q_PASS_EXCEPTION, format, av) )
   { Sclose(fd);
     if ( *out )
       PL_free(*out);
@@ -652,7 +672,7 @@ nulldef_spec(term_t t)
 	      PL_functor_arity(f) == 1 )
   { term_t a1 = PL_new_term_ref();
 
-    PL_get_arg(1, t, a1);
+    _PL_get_arg(1, t, a1);
     if ( PL_is_variable(a1) )
     { nd->nulltype = NULL_FUNCTOR;
       nd->nullvalue.functor = f;
@@ -698,23 +718,24 @@ free_nulldef(nulldef *nd)
 }
 
 
-static void
+WUNUSED static int
 put_sql_null(term_t t, nulldef *nd)
 { if ( nd )
   { switch(nd->nulltype)
     { case NULL_VAR:
-	break;
+	return TRUE;
       case NULL_ATOM:
-	PL_put_atom(t, nd->nullvalue.atom);
-        break;
+	return PL_put_atom(t, nd->nullvalue.atom);
       case NULL_FUNCTOR:
-	PL_put_functor(t, nd->nullvalue.functor);
-        break;
+	return PL_put_functor(t, nd->nullvalue.functor);
       case NULL_RECORD:
-	PL_recorded(nd->nullvalue.record, t);
+	return PL_recorded(nd->nullvalue.record, t);
+      default:
+	assert(0);
+        return FALSE;
     }
   } else
-    PL_put_atom(t, ATOM_null);
+    return PL_put_atom(t, ATOM_null);
 }
 
 
@@ -794,7 +815,7 @@ nth_row_arg(compile_info *info, term_t var)
 { int i;
 
   for(i=1; i<=info->columns; i++)
-  { PL_get_arg(i, info->row, info->tmp);
+  { _PL_get_arg(i, info->row, info->tmp);
     if ( PL_compare(info->tmp, var) == 0 )
       return i;
   }
@@ -825,7 +846,9 @@ compile_arg(compile_info *info, term_t t)
     }
     case PL_ATOM:
     { atom_t val;
-      PL_get_atom(t, &val);
+
+      if ( !PL_get_atom(t, &val) )
+	assert(0);
       ADDCODE_1(info, PL_ATOM, val);
       if ( true(info, CTX_PERSISTENT) )
 	PL_register_atom(val);
@@ -838,7 +861,8 @@ compile_arg(compile_info *info, term_t t)
 	{ u_double v;
 	  unsigned int i;
 
-	  PL_get_float(t, &v.asdouble);
+	  if ( !PL_get_float(t, &v.asdouble) )
+	    assert(0);
 	  ADDCODE(info, PL_FLOAT);
 	  for(i=0; i<sizeof(double)/sizeof(code); i++)
 	    ADDCODE(info, v.ascode[i]);
@@ -861,7 +885,9 @@ compile_arg(compile_info *info, term_t t)
       break;
     case PL_INTEGER:
     { long v;
-      PL_get_long(t, &v);
+
+      if ( !PL_get_long(t, &v) )
+	assert(0);
       ADDCODE_1(info, PL_INTEGER, v);
       break;
     }
@@ -870,11 +896,12 @@ compile_arg(compile_info *info, term_t t)
       int i, arity;
       term_t a = PL_new_term_ref();
 
-      PL_get_functor(t, &f);
+      if ( !PL_get_functor(t, &f) )
+	assert(0);
       arity = PL_functor_arity(f);
       ADDCODE_1(info, PL_FUNCTOR, f);
       for(i=1; i<=arity; i++)
-      { PL_get_arg(i, t, a);
+      { _PL_get_arg(i, t, a);
 	compile_arg(info, a);
       }
       break;
@@ -900,18 +927,21 @@ compile_findall(term_t all, unsigned flags)
   info.size  = 0;
   info.flags = flags;
 
-  PL_get_arg(2, all, info.row);
-  PL_get_name_arity(info.row, &a, &info.columns);
+  if ( !PL_get_arg(2, all, info.row) ||
+       !PL_get_name_arity(info.row, &a, &info.columns) )
+    return NULL;
 
   for(i=1; i<=info.columns; i++)
-  { PL_get_arg(i, info.row, t);
+  { if ( !PL_get_arg(i, info.row, t) )
+      return NULL;
     if ( !PL_is_variable(t) )
     { type_error(t, "unbound");
       return NULL;
     }
   }
 
-  PL_get_arg(1, all, t);
+  if ( !PL_get_arg(1, all, t) )
+    return NULL;
   compile_arg(&info, t);
 
   if ( !(f = odbc_malloc(FND_SIZE(info.size))) )
@@ -1001,17 +1031,20 @@ build_term(context *ctxt, code *PC, term_t result)
 
       for(i=0; i<sizeof(double)/sizeof(code); i++)
 	v.ascode[i] = *PC++;
-      PL_put_float(result, v.asdouble);
+      if ( !PL_put_float(result, v.asdouble) )
+	return NULL;
       return PC;
     }
     case PL_STRING:
     { int len = (int)*PC++;
       char *s = (char*)*PC++;
-      PL_put_string_nchars(result, len, s);
+      if ( !PL_put_string_nchars(result, len, s) )
+	return NULL;
       return PC;
     }
     case PL_INTEGER:
-    { PL_put_integer(result, (long)*PC++);
+    { if ( !PL_put_integer(result, (long)*PC++) )
+	return NULL;
       return PC;
     }
     case PL_TERM:
@@ -1028,7 +1061,8 @@ build_term(context *ctxt, code *PC, term_t result)
 	  return NULL;
       }
 
-      PL_cons_functor_v(result, f, av);
+      if ( !PL_cons_functor_v(result, f, av) )
+	return NULL;
       PL_reset_term_refs(av);
       return PC;
     }
@@ -1140,7 +1174,7 @@ get_connection(term_t tdsn, connection **cn)
   { term_t a = PL_new_term_ref();
     void *ptr;
 
-    PL_get_arg(1, tdsn, a);
+    _PL_get_arg(1, tdsn, a);
     if ( !PL_get_pointer(a, &ptr) )
       return type_error(tdsn, "odbc_connection");
     c = ptr;
@@ -1458,7 +1492,7 @@ odbc_set_connection(connection *cn, term_t option)
   } else if ( PL_is_functor(option, FUNCTOR_null1) )
   { term_t a = PL_new_term_ref();
 
-    PL_get_arg(1, option, a);
+    _PL_get_arg(1, option, a);
     cn->null = nulldef_spec(a);
 
     return TRUE;
@@ -1543,7 +1577,7 @@ odbc_get_connection(term_t conn, term_t option, control_t h)
       opt = PL_foreign_context_address(h);
 
       goto find;
-    case PL_CUTTED:
+    case PL_PRUNED:
     default:
       return TRUE;
   }
@@ -1551,7 +1585,7 @@ odbc_get_connection(term_t conn, term_t option, control_t h)
 find:
   val = PL_new_term_ref();
   a = PL_new_term_ref();
-  PL_get_arg(1, option, a);
+  _PL_get_arg(1, option, a);
 
   for(; opt->name; opt++)
   { if ( !opt->functor )
@@ -1575,8 +1609,11 @@ find:
 	  PL_put_atom_nchars(val, len, buf);
 	  break;
 	case sword:
-	{ SQLSMALLINT v = *((SQLSMALLINT*)buf);
-	  PL_put_integer(val, v);
+	{ SQLSMALLINT *p = (SQLSMALLINT*)buf;
+	  SQLSMALLINT v = *p;
+
+	  if ( !PL_put_integer(val, v) )
+	    return FALSE;
 	  break;
 	}
 	default:
@@ -1588,9 +1625,10 @@ find:
       if ( f )
 	return PL_unify(a, val);
 
-      PL_unify_term(option,
-		    PL_FUNCTOR, opt->functor,
-		    PL_TERM, val);
+      if ( !PL_unify_term(option,
+			  PL_FUNCTOR, opt->functor,
+			  PL_TERM, val) )
+	return FALSE;
 
       if ( opt[1].name )
 	PL_retry_address(opt+1);
@@ -1865,7 +1903,7 @@ get_sql_text(context *ctxt, term_t tquery)
     ctxt->sqltext = q;
     ctxt->sqllen = (SQLINTEGER)qlen;
     set(ctxt, CTX_SQLMALLOCED);
-  } else if ( PL_get_nchars(tquery, &qlen, &q, CVT_ATOM|CVT_STRING|BUF_MALLOC))
+  } else if ( PL_get_nchars(tquery, &qlen, &q, CVT_ATOM|CVT_STRING|REP_MB|BUF_MALLOC))
   { ctxt->sqltext = q;
     ctxt->sqllen = (SQLINTEGER)qlen;
     set(ctxt, CTX_SQLMALLOCED);
@@ -2171,8 +2209,8 @@ set_column_types(context *ctxt, term_t option)
   parameter *p;
   int ntypes;
 
-  PL_get_arg(1, option, tail);
-  if ( (ntypes = list_length(tail)) < 0 )
+  if ( !PL_get_arg(1, option, tail) ||
+       (ntypes = list_length(tail)) < 0 )
     return FALSE;			/* not a proper list */
 
   ctxt->NumCols = ntypes;
@@ -2205,7 +2243,7 @@ set_statement_options(context *ctxt, term_t options)
       } else if ( PL_is_functor(head, FUNCTOR_null1) )
       { term_t arg = PL_new_term_ref();
 
-	PL_get_arg(1, head, arg);
+	_PL_get_arg(1, head, arg);
 	ctxt->null = nulldef_spec(arg);
 	set(ctxt, CTX_OWNNULL);
       } else if ( PL_is_functor(head, FUNCTOR_source1) )
@@ -2230,7 +2268,7 @@ set_statement_options(context *ctxt, term_t options)
 	  set(ctxt, CTX_NOAUTO);
 	else
 	{ term_t a = PL_new_term_ref();
-	  PL_get_arg(1, head, a);
+	  _PL_get_arg(1, head, a);
 	  return domain_error(a, "fetch");
 	}
       } else if ( PL_is_functor(head, FUNCTOR_wide_column_threshold1) )
@@ -2291,7 +2329,7 @@ pl_odbc_query(term_t dsn, term_t tquery, term_t trow, term_t options,
       return odbc_row(PL_foreign_context_address(handle), trow);
 
     default:
-    case PL_CUTTED:
+    case PL_PRUNED:
       free_context(PL_foreign_context_address(handle));
       return TRUE;
   }
@@ -2325,7 +2363,7 @@ odbc_tables(term_t dsn, term_t row, control_t handle)
     case PL_REDO:
       return odbc_row(PL_foreign_context_address(handle), row);
 
-    case PL_CUTTED:
+    case PL_PRUNED:
       free_context(PL_foreign_context_address(handle));
       return TRUE;
 
@@ -2345,7 +2383,7 @@ pl_odbc_column(term_t dsn, term_t db, term_t row, control_t handle)
       size_t len;
       char *s;
 
-      if ( !PL_get_nchars(db, &len, &s, CVT_ATOM|CVT_STRING) )
+      if ( !PL_get_nchars(db, &len, &s, CVT_ATOM|CVT_STRING|REP_MB) )
 	return type_error(db, "atom");
 
       if ( !get_connection(dsn, &cn) )
@@ -2364,7 +2402,7 @@ pl_odbc_column(term_t dsn, term_t db, term_t row, control_t handle)
     case PL_REDO:
       return odbc_row(PL_foreign_context_address(handle), row);
 
-    case PL_CUTTED:
+    case PL_PRUNED:
       free_context(PL_foreign_context_address(handle));
       return TRUE;
 
@@ -2410,7 +2448,7 @@ odbc_types(term_t dsn, term_t sqltype, term_t row, control_t handle)
     case PL_REDO:
       return odbc_row(PL_foreign_context_address(handle), row);
 
-    case PL_CUTTED:
+    case PL_PRUNED:
       free_context(PL_foreign_context_address(handle));
       return TRUE;
 
@@ -2476,7 +2514,7 @@ getStmt(term_t id, context **ctxt)
   { term_t a = PL_new_term_ref();
     void *ptr;
 
-    PL_get_arg(1, id, a);
+    _PL_get_arg(1, id, a);
     if ( PL_get_pointer(a, &ptr) )
     { *ctxt = ptr;
 
@@ -2624,11 +2662,11 @@ declare_parameters(context *ctxt, term_t parms)
     if ( PL_is_functor(head, FUNCTOR_gt2) )
     { term_t a = PL_new_term_ref();
 
-      PL_get_arg(1, head, a);
+      _PL_get_arg(1, head, a);
       if ( !get_pltype(a, &plType) )
 	return FALSE;
 
-      PL_get_arg(2, head, head);
+      _PL_get_arg(2, head, head);
     }
 
     if ( !PL_get_name_arity(head, &name, &arity) )
@@ -2903,8 +2941,8 @@ try_null(context *ctxt, parameter *prm, term_t val, const char *expected)
 
 
 static int
-get_parameter_text(term_t t, parameter *prm, size_t *len, char **s)
-{ unsigned int flags = CVT_ATOM|CVT_STRING;
+get_parameter_text(term_t t, parameter *prm, int rep, size_t *len, char **s)
+{ unsigned int flags;
   const char *expected = "text";
 
   switch(prm->plTypeID)
@@ -2928,7 +2966,7 @@ get_parameter_text(term_t t, parameter *prm, size_t *len, char **s)
       assert(0);
   }
 
-  if ( !PL_get_nchars(t, len, s, flags) )
+  if ( !PL_get_nchars(t, len, s, flags|rep) )
     return type_error(t, expected);
 
   return TRUE;
@@ -2984,6 +3022,7 @@ bind_parameters(context *ctxt, term_t parms)
       { SQLLEN len;
 	size_t l;
 	char *s;
+	int rep = prm->cTypeID == (SQL_C_CHAR ? REP_MB : REP_ISO_LATIN_1);
 
 					/* check for NULL */
 	if ( is_sql_null(head, ctxt->null) )
@@ -2991,7 +3030,7 @@ bind_parameters(context *ctxt, term_t parms)
 	  break;
 	}
 
-	if ( !get_parameter_text(head, prm, &l, &s) )
+	if ( !get_parameter_text(head, prm, rep, &l, &s) )
 	  return FALSE;
 	len = l;
 
@@ -3070,7 +3109,9 @@ odbc_execute(term_t qid, term_t args, term_t row, control_t handle)
 	  { s = NULL;
 	    len = SQL_NULL_DATA;
 	  } else
-	  { if ( !get_parameter_text(p->put_data, p, &len, &s) )
+	  { int rep = (p->cTypeID == SQL_C_BINARY ? REP_ISO_LATIN_1 : REP_MB);
+
+	    if ( !get_parameter_text(p->put_data, p, rep, &len, &s) )
 	      return FALSE;
 	  }
 	  SQLPutData(ctxt->hstmt, s, len);
@@ -3090,7 +3131,7 @@ odbc_execute(term_t qid, term_t args, term_t row, control_t handle)
     case PL_REDO:
       return odbc_row(PL_foreign_context_address(handle), row);
 
-    case PL_CUTTED:
+    case PL_PRUNED:
       close_context(PL_foreign_context_address(handle));
       return TRUE;
 
@@ -3495,22 +3536,27 @@ Microsoft SQL Server sometimes gives the wrong length indication as well
 as the wrong number of pad bytes for the first part of the data.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static void
-put_chars(term_t val, int plTypeID, size_t len, const char *chars)
-{ switch( plTypeID )
+WUNUSED static int
+put_chars(term_t val, int plTypeID, int rep, size_t len, const char *chars)
+{ int pltype;
+
+  switch( plTypeID )
   { case SQL_PL_DEFAULT:
     case SQL_PL_ATOM:
-      PL_put_atom_nchars(val, len, chars);
+      pltype = PL_ATOM;
       break;
     case SQL_PL_STRING:
-      PL_put_string_nchars(val, len, chars);
+      pltype = PL_STRING;
       break;
     case SQL_PL_CODES:
-      PL_put_list_ncodes(val, len, chars);
+      pltype = PL_CODE_LIST;
       break;
     default:
       assert(0);
+      return FALSE;
   }
+
+  return PL_unify_chars(val, pltype|rep, len, chars);
 }
 
 
@@ -3545,7 +3591,8 @@ pl_put_column(context *c, int nth, term_t col)
     if ( c->rc == SQL_SUCCESS || c->rc == SQL_SUCCESS_WITH_INFO )
     { DEBUG(2, Sdprintf("Got %ld bytes\n", len));
       if ( len == SQL_NULL_DATA )
-      { put_sql_null(val, c->null);
+      { if ( !put_sql_null(val, c->null) )
+	  return FALSE;
 	goto ok;
       } else if ( len == SQL_NO_TOTAL )
       { /* Don't know what to do here */
@@ -3597,49 +3644,60 @@ pl_put_column(context *c, int nth, term_t col)
     }
 
   got_all_data:
-    put_chars(val, p->plTypeID, len, data);
+    { int rep = (p->cTypeID == SQL_C_BINARY ? REP_ISO_LATIN_1 : REP_MB);
+
+      if ( !put_chars(val, p->plTypeID, rep, len, data) )
+	return FALSE;
+    }
     if ( data != buf )
       free(data);
     goto ok;
   }
 
   if ( p->length_ind == SQL_NULL_DATA )
-  { put_sql_null(val, c->null);
+  { if ( !put_sql_null(val, c->null) )
+      return FALSE;
   } else
-  { switch( p->cTypeID )
+  { int rc;
+
+    switch( p->cTypeID )
     { case SQL_C_CHAR:
+	rc = put_chars(val, p->plTypeID, REP_MB, p->length_ind, (char*)p->ptr_value);
+        break;
       case SQL_C_BINARY:
-	put_chars(val, p->plTypeID, p->length_ind, (char*)p->ptr_value);
+	rc = put_chars(val, p->plTypeID, REP_ISO_LATIN_1, p->length_ind, (char*)p->ptr_value);
 	break;
       case SQL_C_SLONG:
-	PL_put_integer(val,*(SQLINTEGER *)p->ptr_value);
+	rc = PL_put_integer(val,*(SQLINTEGER *)p->ptr_value);
 	break;
       case SQL_C_SBIGINT:
-	PL_put_int64(val, *(SQLBIGINT *)p->ptr_value);
+	rc = PL_put_int64(val, *(SQLBIGINT *)p->ptr_value);
 	break;
       case SQL_C_DOUBLE:
-	PL_put_float(val,*(SQLDOUBLE *)p->ptr_value);
+	rc = PL_put_float(val,*(SQLDOUBLE *)p->ptr_value);
 	break;
       case SQL_C_TYPE_DATE:
       { DATE_STRUCT* ds = (DATE_STRUCT*)p->ptr_value;
-	term_t av = PL_new_term_refs(3);
+	term_t av;
 
-	PL_put_integer(av+0, ds->year);
-	PL_put_integer(av+1, ds->month);
-	PL_put_integer(av+2, ds->day);
-
-	PL_cons_functor_v(val, FUNCTOR_date3, av);
+	rc = ( (av=PL_new_term_refs(3)) &&
+	       PL_put_integer(av+0, ds->year) &&
+	       PL_put_integer(av+1, ds->month) &&
+	       PL_put_integer(av+2, ds->day) &&
+	       PL_cons_functor_v(val, FUNCTOR_date3, av)
+	     );
 	break;
       }
       case SQL_C_TYPE_TIME:
       { TIME_STRUCT* ts = (TIME_STRUCT*)p->ptr_value;
-	term_t av = PL_new_term_refs(3);
+	term_t av;
 
-	PL_put_integer(av+0, ts->hour);
-	PL_put_integer(av+1, ts->minute);
-	PL_put_integer(av+2, ts->second);
-
-	PL_cons_functor_v(val, FUNCTOR_time3, av);
+	rc = ( (av=PL_new_term_refs(3)) &&
+	       PL_put_integer(av+0, ts->hour) &&
+	       PL_put_integer(av+1, ts->minute) &&
+	       PL_put_integer(av+2, ts->second) &&
+	       PL_cons_functor_v(val, FUNCTOR_time3, av)
+	     );
 	break;
       }
       case SQL_C_TIMESTAMP:
@@ -3648,17 +3706,18 @@ pl_put_column(context *c, int nth, term_t col)
 	switch( p->plTypeID )
 	{ case SQL_PL_DEFAULT:
 	  case SQL_PL_TIMESTAMP:
-	  { term_t av = PL_new_term_refs(7);
+	  { term_t av;
 
-	    PL_put_integer(av+0, ts->year);
-	    PL_put_integer(av+1, ts->month);
-	    PL_put_integer(av+2, ts->day);
-	    PL_put_integer(av+3, ts->hour);
-	    PL_put_integer(av+4, ts->minute);
-	    PL_put_integer(av+5, ts->second);
-	    PL_put_integer(av+6, ts->fraction);
-
-	    PL_cons_functor_v(val, FUNCTOR_timestamp7, av);
+	    rc = ( (av=PL_new_term_refs(7)) &&
+		   PL_put_integer(av+0, ts->year) &&
+		   PL_put_integer(av+1, ts->month) &&
+		   PL_put_integer(av+2, ts->day) &&
+		   PL_put_integer(av+3, ts->hour) &&
+		   PL_put_integer(av+4, ts->minute) &&
+		   PL_put_integer(av+5, ts->second) &&
+		   PL_put_integer(av+6, ts->fraction) &&
+		   PL_cons_functor_v(val, FUNCTOR_timestamp7, av)
+		 );
 	    break;
 	  }
 	  case SQL_PL_INTEGER:
@@ -3687,9 +3746,9 @@ pl_put_column(context *c, int nth, term_t col)
 #endif
 
 	    if ( p->plTypeID == SQL_PL_INTEGER )
-	      PL_put_integer(val, t);
+	      rc = PL_put_integer(val, t);
 	    else
-	      PL_put_float(val, (double)t); /* TBD: fraction */
+	      rc = PL_put_float(val, (double)t); /* TBD: fraction */
 	  }
 #else
 	    return PL_warning("System doesn't support mktime()/timegm()");
@@ -3704,11 +3763,13 @@ pl_put_column(context *c, int nth, term_t col)
 	return PL_warning("ODBC: Unknown cTypeID: %d",
 			  p->cTypeID);
     }
+    if ( !rc )
+      return FALSE;
   }
 
 ok:
   if ( true(c, CTX_SOURCE) )
-    PL_cons_functor_v(col, FUNCTOR_column3, cell);
+    return PL_cons_functor_v(col, FUNCTOR_column3, cell);
 
   return TRUE;
 }
@@ -3730,9 +3791,7 @@ pl_put_row(term_t row, context *c)
       return FALSE;			/* with exception */
   }
 
-  PL_cons_functor_v(row, c->db_row, columns);
-
-  return TRUE;
+  return PL_cons_functor_v(row, c->db_row, columns);
 }
 
 
