@@ -126,14 +126,14 @@ doc_server(Port, _) :-
 doc_server(Port, Options) :-
 	prepare_editor,
 	host_access_options(Options, ServerOptions),
-	append(ServerOptions,		% Put provides options first,
-	       [ port(Port),		% so they override our defaults
-		 timeout(60),
-		 keep_alive_timeout(1),
-		 local(4000),		% keep stack sizes independent
-		 global(4000),		% from main application
-		 trail(4000)
-	       ], HTTPOptions),
+	merge_options(ServerOptions,
+		      [ port(Port),
+			timeout(60),
+			keep_alive_timeout(1),
+			local(4000),	% small stacks for now
+			global(4000),
+			trail(4000)
+		      ], HTTPOptions),
 	http_server(http_dispatch, HTTPOptions),
 	assert(doc_server_port(Port)),
 	print_message(informational, pldoc(server_started(Port))).
@@ -266,7 +266,8 @@ pldoc_root(Request, _) :-
 %	of the available documentation.
 
 pldoc_index(_Request) :-
-	reply_html_page(title('SWI-Prolog documentation'),
+	reply_html_page(pldoc(index),
+			title('SWI-Prolog documentation'),
 			[ \doc_links('', []),
 			   h1('SWI-Prolog documentation'),
 			  \man_overview([])
@@ -325,7 +326,8 @@ pldoc_edit(Request) :-
 	),
 	format(string(Cmd), '~q', [edit(Edit)]),
 	edit(Edit),
-	reply_html_page(title('Edit'),
+	reply_html_page(pldoc(edit),
+			title('Edit'),
 			p(['Started ', Cmd])).
 pldoc_edit(_Request) :-
 	throw(http_reply(forbidden('/edit'))).
@@ -526,7 +528,8 @@ pldoc_man(Request) :-
 	;   ground(F)
 	->  Obj = c(F)
 	),
-	reply_html_page(title(Title),
+	reply_html_page(pldoc(man),
+			title(Title),
 			\man_page(Obj, [])).
 
 %%	pldoc_object(+Request)
@@ -544,7 +547,8 @@ pldoc_object(Request) :-
 	;   Title = Atom
 	),
 	edit_options(Request, EditOptions),
-	reply_html_page(title(Title),
+	reply_html_page(pldoc(object),
+			title(Title),
 			\object_page(Obj, EditOptions)).
 
 
@@ -576,7 +580,8 @@ pldoc_search(Request) :-
 			]),
 	edit_options(Request, EditOptions),
 	format(string(Title), 'Prolog search -- ~w', [For]),
-	reply_html_page(title(Title),
+	reply_html_page(pldoc(search),
+			title(Title),
 			\search_reply(For,
 				      [ resultFormat(Format),
 					search_in(In),
