@@ -88,6 +88,7 @@ blocked('/chr').
 blocked('/clpq').
 blocked('/clpr').
 blocked('/pldoc').
+blocked('/ciao').
 blocked('/checkselect.pl').
 blocked('/checklast.pl').
 %blocked('/jpl.pl').			% should be added
@@ -97,3 +98,4 @@ blocked('/index.pl').
 blocked('/ciao.pl').			% is an include-file.  We must
 					% find a more general solution here
 blocked('/commons.pl').
+blocked('/swipl-lfr.pl').
