@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2010, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -37,6 +38,7 @@
 	    rdf_has/3,			% ?Subject, +Pred, ?Obj
 	    rdf_has/4,			% ?Subject, +Pred, ?Obj, -RealPred
 	    rdf_reachable/3,		% ?Subject, +Pred, ?Object
+	    rdf_reachable/5,		% ?Subject, +Pred, ?Object, +MaxD, ?D
 	    rdf_subject/1,		% ?Subject
 
 	    rdf_member_property/2,	% ?Property, ?Index
@@ -66,6 +68,9 @@
 	    rdf_bnode/1,		% -Id
 	    rdf_is_bnode/1,		% +Id
 
+	    rdf_is_resource/1,		% +Term
+	    rdf_is_literal/1,		% +Term
+
 	    rdf_load/1,			% +File
 	    rdf_load/2,			% +File, +Options
 	    rdf_save/1,			% +File
@@ -90,6 +95,8 @@
 	    rdf_save_footer/1,		% +Out
 
 	    rdf_equal/2,		% ?Resource, ?Resource
+	    lang_equal/2,		% +Lang1, +Lang2
+	    lang_matches/2,		% +Lang, +Pattern
 
 	    rdf_current_ns/2,		% ?Alias, ?URI
 	    rdf_register_ns/2,		% +Alias, +URI
@@ -139,7 +146,7 @@
 
 :- multifile
 	ns/2,
-	rdf_meta_specification/2.	% UnboundHead, Head
+	rdf_meta_specification/3.	% UnboundHead, Module, Head
 :- dynamic
 	ns/2,			% ID, URL
 	rdf_source/5.		% DB, SourceURL, ModTimeAtLoad, Triples, MD5
@@ -170,13 +177,15 @@ rdf_current_ns(Alias, URI) :-
 %	Dynamic  predicate  that  maintains   the  registered  namespace
 %	aliases.
 
-ns(rdf,  'http://www.w3.org/1999/02/22-rdf-syntax-ns#').
-ns(rdfs, 'http://www.w3.org/2000/01/rdf-schema#').
-ns(owl,  'http://www.w3.org/2002/07/owl#').
-ns(xsd,  'http://www.w3.org/2001/XMLSchema#').
-ns(dc,   'http://purl.org/dc/elements/1.1/').
-ns(eor,  'http://dublincore.org/2000/03/13/eor#').
-ns(serql,'http://www.openrdf.org/schema/serql#').
+ns(rdf,	    'http://www.w3.org/1999/02/22-rdf-syntax-ns#').
+ns(rdfs,    'http://www.w3.org/2000/01/rdf-schema#').
+ns(owl,	    'http://www.w3.org/2002/07/owl#').
+ns(xsd,	    'http://www.w3.org/2001/XMLSchema#').
+ns(dc,	    'http://purl.org/dc/elements/1.1/').
+ns(dcterms, 'http://purl.org/dc/terms/').
+ns(eor,	    'http://dublincore.org/2000/03/13/eor#').
+ns(skos,    'http://www.w3.org/2004/02/skos/core#').
+ns(serql,   'http://www.openrdf.org/schema/serql#').
 
 %%	rdf_register_ns(+Alias, +URI) is det.
 %%	rdf_register_ns(+Alias, +URI, +Options) is det.
@@ -312,15 +321,17 @@ rdf_global_term(Term, Term).
 	system:goal_expansion/2.
 
 system:term_expansion((:- rdf_meta(Heads)), Clauses) :-
-	mk_clauses(Heads, Clauses).
+	prolog_load_context(module, M),
+	mk_clauses(Heads, M, Clauses).
 
-mk_clauses((A,B), [H|T]) :- !,
-	mk_clause(A, H),
-	mk_clauses(B, T).
-mk_clauses(A, [C]) :-
-	mk_clause(A, C).
+mk_clauses((A,B), M, [H|T]) :- !,
+	mk_clause(A, M, H),
+	mk_clauses(B, M, T).
+mk_clauses(A, M, [C]) :-
+	mk_clause(A, M, C).
 
-mk_clause(Head, rdf_db:rdf_meta_specification(Unbound, Head)) :-
+mk_clause(Head0, M0, rdf_db:rdf_meta_specification(Unbound, Module, Head)) :-
+	strip_module(M0:Head0, Module, Head),
 	valid_rdf_meta_head(Head),
 	functor(Head, Name, Arity),
 	functor(Unbound, Name, Arity).
@@ -359,9 +370,33 @@ rdf_meta(Heads) :-
 
 
 system:goal_expansion(G, Expanded) :-
-	rdf_meta_specification(G, Spec), !,
+	rdf_meta_specification(G, _, _), !,
+	prolog_load_context(module, LM),
+	(   rdf_meta_specification(G, Module, Spec),
+	    right_module(LM, G, Module)
+	->  rdf_expand(G, Spec, Expanded)
+	;   debugging(rdf_meta),
+	    sub_term(G, NS:Local),
+	    atom(NS), atom(Local)
+	->  print_message(warning, rdf_meta(not_expanded(LM:G))),
+	    fail
+	),
 	rdf_expand(G, Spec, Expanded).
 
+system:term_expansion(Fact, Expanded) :-
+	rdf_meta_specification(Fact, Module, Spec),
+	prolog_load_context(module, Module), !,
+	rdf_expand(Fact, Spec, Expanded).
+system:term_expansion((Head :- Body), (Expanded :- Body)) :-
+	rdf_meta_specification(Head, Module, Spec),
+	prolog_load_context(module, Module), !,
+	rdf_expand(Head, Spec, Expanded).
+
+
+right_module(M, _, M) :- !.
+right_module(LM, G, M) :-
+	predicate_property(LM:G, imported_from(M)).
+
 rdf_expand(G, Spec, Expanded) :-
 	functor(G, Name, Arity),
 	functor(Expanded, Name, Arity),
@@ -396,18 +431,13 @@ mk_global(X, X) :-
 mk_global(X, X) :-
 	atom(X), !.
 mk_global(NS:Local, Global) :-
-	must_be_atom(NS),
-	must_be_atom(Local),
+	must_be(atom, NS),
+	must_be(atom, Local),
 	(   ns(NS, Full)
 	->  atom_concat(Full, Local, Global)
-	;   throw(error(existence_error(namespace, NS), _))
+	;   existence_error(namespace, NS)
 	).
 
-must_be_atom(X) :-
-	atom(X), !.
-must_be_atom(X) :-
-	throw(error(type_error(atom, X), _)).
-
 :- rdf_meta
 	rdf(r,r,o),
 	rdf_has(r,r,o,r),
@@ -418,6 +448,7 @@ must_be_atom(X) :-
 	rdf_assert(r,r,o,+),
 	rdf_retractall(r,r,o,?),
 	rdf_reachable(r,r,r),
+	rdf_reachable(r,r,r,+,?),
 	rdf_update(r,r,o,t),
 	rdf_update(r,r,o,+,t),
 	rdf_equal(r,r),
@@ -433,6 +464,17 @@ must_be_atom(X) :-
 
 rdf_equal(Resource, Resource).
 
+%%	lang_equal(+Lang1, +Lang2) is semidet.
+%
+%	True if two RFC language specifiers denote the same language
+%
+%	@see lang_matches/2.
+
+lang_equal(Lang, Lang) :- !.
+lang_equal(Lang1, Lang2) :-
+	downcase_atom(Lang1, LangCannon),
+	downcase_atom(Lang2, LangCannon).
+
 
 %%	rdf_has(?Subject, +Predicate, ?Object)
 %
@@ -492,6 +534,11 @@ rdf_bnode(Value) :-
 	\+ rdf(_, Value, _), !.
 
 
+
+		 /*******************************
+		 *	       TYPES		*
+		 *******************************/
+
 %%	rdf_is_bnode(+Id)
 %
 %	Tests if a resource is a blank node (i.e. is an anonymous
@@ -503,6 +550,25 @@ rdf_is_bnode(Id) :-
 	atom(Id),
 	sub_atom(Id, 0, _, _, '__').
 
+%%	rdf_is_resource(@Term) is semidet.
+%
+%	True if Term is an RDF  resource.   Note  that  this is merely a
+%	type-test; it does not mean  this   resource  is involved in any
+%	triple.  Blank nodes are also considered resources.
+%
+%	@see rdf_is_bnode/1
+
+rdf_is_resource(Term) :-
+	atom(Term).
+
+%%	rdf_is_literal(@Term) is semidet.
+%
+%	True if Term is an RDF literal object. Currently only checks for
+%	groundness and the literal functor.
+
+rdf_is_literal(literal(Value)) :-
+	ground(Value).
+
 
 		 /*******************************
 		 *	      SOURCE		*
@@ -543,7 +609,7 @@ rdf_statistics(rehash(Count, Time)) :-
 rdf_statistics(core(Bytes)) :-
 	rdf_statistics_(core(Bytes)).
 rdf_statistics(lookup(Index, Count)) :-
-	functor(Indexed, indexed, 8),
+	functor(Indexed, indexed, 16),
 	rdf_statistics_(Indexed),
 	index(Index, I),
 	Arg is I + 1,
@@ -555,21 +621,30 @@ rdf_statistics(literals(Count)) :-
 	rdf_statistics_(literals(Count)).
 rdf_statistics(triples_by_file(File, Count)) :-
 	(   var(File)
-	->  rdf_source(File),
+	->  rdf_graph(File),
 	    rdf_statistics_(triples(File, Count))
 	;   rdf_statistics_(triples(File, Count))
 	).
 rdf_statistics(duplicates(Count)) :-
 	rdf_statistics_(duplicates(Count)).
 
-index(rdf(-,-,-), 0).
-index(rdf(+,-,-), 1).
-index(rdf(-,+,-), 2).
-index(rdf(+,+,-), 3).
-index(rdf(-,-,+), 4).
-index(rdf(+,-,+), 5).
-index(rdf(-,+,+), 6).
-index(rdf(+,+,+), 7).
+index(rdf(-,-,-,-), 0).
+index(rdf(+,-,-,-), 1).
+index(rdf(-,+,-,-), 2).
+index(rdf(+,+,-,-), 3).
+index(rdf(-,-,+,-), 4).
+index(rdf(+,-,+,-), 5).
+index(rdf(-,+,+,-), 6).
+index(rdf(+,+,+,-), 7).
+
+index(rdf(-,-,-,+), 8).
+index(rdf(+,-,-,+), 9).
+index(rdf(-,+,-,+), 10).
+index(rdf(+,+,-,+), 11).
+index(rdf(-,-,+,+), 12).
+index(rdf(+,-,+,+), 13).
+index(rdf(-,+,+,+), 14).
+index(rdf(+,+,+,+), 15).
 
 
 		 /*******************************
@@ -767,13 +842,12 @@ rdf_load_db(File) :-
 		 *******************************/
 
 :- multifile
-	rdf_open_hook/3,
-	rdf_load_stream/3,
-	rdf_input_info_hook/3,
-	rdf_file_type/2,
-	rdf_storage_encoding/2,
-	url_protocol/1,
-	exists_url/1.
+	rdf_open_hook/8,
+	rdf_open_decode/4,		% +Encoding, +File, -Stream, -Cleanup
+	rdf_load_stream/3,		% +Format, +Stream, +Options
+	rdf_file_type/2,		% ?Extension, ?Format
+	rdf_storage_encoding/2,		% ?Extension, ?Encoding
+	url_protocol/1.			% ?Protocol
 
 %%	rdf_load(+FileOrList) is det.
 %%	rdf_load(+FileOrList, +Options) is det.
@@ -781,11 +855,6 @@ rdf_load_db(File) :-
 %	Load RDF file.  Options provides additional processing options.
 %	Currently defined options are:
 %
-%	    * result(-Action, -Triples, -MD5)
-%	    Return action taken (load, reload, none) and number
-%	    of triples loaded from the file as well as the MD5
-%	    digest.
-%
 %	    * blank_nodes(+ShareMode)
 %	    How to handle equivalent blank nodes.  If =share= (default),
 %	    equivalent blank nodes are shared in the same resource.
@@ -795,14 +864,23 @@ rdf_load_db(File) :-
 %	    that are relative to the base uri.  Default is the source
 %	    URL.
 %
-%	    * db(+DB)
+%	    * graph(?Graph)
 %	    Named graph in which to load the data.  It is *not* allowed
-%	    to load two sources into the same named graph.
+%	    to load two sources into the same named graph.  If Graph is
+%	    unbound, it is unified to the graph into which the data is
+%	    loaded.
+%
+%	    * db(?Graph)
+%	    Deprecated.  New code must use graph(Graph).
 %
 %	    * if(Condition)
 %	    When to load the file. One of =true=, =changed= (default) or
 %	    =not_loaded=.
 %
+%	    * modified(-Modified)
+%	    Unify Modified with one of =not_modified=, cached(File),
+%	    last_modified(Stamp) or =unknown=.
+%
 %	    * cache(Bool)
 %	    If =false=, do not use or create a cache file.
 %
@@ -819,162 +897,277 @@ rdf_load([], _) :- !.
 rdf_load([H|T], Options) :- !,
 	rdf_load(H, Options),
 	rdf_load(T, Options).
-rdf_load(Spec, M:Options0) :-
-	fix_options(Options0, Options),
+rdf_load(Spec, M:Options) :-
+	must_be(list, Options),
 	statistics(cputime, T0),
-	(   select(result(Action, Triples, MD5), Options, Options1)
-	->  true
-	;   Options1 = Options
+	rdf_open_input(Spec, In, Cleanup, SourceURL, Graph, Modified,
+		       Format, Options),
+	return_modified(Modified, Options),
+	(   Modified == not_modified
+	->  Action = none
+	;   Modified = cached(CacheFile)
+	->  do_unload(Graph),
+	    catch(rdf_load_db_no_admin(CacheFile, cache(Graph), Graphs), _, fail),
+	    check_loaded_cache(Graph, Graphs, Modified),
+	    rdf_statistics_(triples(Graph, Triples)),
+	    Action = load
+	;   option(base_uri(BaseURI), Options, Graph),
+	    (	var(BaseURI)
+	    ->	BaseURI = SourceURL
+	    ;	true
+	    ),
+	    once(phrase(derived_options(Options, NSList), Extra)),
+	    merge_options([ base_uri(BaseURI),
+			    graph(Graph),
+			    format(Format)
+			  | Extra
+			  ], Options, RDFOptions),
+	    do_unload(Graph),
+	    graph_modified(Modified, ModifiedStamp),
+	    rdf_set_graph_source(Graph, SourceURL, ModifiedStamp),
+	    call_cleanup(rdf_load_stream(Format, In, M:RDFOptions),
+			 Cleanup),
+	    save_cache(Graph, SourceURL, Options),
+	    register_file_ns(NSList),
+	    rdf_statistics_(triples(Graph, Triples)),
+	    rdf_md5(Graph, MD5),
+	    assert(rdf_source(Graph, SourceURL, Modified, Triples, MD5)),
+	    format_action(Format, Action)
 	),
-	rdf_input(Spec, Input, SourceURL),
-	(   rdf_input_info(Input, Modified, DefFormat)
-	->  true
-	;   Modified = 0,
-	    DefFormat = xml
+	report_loaded(Action, SourceURL, Graph, Triples, T0, Options).
+
+format_action(triples, load) :- !.
+format_action(_, parsed).
+
+save_cache(Graph, SourceURL, Options) :-
+	option(cache(true), Options, true),
+	rdf_cache_file(SourceURL, write, CacheFile), !,
+	catch(save_cache(Graph, CacheFile), E,
+	      print_message(warning, E)).
+save_cache(_, _, _).
+
+derived_options([], _) -->
+	[].
+derived_options([H|T], NSList) -->
+	(   {   H == register_namespaces(true)
+	    ;   H == (register_namespaces = true)
+	    }
+	->  [ namespaces(NSList) ]
+	;   []
 	),
-	select_option(base_uri(BaseURI), Options1, Options2, SourceURL),
-	select_option(format(Format), Options2, Options3, DefFormat),
-	select_option(blank_nodes(ShareMode), Options3, Options4, noshare),
-	select_option(cache(Cache), Options4, Options5, true),
-	select_option(if(If), Options5, Options6, changed),
-	select_option(db(DB), Options6, Options7, SourceURL),
-	select_option(register_namespaces(RegNS), Options7, RDFOptions0, false),
-	(   var(BaseURI)
-	->  BaseURI = SourceURL
+	derived_options(T, NSList).
+
+graph_modified(last_modified(Stamp), Stamp).
+graph_modified(unknown, Stamp) :-
+	get_time(Stamp).
+
+return_modified(Modified, Options) :-
+	option(modified(M0), Options), !,
+	M0 = Modified.
+return_modified(_, _).
+
+
+		 /*******************************
+		 *	  INPUT HANDLING	*
+		 *******************************/
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+This section deals with pluggable input sources.  The task of the input
+layer is
+
+    * Decide on the graph-name
+    * Decide on the source-location
+    * Decide whether loading is needed (if-modified)
+    * Decide on the serialization in the input
+
+The protocol must ensure minimal  overhead,   in  particular for network
+protocols. E.g. for HTTP we want to make a single call on the server and
+use If-modified-since to verify that we need not reloading this file.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+%%	rdf_open_input(+Spec, -Stream, -Cleanup,
+%%		       -Source, -Graph, -Modified, -Format,
+%%		       +Options)
+%
+%	Open an input source.
+%
+%	Options processed:
+%
+%	    * graph(Graph)
+%	    * db(Graph)
+%	    * if(Condition)
+%	    * cache(Cache)
+%	    * format(Format)
+%
+%	@param	Modified is one of =not_modified=, last_modified(Time),
+%		cached(CacheFile) or =unknown=
+
+rdf_open_input(Spec, Stream, Cleanup,
+	       SourceURL, Graph, Modified, Format,
+	       Options) :-
+	source_url(Spec, Protocol, SourceURL),
+	load_graph(SourceURL, Graph, Options),
+	option(if(If), Options, changed),
+	(   If == true
+	->  true
+	;   rdf_graph_source_(Graph, SourceURL, HaveModified)
+	->  true
+	;   option(cache(true), Options, true),
+	    rdf_cache_file(SourceURL, read, CacheFile)
+	->  time_file(CacheFile, HaveModified)
 	;   true
 	),
-	(   RegNS == true
-	->  RDFOptions = [ namespaces(NSList)|RDFOptions0]
-	;   RDFOptions = RDFOptions0
-	),
-	(   must_load(If, DB, Modified)
-	->  do_unload(DB),		% unload old
-	    (   Cache == true,
-		read_cache(SourceURL, Modified, CacheFile),
-	        catch(rdf_load_db_no_admin(CacheFile, cache(DB), Graphs), _, fail),
-		check_loaded_cache(DB, Graphs, Modified)
-	    ->  Action = load
-	    ;   rdf_input_open(Input, Stream, Format),
-		must_be(ground, Format),
-		rdf_format(Format, RDFFormat),
-		rdf_set_graph_source(DB, SourceURL, Modified),
-		call_cleanup(rdf_load_stream(RDFFormat, Stream,
-					     M:[ base_uri(BaseURI),
-						 blank_nodes(ShareMode),
-						 db(DB)
-					       | RDFOptions
-					       ]),
-			     close_input(Input, Stream)), !,
-		register_file_ns(NSList),
-		(   Cache == true,
-		    rdf_cache_file(SourceURL, write, CacheFile)
-		->  catch(save_cache(DB, CacheFile), E,
-			  print_message(warning, E))
-		;   true
-		),
-		(   Format = triples
-		->  Action = load
-		;   Action = parsed
-		)
-	    ),
-	    rdf_statistics_(triples(DB, Triples)),
-	    rdf_md5(DB, MD5),
-	    assert(rdf_source(DB, SourceURL, Modified, Triples, MD5))
-	;   Action = none,
-	    rdf_source(DB, _, _, Triples, MD5)
-	),
-	report_loaded(Action, Input, DB, Triples, T0, Options).
-
-%%	fix_options(+Spec, -Options) is det.
-%
-%	Fix options for old format
-
-fix_options(DB, Options) :-
-	atom(DB), DB \== [], !,
-	Options = [db(DB)].
-fix_options(Options, Options).
-
-%%	rdf_format(+FormatIn, -Format)
-%
-%	Remove format envelopes, such as gzip(Format).
-
-rdf_format(Format, Format) :-
-	atom(Format), !.
-rdf_format(Format0, Format) :-
-	arg(1, Format0, Format1),
-	rdf_format(Format1, Format).
-
-%%	close_input(+Input, +Stream) is det.
-%
-%	Close input if it was not specified as a stream.
-
-close_input(stream(_), _) :- !.
-close_input(_, Stream) :-
-	close(Stream).
-
-%%	rdf_input(+Term, -Input, -BaseURI) is semidet.
-%
-%	Resolve input description Term.  Unify   Input  with a canonical
-%	description of the input, which is one of:
-%
-%	  * stream(Stream)
-%	  * file(AbsolutePath)
-%	  * url(Protocol, URL)
-%
-%	BaseURI is unified with a default base   URI. Note that this may
-%	be overruled from the options of rdf_load/2.
-
-rdf_input(stream(Stream), stream(Stream), BaseURI) :- !,
-	(   stream_property(Stream, file_name(File))
-	->  uri_file_name(BaseURI, File)
-	;   gensym('stream://', BaseURI)
+	option(format(Format), Options, _),
+	open_input_if_modified(Protocol, SourceURL, HaveModified,
+			       Stream, Cleanup, Modified0, Format, Options),
+	(   Modified0 == not_modified
+	->  (   nonvar(CacheFile)
+	    ->	Modified = cached(CacheFile)
+	    ;	Modified = not_modified
+	    )
+	;   Modified = Modified0
 	).
-rdf_input(Stream, stream(Stream), BaseURI) :-
+
+
+%%	source_url(+Spec, -Class, -SourceURL) is det.
+%
+%	Determine class and url of the source.  Class is one of
+%
+%	    * stream(Stream)
+%	    * file
+%	    * a url-protocol (e.g., =http=)
+
+source_url(stream(In), stream(In), SourceURL) :- !,
+	(   stream_property(In, file_name(File))
+	->  to_url(File, SourceURL)
+	;   gensym('stream://', SourceURL)
+	).
+source_url(Stream, Class, SourceURL) :-
 	is_stream(Stream), !,
-	rdf_input(stream(Stream), _, BaseURI).
-rdf_input(FileURL, file(File), BaseURI) :-
-	atom(FileURL),
-	uri_file_name(FileURL, File0), !,
-	file_input(File0, File, BaseURI).
-rdf_input(URL0, url(Protocol, URL), URL0) :-
-	is_url(URL0, Protocol), !,
-	(   storage_extend(URL0, URL),
-	    exists_url(url(Protocol, URL))
-	->  true
-	;   URL=URL0
-	).
-rdf_input(Spec, file(Path), BaseURI) :-
-	file_input(Spec, Path, BaseURI).
+	source_url(stream(Stream), Class, SourceURL).
+source_url(Spec, Protocol, SourceURL) :-
+	compound(Spec), !,
+	source_file(Spec, Protocol, SourceURL).
+source_url(FileURL, Protocol, SourceURL) :-		% or return FileURL?
+	uri_file_name(FileURL, File), !,
+	source_file(File, Protocol, SourceURL).
+source_url(SourceURL, Protocol, SourceURL) :-
+	is_url(SourceURL, Protocol), !.
+source_url(File, Protocol, SourceURL) :-
+	source_file(File, Protocol, SourceURL).
 
-file_input(Spec, Path, BaseURI) :-
+source_file(Spec, file(SExt), SourceURL) :-
 	findall(Ext, valid_extension(Ext), Exts),
-	absolute_file_name(Spec, Path,
-			   [ access(read),
-			     extensions(Exts)
-			   ]),
-	uri_file_name(BaseURI0, Path),
-	clean_base_uri(BaseURI0, BaseURI).
+	absolute_file_name(Spec, File, [access(read), extensions(Exts)]),
+	storage_extension(Plain, SExt, File),
+	uri_file_name(SourceURL, Plain).
+
+to_url(URL, URL) :-
+	is_url(URL, _Protocol), !.
+to_url(File, URL) :-
+	absolute_file_name(File, Path),
+	uri_file_name(URL, Path).
+
+storage_extension(Plain, SExt, File) :-
+	file_name_extension(Plain, SExt, File),
+	SExt \== '',
+	rdf_storage_encoding(SExt, _), !.
+storage_extension(File, '', File).
+
+%%	rdf_input(URL, Source, _BaseURI) is semidet.
+%
+%	@deprecated Only exists to make old ClioPatria rdf_library.pl
+%	work
+
+rdf_input(Spec, Source, _BaseURI) :-
+	source_url(Spec, Class, SourceURL),
+	compat_input(Class, SourceURL, Source).
+
+compat_input(file(Ext), SourceURL, file(Path)) :-
+	uri_file_name(SourceURL, File),
+	file_name_extension(File, Ext, Path).
+compat_input(http, SourceURL, url(http, SourceURL)).
+
+%%	load_graph(+SourceURL, -Graph, +Options) is det.
+%
+%	Graph is the graph into which we load the data.  Processes
+%	the graph(?Graph) option.
+
+load_graph(Source, Graph, Options) :-
+	(   option(graph(Graph), Options)
+	->  true
+	;   option(db(Graph), Options)
+	), !,
+	(   ground(Graph)
+	->  true
+	;   load_graph(Source, Graph)
+	).
+load_graph(Source, Graph, _) :-
+	load_graph(Source, Graph).
+
+load_graph(SourceURL, BaseURI) :-
+	file_name_extension(BaseURI, Ext, SourceURL),
+	rdf_storage_encoding(Ext, _), !.
+load_graph(SourceURL, SourceURL).
+
+
+open_input_if_modified(stream(In), SourceURL, _, In, true,
+		       unknown, Format, _) :- !,
+	(   var(Format)
+	->  guess_format(SourceURL, Format)
+	;   true
+	).
+open_input_if_modified(file(SExt), SourceURL, HaveModified, Stream, Cleanup,
+		       Modified, Format, _) :- !,
+	uri_file_name(SourceURL, File0),
+	file_name_extension(File0, SExt, File),
+	time_file(File, LastModified),
+	(   nonvar(HaveModified),
+	    HaveModified >= LastModified
+	->  Modified = not_modified,
+	    Cleanup = true
+	;   storage_open(SExt, File, Stream, Cleanup),
+	    Modified = last_modified(LastModified),
+	    (	var(Format)
+	    ->	guess_format(File0, Format)
+	    ;	true
+	    )
+	).
+open_input_if_modified(file, SourceURL, HaveModified, Stream, Cleanup,
+		       Modified, Format, Options) :- !,
+	open_input_if_modified(file(''), SourceURL, HaveModified,
+			       Stream, Cleanup,
+			       Modified, Format, Options).
+open_input_if_modified(Protocol, SourceURL, HaveModified, Stream, Cleanup,
+		       Modified, Format, Options) :-
+	rdf_open_hook(Protocol, SourceURL, HaveModified, Stream, Cleanup,
+		      Modified, Format, Options).
+
+guess_format(File, Format) :-
+	file_name_extension(_, Ext, File),
+	(   rdf_file_type(Ext, Format)
+	->  true
+	;   Format = xml,
+	    print_message(warning, rdf(guess_format(Ext)))
+	).
+
+%%	storage_open(+Extension, +File, -Stream, -Cleanup)
+%
+%	Open the low-level storage. Note  that   the  file  is opened as
+%	binary. This is the same  as   for  HTTP  resources. The correct
+%	encoding will be set by the XML parser or the Turtle parser.
+
+storage_open('', File, Stream, close(Stream)) :- !,
+	open(File, read, Stream, [type(binary)]).
+storage_open(Ext, File, Stream, Cleanup) :-
+	rdf_storage_encoding(Ext, Encoding),
+	rdf_open_decode(Encoding, File, Stream, Cleanup).
 
 valid_extension(Ext) :-
 	rdf_file_type(Ext, _).
 valid_extension(Ext) :-
 	rdf_storage_encoding(Ext, _).
 
-
-storage_extend(Path, Path).
-storage_extend(Path0, Path) :-
-	rdf_storage_encoding(Ext, _), Ext \== '',
-	file_name_extension(Path0, Ext, Path).
-
-%%	clean_base_uri(+BaseURI0, -BaseURI) is det.
-%
-%	BaseURI  is  BaseURI0  after    removing  storage  extensions.
-
-clean_base_uri(BaseURI0, BaseURI) :-
-	rdf_storage_encoding(Ext, _), Ext \== '',
-	file_name_extension(BaseURI, Ext, BaseURI0), !.
-clean_base_uri(BaseURI, BaseURI).
-
 %%	is_url(+Term, -Protocol) is semidet.
 %
 %	True if Term is an atom denoting a URL of the given Protocol.
@@ -988,46 +1181,7 @@ is_url(URL, Protocol) :-
 	downcase_atom(RawProtocol, Protocol),
 	url_protocol(Protocol).
 
-%%	exists_url(+URL) is semidet.
-%
-%	True if URL exists.
-
-%	(no default clauses)
-
-%%	url_protocol(+Protocol) is det.
-%
-%	True  if  Protocol  is  the  (lowercase)  name  of  a  supported
-%	protocol.  New  protocols  can  be    added  to  this  multifile
-%	predicate. In addition the plugin must define rdf_open_hook/3 to
-%	create a stream from the added protocol.
-
-%	(no default clauses)
-
-%%	rdf_input_info(+Input, -Modified, -Format) is semidet.
-%
-%	Return the last modification  time  of   Input  as  a POSIX time
-%	stamp as well as the format of the input.
-
-rdf_input_info(Input, Modified, Format) :-
-	rdf_input_info_hook(Input, Modified, Format), !.
-rdf_input_info(file(File), Modified, Format) :-
-	time_file(File, Modified),
-	file_name_extension(_, Ext, File),
-	(   rdf_file_type(Ext, Format)
-	->  true
-	;   Format = xml
-	).
-
-%%	rdf_input_open(+Input, -Stream, ?Format) is det.
-%
-%	Open given input as a stream.
-
-rdf_input_open(Input, Stream, Format) :-
-	rdf_open_hook(Input, Stream, Format), !.
-rdf_input_open(stream(Stream), Stream, _) :- !.
-rdf_input_open(file(File), Stream, _) :-
-	open(File, read, Stream, [type(binary)]).
-
+url_protocol(file).			% built-in
 
 %%	rdf_file_type(+Extension, -Format) is semidet.
 %
@@ -1035,6 +1189,7 @@ rdf_input_open(file(File), Stream, _) :-
 %	extension.  This predicate is multifile and can thus be extended
 %	by plugins.
 
+rdf_file_type(xml,   xml).
 rdf_file_type(rdf,   xml).
 rdf_file_type(rdfs,  xml).
 rdf_file_type(owl,   xml).
@@ -1061,41 +1216,16 @@ rdf_load_stream(xml, Stream, Options) :- !,
 	graph(Options, Graph),
 	rdf_transaction(process_rdf(Stream, assert_triples, Options),
 			parse(Graph)).
-rdf_load_stream(xhtml, Stream, Options) :- !,
+rdf_load_stream(xhtml, Stream, M:Options) :- !,
 	graph(Options, Graph),
-	rdf_transaction(process_rdf(Stream, assert_triples, [embedded(true)|Options]),
+	rdf_transaction(process_rdf(Stream, assert_triples,
+				    M:[embedded(true)|Options]),
 			parse(Graph)).
 rdf_load_stream(triples, Stream, Options) :- !,
 	graph(Options, Graph),
 	rdf_load_db_(Stream, Graph, _Graphs).
 
 
-%%	read_cache(+BaseURI, +SourceModified, -CacheFile) is semidet.
-%
-%	True if CacheFile is a cache-file modified after SourceModified.
-
-read_cache(BaseURI, Modified, CacheFile) :-
-	rdf_cache_file(BaseURI, read, CacheFile),
-	time_file(CacheFile, CacheModified),
-	CacheModified >= Modified.
-
-
-%%	must_load(+Condition, +DB, +ModifiedSrc) is semidet.
-%
-%	True if source must be reloaded.
-
-must_load(true, _, _) :- !.
-must_load(changed, DB, ModifiedSrc) :- !,
-	(   rdf_graph_source_(DB, _SourceURL, ModifiedLoaded)
-	->  ModifiedSrc > ModifiedLoaded
-	;   true
-	).
-must_load(not_loaded, DB, _) :- !,
-	\+ rdf_source(DB).
-must_load(Cond, _, _) :-
-	throw(eror(domain_error(condition, Cond), _)).
-
-
 %%	report_loaded(+Action, +Source, +DB, +Triples, +StartCPU, +Options)
 
 report_loaded(none, _, _, _, _, _) :- !.
@@ -1110,26 +1240,24 @@ report_loaded(Action, Source, DB, Triples, T0, Options) :-
 		      rdf(loaded(Action, Source, DB, Triples, Time))).
 
 
-
-%%	rdf_unload(+Spec)
+%%	rdf_unload(+Spec) is det.
 %
 %	Remove the triples loaded from the specified source and remove
 %	the source from the database.
 
-rdf_unload(DB) :-
-	atom(DB),
-	rdf_statistics_(triples(DB, _)),
-	(   rdf(_,_,_,DB)
-	;   rdf(_,_,_,DB:_)
-	), !,
-	do_unload(DB).
+rdf_unload(Graph) :-
+	atom(Graph),
+	rdf_statistics_(triples(Graph, Triples)),
+	Triples > 0, !,
+	do_unload(Graph).
 rdf_unload(Spec) :-
-	rdf_input(Spec, _, BaseURI),
-	do_unload(BaseURI).
+	source_url(Spec, _Protocol, SourceURL),
+	rdf_graph_source_(Graph, SourceURL, _), !,
+	do_unload(Graph).
+rdf_unload(_).
 
 do_unload(DB) :-
-	rdf_transaction((rdf_retractall(_,_,_,DB:_),
-			 rdf_retractall(_,_,_,DB)),
+	rdf_transaction(rdf_retractall(_,_,_,DB),
 			unload(DB)),
 	retractall(rdf_source(DB, _, _, _, _)),
 	rdf_unset_graph_source(DB).
@@ -1150,13 +1278,13 @@ rdf_graph(DB) :-
 	rdf_statistics_(triples(DB, Triples)),
 	Triples > 0.
 
-%%	rdf_source(?DB, ?Source) is nondet.
+%%	rdf_source(?Graph, ?SourceURL) is nondet.
 %
-%	True if named graph DB is loaded from Source.
+%	True if named Graph is loaded from SourceURL.
 
-rdf_source(DB, SourceURL) :-
-	rdf_graph(DB),
-	rdf_graph_source_(DB, SourceURL, _Modified).
+rdf_source(Graph, SourceURL) :-
+	rdf_graph(Graph),
+	rdf_graph_source_(Graph, SourceURL, _Modified).
 
 %%	rdf_source(?Source)
 %
@@ -1164,14 +1292,8 @@ rdf_source(DB, SourceURL) :-
 %
 %	@deprecated	Use rdf_graph/1 or rdf_source/2.
 
-rdf_source(DB) :-
-	rdf_source(DB,_,_,_,_).		% loaded files
-rdf_source(DB) :-
-	rdf_graphs_(Sources),		% other sources
-	member(DB, Sources),
-	\+ rdf_source(DB,_,_,_,_),
-	rdf_statistics_(triples(DB, Triples)),
-	Triples > 0.
+rdf_source(SourceURL) :-
+	rdf_source(_Graph, SourceURL).
 
 %%	rdf_make
 %
@@ -1179,9 +1301,15 @@ rdf_source(DB) :-
 %	time they were loaded.
 
 rdf_make :-
-	forall((rdf_source(DB, SourceURL, Time, _Triples, _),
-		Time \== 0),
-	       catch(rdf_load(SourceURL, [db(DB)]), _, true)).
+	findall(Source-Graph, modified_graph(Source, Graph), Modified),
+	forall(member(Source-Graph, Modified),
+	       catch(rdf_load(Source, [graph(Graph), if(changed)]), E,
+		     print_message(error, E))).
+
+modified_graph(SourceURL, Graph) :-
+	rdf_graph(Graph),
+	rdf_graph_source_(Graph, SourceURL, Modified),
+	Modified > 0.
 
 %%	save_cache(+DB, +Cache) is det.
 %
@@ -1314,7 +1442,7 @@ rdf_save2(File, Options) :-
 				  Close)).
 
 open_output(stream(Out), Encoding, Out,
-	    set_stream(Out, encoding(Old))) :-
+	    set_stream(Out, encoding(Old))) :- !,
 	stream_property(Out, encoding(Old)),
 	set_stream(Out, encoding(Encoding)).
 open_output(File, Encoding, Out,
@@ -1343,8 +1471,8 @@ cleanup_save(Reason,
 	;   format(user_error, 'Reason = ~w~n', [Reason])
 	).
 
-rdf_do_save(Out, Options) :-
-	rdf_save_header(Out, Options),
+rdf_do_save(Out, Options0) :-
+	rdf_save_header(Out, Options0, Options),
 	(   option(sorted(true), Options, false)
 	->  setof(Subject, rdf_subject(Subject, Options), Subjects),
 	    forall(member(Subject, Subjects),
@@ -1382,24 +1510,26 @@ graph(Options0, DB) :-
 %	This predicate also sets up the namespace notation.
 
 rdf_save_header(Out, Options) :-
+	rdf_save_header(Out, Options, _).
+
+rdf_save_header(Out, Options, OptionsOut) :-
 	is_list(Options), !,
 	stream_property(Out, encoding(Enc)),
 	xml_encoding(Enc, Encoding),
 	format(Out, '<?xml version=\'1.0\' encoding=\'~w\'?>~n', [Encoding]),
 	format(Out, '<!DOCTYPE rdf:RDF [', []),
-	header_namespaces(Options, NSList),
-	(   member(Id, NSList),
-	    ns(Id, NS),
-	    rdf_quote_uri(NS, QNS),
-	    xml_quote_attribute(QNS, NSText0, Enc),
-	    xml_escape_parameter_entity(NSText0, NSText),
-	    format(Out, '~N    <!ENTITY ~w \'~w\'>', [Id, NSText]),
-	    fail
-	;   true
-	),
+	header_namespaces(Options, NSIdList),
+	nsmap(NSIdList, NsMap),
+	append(Options, [nsmap(NsMap)], OptionsOut),
+	forall(member(Id=URI, NsMap),
+	       (   rdf_quote_uri(URI, QURI),
+		   xml_quote_attribute(QURI, NSText0, Enc),
+		   xml_escape_parameter_entity(NSText0, NSText),
+		   format(Out, '~N    <!ENTITY ~w \'~w\'>', [Id, NSText])
+	       )),
 	format(Out, '~N]>~n~n', []),
 	format(Out, '<rdf:RDF', []),
-	(   member(Id, NSList),
+	(   member(Id, NSIdList),
 	    format(Out, '~N    xmlns:~w="&~w;"~n', [Id, Id]),
 	    fail
 	;   true
@@ -1416,9 +1546,9 @@ rdf_save_header(Out, Options) :-
 	;   true
 	),
 	format(Out, '>~n', []).
-rdf_save_header(Out, FileRef) :-	% compatibility
+rdf_save_header(Out, FileRef, OptionsOut) :-	% compatibility
 	atom(FileRef),
-	rdf_save_header(Out, [graph(FileRef)]).
+	rdf_save_header(Out, [graph(FileRef)], OptionsOut).
 
 xml_encoding(Enc, Encoding) :-
 	(   xml_encoding_name(Enc, Encoding)
@@ -1430,6 +1560,15 @@ xml_encoding_name(ascii,       'US-ASCII').
 xml_encoding_name(iso_latin_1, 'ISO-8859-1').
 xml_encoding_name(utf8,        'UTF-8').
 
+%%	nsmap(+NSIds, -Map:list(id=uri)) is det.
+%
+%	Create a namespace-map that is compatible to xml_write/2
+%	for dealing with XML-Literals
+
+nsmap([], []).
+nsmap([Id|T0], [Id=URI|T]) :-
+	ns(Id, URI),
+	nsmap(T0, T).
 
 %%	xml_escape_parameter_entity(+In, -Out) is det.
 %
@@ -1806,29 +1945,7 @@ save_attribute(body, Name=literal(Literal0), BaseURI, Out, Indent, Options) :- !
 	->  Literal = type(Type, Content)
 	;   Literal = Literal0
 	),
-	format(Out, '~N~*|<', [Indent]),
-	rdf_write_id(Out, NameText),
-	(   Literal = lang(Lang, Value)
-	->  (   memberchk(document_language(Lang), Options)
-	    ->  write(Out, '>')
-	    ;   rdf_id(Lang, BaseURI, LangText),
-		format(Out, ' xml:lang="~w">', [LangText])
-	    )
-	;   Literal = type(Type, Value)
-	->  (   rdf_equal(Type, rdf:'XMLLiteral')
-	    ->	write(Out, ' rdf:parseType="Literal">')
-	    ;	stream_property(Out, encoding(Encoding)),
-		rdf_value(Type, BaseURI, QVal, Encoding),
-		format(Out, ' rdf:datatype="~w">', [QVal])
-	    )
-	;   atomic(Literal)
-	->  write(Out, '>'),
-	    Value = Literal
-	;   write(Out, ' rdf:parseType="Literal">'),
-	    Value = Literal
-	),
-	save_attribute_value(Value, Out, Indent),
-	write(Out, '</'), rdf_write_id(Out, NameText), write(Out, '>').
+	save_body_literal(Literal, NameText, BaseURI, Out, Indent, Options).
 save_attribute(body, Name=Value, BaseURI, Out, Indent, Options) :-
 	rdf_is_bnode(Value), !,
 	rdf_id(Name, BaseURI, NameText),
@@ -1864,6 +1981,47 @@ save_attribute(body, Name=Value, BaseURI, Out, Indent, _DB) :-
 	rdf_write_id(Out, NameText),
 	format(Out, ' rdf:resource="~w"/>', [QVal]).
 
+%%	save_body_literal(+Literal, +NameText, +BaseURI,
+%%			  +Out, +Indent, +Options).
+
+save_body_literal(lang(Lang, Value),
+		  NameText, BaseURI, Out, Indent, Options) :- !,
+	format(Out, '~N~*|<', [Indent]),
+	rdf_write_id(Out, NameText),
+	(   memberchk(document_language(Lang), Options)
+	->  write(Out, '>')
+	;   rdf_id(Lang, BaseURI, LangText),
+	    format(Out, ' xml:lang="~w">', [LangText])
+	),
+	save_attribute_value(Value, Out, Indent),
+	write(Out, '</'), rdf_write_id(Out, NameText), write(Out, '>').
+save_body_literal(type(Type, DOM),
+		  NameText, _BaseURI, Out, Indent, Options) :-
+	rdf_equal(Type, rdf:'XMLLiteral'), !,
+	save_xml_literal(DOM, NameText, Out, Indent, Options).
+save_body_literal(type(Type, Value),
+		  NameText, BaseURI, Out, Indent, _) :- !,
+	format(Out, '~N~*|<', [Indent]),
+	rdf_write_id(Out, NameText),
+	stream_property(Out, encoding(Encoding)),
+	rdf_value(Type, BaseURI, QVal, Encoding),
+	format(Out, ' rdf:datatype="~w">', [QVal]),
+	save_attribute_value(Value, Out, Indent),
+	write(Out, '</'), rdf_write_id(Out, NameText), write(Out, '>').
+save_body_literal(Literal,
+		  NameText, _, Out, Indent, _) :-
+	atomic(Literal), !,
+	format(Out, '~N~*|<', [Indent]),
+	rdf_write_id(Out, NameText),
+	write(Out, '>'),
+	save_attribute_value(Literal, Out, Indent),
+	write(Out, '</'), rdf_write_id(Out, NameText), write(Out, '>').
+save_body_literal(DOM,
+		  NameText, BaseURI, Out, Indent, Options) :-
+	rdf_equal(Type, rdf:'XMLLiteral'),
+	save_body_literal(type(Type, DOM),
+			  NameText, BaseURI, Out, Indent, Options).
+
 save_attribute_value(Value, Out, _) :-	% strings
 	atom(Value), !,
 	stream_property(Out, encoding(Encoding)),
@@ -1872,16 +2030,41 @@ save_attribute_value(Value, Out, _) :-	% strings
 save_attribute_value(Value, Out, _) :-	% numbers
 	number(Value), !,
 	writeq(Out, Value).		% quoted: preserve floats
-save_attribute_value(Value, Out, Indent) :-
-	xml_is_dom(Value), !,
-	XMLIndent is Indent+2,
-	xml_write(Out, Value,
-		  [ header(false),
-		    indent(XMLIndent)
-		  ]).
 save_attribute_value(Value, _Out, _) :-
 	throw(error(save_attribute_value(Value), _)).
 
+%%	save_xml_literal(+DOM, +Attr, +Out, +Indent, +Options) is det.
+%
+%	Save an XMLLiteral value. We already emitted
+%
+%		==
+%		<prop parseType="literal"
+%		==
+%
+%	but  not  the  terminating  =|>|=.  We  need  to  establish  the
+%	namespaces used in the DOM. The   namespaces in the rdf document
+%	are in the nsmap-option of Options.
+
+save_xml_literal(DOM, Attr, Out, Indent, Options) :-
+	xml_is_dom(DOM), !,
+	memberchk(nsmap(NsMap), Options),
+	id_to_atom(Attr, Atom),
+	xml_write(Out,
+		  element(Atom, ['rdf:parseType'='Literal'], DOM),
+		  [ header(false),
+		    indent(Indent),
+		    nsmap(NsMap)
+		  ]).
+save_xml_literal(NoDOM, _, _, _, _) :-
+	must_be(xml_dom, NoDOM).
+
+id_to_atom(NS:Local, Atom) :- !,
+	atomic_list_concat([NS,Local], :, Atom).
+id_to_atom(ID, ID).
+
+
+%%	rdf_save_list(+Out, +List, +BaseURI, +Indent, +Options)
+
 rdf_save_list(_, List, _, _, _) :-
 	rdf_equal(List, rdf:nil), !.
 rdf_save_list(Out, List, BaseURI, Indent, Options) :-
@@ -2007,24 +2190,21 @@ prolog:message(rdf(inconsistent_cache(DB, Graphs))) -->
 	[ 'RDF cache file for ~w contains the following graphs'-[DB], nl,
 	  '~t~8|~p'-[Graphs]
 	].
+prolog:message(rdf(guess_format(Ext))) -->
+	[ 'Unknown file-extension: ~w.  Assuming RDF/XML'-[Ext] ].
+prolog:message(rdf_meta(not_expanded(G))) -->
+	[ 'rdf_meta: ~p is not expanded'-[G] ].
 
 how(load)   --> [ 'Loaded' ].
 how(parsed) --> [ 'Parsed' ].
 
-source(stream(Stream)) -->
-	{ stream_property(Stream, file_name(Path)), !,
-	  file_base_name(Path, Base)
+source(SourceURL) -->
+	{ uri_file_name(SourceURL, File), !,
+	  file_base_name(File, Base)	% TBD: relative file?
 	},
-	[ ' "~w"'-[Base] ].
-source(file(Path)) -->
-	{ atom(Path), !,
-	  file_base_name(Path, Base)
-	},
-	[ ' "~w"'-[Base] ].
-source(url(_Protocol, URL)) -->
-	[ ' "~w"'-[URL] ].
-source(Spec) -->
-	[ ' "~p"'-[Spec] ].
+	[ ' "~p"'-[Base] ].
+source(SourceURL) -->
+	[ ' "~p"'-[SourceURL] ].
 
 into(_, _) --> [].			% TBD
 
