@@ -1,4 +1,4 @@
-/*  $Id: shell.pl,v 1.17 2005/04/06 20:22:38 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -112,7 +112,7 @@ update_label(B) :->
 	    send(DictItem, label, string('%s\t%s', Name, Indicator)),
 	    new(FrameLabel, string('%s [%s]', Name, Status)),
 	    send(B?editors, for_all,
-		 message(@arg1?frame, label, FrameLabel))
+		 message(@arg1, label, FrameLabel))
 	;   true
 	).
 
@@ -279,7 +279,7 @@ properties(B) :->
 	),
 	send(V, appendf, '\n'),
 	send(V, appendf, 'Status:\t%s\n', Process?status).
-	
+
 
 		 /*******************************
 		 *      WINDOW INTERFACE	*
@@ -292,25 +292,13 @@ start_process(B) :->
 	    send(B, update_label)
 	;   send(B, report, warning, 'No process')
 	).
-	
 
-open(B, New:[bool]) :->
+
+open(B, Where:[{here,tab,window}]) :->
 	"Create window for buffer"::
-	(   New == @on
-	->  send(new(Window, emacs_frame(B)), open)
-	;   (	\+ send(B?editors, empty)
-	    ->	send(B?editors?head?frame, expose)
-	    ;	get(@emacs, free_window, B?pool, Window)
-	    ->	send(Window, buffer, B)
-	    ;	send(emacs_frame(B), open)
-	    )
-	),
-	(   nonvar(Window)
-	->  send(Window?editor, caret, @default)
-	;   true
-	),
+	get(B, open, Where, Frame),
+	send(Frame?editor, caret, @default),
 	send(B, update_label).
-	
 
 :- pce_end_class.
 
@@ -352,7 +340,7 @@ open(B, New:[bool]) :->
 	send(MM, append, shell, end_of_file),
 	send(MM, append, shell, interrupt),
 	send(MM, append, shell, kill_subjob).
-	
+
 
 variable(current_history, number*, both, "Current history number").
 
@@ -429,7 +417,7 @@ insert_history(N, Chain, E) :-
 	),
 	NN is N - 1,
 	insert_history(NN, Chain, E).
-	
+
 
 backward_history(E, Id:event_id) :->
 	"Insert last command"::
@@ -604,7 +592,7 @@ mark_errors(M, From:[int], To:[int]) :->
 		 ;   send(M, report, status, 'No errors')
 		 )
 	     ).
-	     
+
 
 goto_error(M) :->
 	"Goto error from current_line"::
@@ -613,10 +601,10 @@ goto_error(M) :->
 		and(message(@arg1, instance_of, emacs_link_fragment),
 		    message(@arg1, overlap, Caret)),
 		Fragment)
-	->  send(Fragment, follow)
+	->  send(Fragment, follow, tab)
 	;   get(M, error, Caret, tuple(File, Line))
 	->  new(B, emacs_buffer(File)),
-	    send(B, open),
+	    send(B, open, tab),
 	    send(B?editors?head, line_number, Line)
 	;   send(M, report, warning, 'No error on this line')
 	).
@@ -655,18 +643,18 @@ unlink(F) :->
 	;   true
 	),
 	send(F, send_super, unlink).
-	
+
 
 link(F, F2:emacs_link_fragment) :->
 	send(F, slot, link, F2),
 	send(F2, slot, link, F).
 
 
-follow(F) :->
+follow(F, Where:[{here,tab,window}]) :->
 	"Goto other end of link"::
 	get(F, link, F2),
 	get(F2, text_buffer, B),
-	send(B, open),
+	send(B, open, Where),
 	send(B?editors?head, caret, F2?start).
 
 :- pce_end_class.
