@@ -54,8 +54,8 @@ This module is normally hooked into help/1 by the module swi_hooks.pl.
 	   , atom_length/2
 	   , between/3
 	   , chain_list/2
-	   , concat_atom/2
-	   , concat_atom/3
+	   , atomic_list_concat/2
+	   , atomic_list_concat/3
 	   , default/3
 	   , explain/2
 	   , forall/2
@@ -83,7 +83,7 @@ resource(predicate,	image,  image('16x16/preddoc.xpm')).
 
 prolog_help :-
 	section(S, 'Getting started quickly', _, _),
-	concat_atom(S, -, Id),
+	atomic_list_concat(S, -, Id),
 	term_to_atom(Spec, Id),
 	prolog_help(Spec).
 
@@ -230,7 +230,7 @@ fill_menu_bar(F) :->
 		    menu_item('SWI-Prolog Download (on www) ...',
 			      message(F, open_url, pl_download),
 			      end_group := @on),
-		    
+
 		    menu_item('XPCE (GUI) Manual ...',
 			      message(@prolog, manpce)),
 		    menu_item('XPCE User Guide (on www) ...',
@@ -386,7 +386,7 @@ setup_isearch(V) :->
 				message(@receiver, focus_function, @nil)),
 			    message(@receiver, send_class, '_isearch',
 				    @arg1)))).
-	
+
 
 displayed_ranges(V, Displayed:prolog) :->
 	send(V, slot, displayed_ranges, Displayed),
@@ -410,7 +410,7 @@ give_help(V, What:name,
 	    ;	send(V, clear)
 	    ),
 	    manual_file(ManFile),
-	    new(F, file(ManFile)),
+	    new(F, file(ManFile, binary)),
 	    get(V, caret, Start),
 	    send(F, open, read),
 	    append_ranges(V, F, Ranges),
@@ -437,13 +437,23 @@ give_help(V, What:name,
 	),
 	send(V, editable, @off).
 
+%%	append_ranges(+View, +File, +Ranges) is det.
+%
+%	Note that the file is opened in binary mode to allow seeking. We
+%	must delete \r from the input   to  compensate for Windows cr/lf
+%	line ends. This is all ok as long  as the contents of the manual
+%	file is ISO Latin 1.
+
 append_ranges(V, F, [H|T]) :- !,
 	append_ranges(V, F, H),
 	append_ranges(V, F, T).
 append_ranges(_, _, []) :- !.
-append_ranges(V, F, From-To) :-
+append_ranges(V, F, FromLine-ToLine) :-
+	line_start(FromLine, From),
+	line_start(ToLine, To),
 	send(F, seek, From),
 	get(F, read, To-From, Text),
+	send(Text, translate, 13, @nil),
 	send(V, insert, Text),
 	send(V, insert, @pui_ff),
 	send(V, newline).
@@ -480,7 +490,7 @@ mark_titles(TB, From, To) :-
 	->  mark_titles(TB, Idx, To)
 	;   true
 	).
-	
+
 		 /*******************************
 		 *     CROSS-REFERENCE LINKS	*
 		 *******************************/
@@ -601,6 +611,42 @@ manual_file(_File) :-
 	send(@nil, report, error, 'Can''t find manual database MANUAL'),
 	fail.
 
+%%	line_start(Line, Start) is det.
+%
+%	True if Start is the byte position at which Line starts.
+
+:- dynamic
+	start_of_line/2.
+
+line_start(Line, Start) :-
+	start_of_line(Line, Start), !.
+line_start(Line, Start) :-
+	line_index,
+	start_of_line(Line, Start).
+
+
+%%	line_index
+%
+%	Create index holding the byte positions for the line starts
+
+line_index :-
+	start_of_line(_,_), !.
+line_index :-
+	manual_file(File),
+	open(File, read, Stream, [type(binary)]),
+	call_cleanup(line_index(Stream, 1), close(Stream)).
+
+line_index(Stream, LineNo) :-
+	byte_count(Stream, ByteNo),
+	assert(start_of_line(LineNo, ByteNo)),
+	(   at_end_of_stream(Stream)
+	->  true
+	;   LineNo2 is LineNo+1,
+	    skip(Stream, 10),
+	    line_index(Stream, LineNo2)
+	).
+
+
 		 /*******************************
 		 *     ATOMIC SPECIFICATIONS	*
 		 *******************************/
@@ -609,7 +655,7 @@ manual_file(_File) :-
 %
 %	Find the character range for the given help topic, which is of
 %	the form
-%	
+%
 %		* name/arity
 %		* function()
 %		* n.m..
@@ -751,11 +797,11 @@ alnum(C) -->
 help_atom(Name/Arity, Atom) :- !,
 	(   var(Arity)
 	->  atom_concat(Name, '/_', Atom)
-	;   concat_atom([Name, /, Arity], Atom)
+	;   atomic_list_concat([Name, /, Arity], Atom)
 	).
 help_atom(S1-S0, Atom) :- !,
 	help_atom(S1, A0),
-	concat_atom([A0, '.', S0], Atom).
+	atomic_list_concat([A0, '.', S0], Atom).
 help_atom(C, C) :-
 	integer(C), !.
 help_atom(F, F) :-
@@ -816,7 +862,7 @@ section_apropos(V, Pattern) :-
 	send(V, insert_section,
 	     string('Sections from "%s":', Pattern)),
 	forall(member(Index-Title, Names),
-	       (   concat_atom(Index, '.', Jump),
+	       (   atomic_list_concat(Index, '.', Jump),
 		   append_apropos(V, Jump, Title)
 	       )).
 
@@ -907,8 +953,8 @@ source(File, Line) -->			% XPCE method
 	  get(Location, file_name, File),
 	  get(Location, line_no, Line)
 	}.
-	
-	
+
+
 		 /*******************************
 		 *       TABLE-OF-CONTENTS	*
 		 *******************************/
@@ -986,7 +1032,7 @@ function_in_section(Section, Func) :-
 
 subsection(Sec, Sub) :-
 	append(Sec, [_], Sub).
-	
+
 select_range(PT, Ranges:prolog) :->
 	"Select selections in range"::
 	(   Ranges == []
@@ -998,7 +1044,7 @@ select_range(PT, Ranges:prolog) :->
 	    get(PT, selection, Nodes),
 	    send(PT, normalise, Nodes, y)
 	).
-	    
+
 in_range(F-T, Node, How) :-
 	get(Node, start, S),
 	get(Node, end, E),
@@ -1029,7 +1075,7 @@ initialise(N, Index:prolog) :->
 	->  CanExpand = @on
 	;   CanExpand = @off
 	),
-	concat_atom(Index, '.', Id),
+	atomic_list_concat(Index, '.', Id),
 	send_super(N, initialise, Title?capitalise, Id,
 		   resource(manual), resource(book),
 		   CanExpand),
@@ -1053,7 +1099,7 @@ select_range(N, Ranges:prolog) :->
 	;   true
 	).
 
-:- pce_end_class(pui_section_node). 
+:- pce_end_class(pui_section_node).
 
 
 :- pce_begin_class(pui_predicate_node, toc_file,
