@@ -1,4 +1,4 @@
-/*  $Id: swi_ide.pl,v 1.8 2006/03/26 16:27:43 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -46,19 +46,19 @@ the IDE components to the autoloading of one single predicate.
 
 :- pce_image_directory(library('trace/icons')).
 
-:- pce_autoload(prolog_debug_status,   library('trace/status')).
-:- pce_autoload(prolog_navigator,      library('trace/browse')).
-:- pce_autoload(prolog_query_frame,    library('trace/query')).
-:- pce_autoload(prolog_thread_monitor, library('swi/thread_monitor')).
-:- pce_autoload(prolog_debug_monitor,  library('swi/pce_debug_monitor')).
-:- pce_autoload(xref_frame,	       library('pce_xref')).
-
+:- pce_autoload(prolog_debug_status,	library('trace/status')).
+:- pce_autoload(prolog_navigator,	library('trace/browse')).
+:- pce_autoload(prolog_query_frame,	library('trace/query')).
+:- pce_autoload(prolog_trace_exception,	library('trace/exceptions')).
+:- pce_autoload(prolog_thread_monitor,	library('swi/thread_monitor')).
+:- pce_autoload(prolog_debug_monitor,	library('swi/pce_debug_monitor')).
+:- pce_autoload(xref_frame,		library('pce_xref')).
 
 		 /*******************************
 		 *	      TOPLEVEL		*
 		 *******************************/
 
-%	prolog_ide(+Action)
+%%	prolog_ide(+Action)
 %
 %	Invoke an action on the (SWI-)Prolog IDE application.  This is a
 %	predicate to ensure optimal delaying of loading and object creation
@@ -66,7 +66,7 @@ the IDE components to the autoloading of one single predicate.
 %	Development Environment.
 
 prolog_ide(Action) :-
-	send(@prolog_ide, Action).
+	in_pce_thread(send(@prolog_ide, Action)).
 
 
 		 /*******************************
@@ -74,6 +74,7 @@ prolog_ide(Action) :-
 		 *******************************/
 
 :- pce_global(@prolog_ide, new(prolog_ide)).
+:- pce_global(@prolog_exception_window, new(prolog_trace_exception)).
 
 :- pce_begin_class(prolog_ide, application, "Prolog IDE application").
 
@@ -89,6 +90,19 @@ open_debug_status(IDE) :->
 	;   send(prolog_debug_status(IDE), open)
 	).
 
+open_exceptions(IDE, Gui:[bool]) :->
+	"Open/show exceptions"::
+	W = @prolog_exception_window,
+	(   object(W)
+	->  send(W, expose)
+	;   (   Gui == @on
+	    ->	catch(tdebug, _, guitracer)
+	    ;	true
+	    ),
+	    send(W, application, IDE),
+	    send(W, open)
+	).
+
 open_navigator(IDE, Where:[directory|source_location]) :->
 	"Open Source Navigator"::
 	(   send(Where, instance_of, directory)
