@@ -1,4 +1,4 @@
-/*  $Id: tabbed_window.pl,v 1.8 2004/09/21 14:22:25 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -59,6 +59,8 @@ initialise(W, Label:label=[name], Size:size=[size],
 	send_super(W, initialise, Label, Size, Display),
 	send(W, hor_stretch, 100),
 	send(W, ver_stretch, 100),
+	send(W, hor_shrink, 100),
+	send(W, ver_shrink, 100),
 	send(W, pen, 0),
 	send(W, border, size(0,0)),
 	send_super(W, append, new(tab_stack)).
@@ -77,7 +79,7 @@ resize(W, Tab:[tab]) :->
 		 message(@arg1, size, size(Width,TabH)))
 	;   send(Tab, size, size(Width,TabH))
 	).
-    
+
 layout_dialog(W, _Gap:[size], _Size:[size], _Border:[size]) :->
 	"Overrule to deal with nested tabbed windows"::
 	new(S0, size(0,0)),
@@ -85,19 +87,39 @@ layout_dialog(W, _Gap:[size], _Size:[size], _Border:[size]) :->
 
 :- pce_group(stack).
 
-on_top(W, Name:name) :->
-	"Put the named tab on top"::
+on_top(W, Top:'name|window') :->
+	"Put the named tab or tab containing Window on top"::
 	get_super(W, member, tab_stack, TS),
-	(   get(TS, member, Name, Tab)
+	(   atom(Top)
+	->  (   get(TS, member, Top, Tab)
+	    ->  send(TS, on_top, Tab)
+	    ;   get(W, hypered, tab, @arg3?name == Top, Window)
+	    ->  send(Window, expose)
+	    )
+	;   get(Top, container, window_tab, Tab)
 	->  send(TS, on_top, Tab)
-	;   get(W, hypered, tab, @arg3?name == Name, Window)
-	->  send(Window, expose)
+	).
+
+
+current(W, Window:window) :<-
+	"Window of currently selected tab"::
+	get_super(W, member, tab_stack, TS),
+	get(TS, on_top, Tab),
+	get(Tab, window, Window).
+
+current(W, Window:window) :->
+	"Window of currently selected tab"::
+	get(Window, container, window_tab, Tab),
+	(   get(Tab, status, on_top)
+	->  send(W, resize, Tab)
+	;   get_super(W, member, tab_stack, TS),
+	    send(TS, on_top, Tab)
 	).
 
 :- pce_group(members).
 
 %	->append: Window, Label, [Expose]
-%	
+%
 %	Append a new tab using Window with the given tab label.
 %
 %	The call to ->'_compute_desired_size' should be properly delayed
@@ -133,7 +155,7 @@ members(W, Windows:chain) :<-
 		    message(Windows, append, @arg1?to)))
 	;   true
 	).
-	
+
 clear(W) :->
 	"Remove all member tabs"::
 	get_super(W, member, tab_stack, TS),
@@ -153,6 +175,17 @@ tab(W, Name:name, Tab:tab) :<-
 	get_super(W, member, tab_stack, TS),
 	get(TS, member, Name, Tab).
 
+empty(_W) :->
+	"Abstract method.  Called if last window disappears"::
+	true.
+
+:- pce_group(frame).
+
+frame_window(TW, Window:window, Name:name, Rank:'1..', Frame:frame) :<-
+	"After un-tabbing, give the window a new frame"::
+	new(Frame, window_tab_frame(Window, Name, Rank)),
+	new(_, partof_hyper(TW, Window, toplevel, tab)).
+
 :- pce_end_class(tabbed_window).
 
 
@@ -197,11 +230,24 @@ initialise(T, Window:window=[window], Name:name=[name]) :->
 	send(T, slot, window, W),
 	new(_, mutual_dependency_hyper(T, W, window, tab)).
 
+unlink(Tab) :->
+	"Trap if I'm the last tab"::
+	(   get(Tab, device, Dev),
+	    Dev \== @nil
+	->  get(Dev?graphicals, size, Count),
+	    (   Count == 1
+	    ->  get(Tab, container, tabbed_window, TabbedWindow),
+		send_super(Tab, unlink),
+		send(TabbedWindow, empty)
+	    ;   send_super(Tab, unlink)
+	    )
+	;   send_super(Tab, unlink)
+	).
 
 :- pce_group(resize).
 
 %	->size
-%	
+%
 %	This method must update the size of  the window. For some, to me
 %	unknown,  reason  this  does  not    work  correctly  when  done
 %	immediately.  Possibly  this  has  something   to  do  with  X11
@@ -233,7 +279,7 @@ status(T, Status:{on_top,hidden}) :->
 	(   Status == on_top,
 	    get(T, is_displayed, @on),
 	    get(T, container, tabbed_window, TabbedWindow)
-	->  send(TabbedWindow, resize, T)
+	->  send(TabbedWindow, current, T?window)
 	;   true
 	).
 
@@ -301,7 +347,7 @@ untab(Tab, W:window) :<-
 	send(Tab, delete_hypers, window),
 	free(Tab),
 	get(W, unlock, _).
-	
+
 untab(Tab) :->
 	"Turn the window into a toplevel window"::
 	get(Tab, rank, Rank),
@@ -309,12 +355,12 @@ untab(Tab) :->
 	get(Tab, container, dialog, TabbedWindow),
 	get(Tab, display_position, point(X, Y)),
 	get(Tab, untab, Window),
-	send(new(window_tab_frame(Window, Name, Rank)), open, point(X, Y+20)),
-	new(_, partof_hyper(TabbedWindow, Window, toplevel, tab)).
+	get(TabbedWindow, frame_window, Window, Name, Rank, Frame),
+	send(Frame, open, point(X, Y+20)).
 
 %	->close_other_tabs
-%	
-%	Close all tabs but be. To work   around scheduled resize for the
+%
+%	Close all tabs but me. To work   around scheduled resize for the
 %	subwindows we first indicate we are about to close the tabs. See
 %	also ->size.
 
