@@ -73,6 +73,9 @@ test(shared, X == '') :-
 	atom_concat(X, X, '').
 test(shared, fail) :-
 	atom_concat(X, X, abac).
+test(attvar, X-Y == ab-ok) :-
+	freeze(X, Y = ok),
+	atom_concat(a, b, X).
 
 :- end_tests(atom_concat).
 
@@ -92,5 +95,7 @@ test(read, T-V =@= foo(A)-['A' = A] ) :-
 	atom_to_term('foo(A)', T, V).
 test(error, error(instantiation_error)) :-
 	atom_to_term(_, _, _).
+test(eof, error(syntax_error(_))) :-
+	atom_to_term('x /* comment', _, _).
 
 :- end_tests(atom_to_term).
