@@ -34,11 +34,13 @@ unification is wrong you won't get as far as running this test :-)
 */
 
 test_unify :-
-	run_tests([ unify
+	run_tests([ unify,
+		    can_compare
 		  ]).
 
 :- begin_tests(unify).
 
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 blam([]).
 blam([L|L]) :- blam(L).
 
@@ -46,4 +48,31 @@ test(blam, X == [[[]], []]) :-
 	blam(X),
 	length(X, 2), !.
 
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+p(X,Y):-				% Bug#436
+  U=U,
+  V=V,
+  findall(_C,
+          q(X,Y,U,V),
+          _Cs).
+
+q(_,_,_,_).
+
+test(unify_self, true) :-
+	p(_,_).
+
 :- end_tests(unify).
+
+:- begin_tests(can_compare).
+
+v(_).
+
+test(ground, true) :-
+	?=(a,b).
+test(ground, true) :-
+	?=(a,a).
+test(ground, fail) :-
+	v(X),
+	?=(a,X).
+
+:- end_tests(can_compare).
