@@ -35,20 +35,53 @@ throw(error(<Formal>, <SWI-Prolog>))
 #include <errno.h>
 #endif
 
-static void
+static int
 put_name_arity(term_t t, functor_t f)
-{ FunctorDef fdef = valueFunctor(f);
-  term_t a = PL_new_term_refs(2);
+{ GET_LD
+  FunctorDef fdef = valueFunctor(f);
+  term_t a;
 
-  PL_put_atom(a+0, fdef->name);
-  PL_put_integer(a+1, fdef->arity);
-  PL_cons_functor(t, FUNCTOR_divide2, a+0, a+1);
+  if ( (a=PL_new_term_refs(2)) )
+  { PL_put_atom(a+0, fdef->name);
+
+    return (PL_put_integer(a+1, fdef->arity) &&
+	    PL_cons_functor(t, FUNCTOR_divide2, a+0, a+1));
+  }
+
+  return FALSE;
+}
+
+
+static void
+rewrite_callable(atom_t *expected, term_t actual)
+{ GET_LD
+  term_t a = 0;
+  int loops = 0;
+
+  while ( PL_is_functor(actual, FUNCTOR_colon2) )
+  { if ( !a )
+     a = PL_new_term_ref();
+
+    _PL_get_arg(1, actual, a);
+    if ( !PL_is_atom(a) )
+    { *expected = ATOM_atom;
+      PL_put_term(actual, a);
+      return;
+    } else
+    { _PL_get_arg(2, actual, a);
+      PL_put_term(actual, a);
+    }
+
+    if ( ++loops > 100 && !PL_is_acyclic(actual) )
+      break;
+  }
 }
 
 
 int
 PL_error(const char *pred, int arity, const char *msg, int id, ...)
-{ Definition caller;
+{ GET_LD
+  Definition caller;
   term_t except, formal, swi;
   va_list args;
   int do_throw = FALSE;
@@ -60,7 +93,8 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
   else
     caller = NULL;
 
-  if ( id == ERR_FILE_OPERATION && !fileerrors )
+  if ( id == ERR_FILE_OPERATION &&
+       !truePrologFlag(PLFLAG_FILEERRORS) )
     fail;
 
   if ( msg == MSG_ERRNO )
@@ -69,7 +103,11 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     msg = OsError();
   }
 
-  fid    = PL_open_foreign_frame();
+  LD->exception.processing = TRUE;	/* allow using spare stack */
+
+  if ( !(fid = PL_open_foreign_frame()) )
+    goto nomem;
+
   except = PL_new_term_ref();
   formal = PL_new_term_ref();
   swi    = PL_new_term_ref();
@@ -79,35 +117,37 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
   switch(id)
   { case ERR_INSTANTIATION:
       err_instantiation:
-      PL_unify_atom(formal, ATOM_instantiation_error);
+      rc = PL_unify_atom(formal, ATOM_instantiation_error);
       break;
     case ERR_MUST_BE_VAR:
     { int argn = va_arg(args, int);
-      term_t bound = va_arg(args, term_t);
       char buf[50];
+      /*term_t bound =*/ va_arg(args, term_t);
 
-      if ( !msg )
+      if ( !msg && argn > 0 )
       { Ssprintf(buf, "%d-%s argument",
 		 argn, argn == 1 ? "st" : argn == 2 ? "nd" : "th");
 	msg = buf;
       }
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_representation_error1,
-		      PL_ATOM, ATOM_variable);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_representation_error1,
+			   PL_ATOM, ATOM_variable);
       break;
     }
     case ERR_TYPE:			/* ERR_INSTANTIATION if var(actual) */
     { atom_t expected = va_arg(args, atom_t);
       term_t actual   = va_arg(args, term_t);
 
+      if ( expected == ATOM_callable )
+	rewrite_callable(&expected, actual);
       if ( PL_is_variable(actual) && expected != ATOM_variable )
 	goto err_instantiation;
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_type_error2,
-		      PL_ATOM, expected,
-		      PL_TERM, actual);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_type_error2,
+			   PL_ATOM, expected,
+			   PL_TERM, actual);
       break;
     }
     case ERR_CHARS_TYPE:		/* ERR_INSTANTIATION if var(actual) */
@@ -117,10 +157,10 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       if ( PL_is_variable(actual) && !streq(expected, "variable") )
 	goto err_instantiation;
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_type_error2,
-		      PL_CHARS, expected,
-		      PL_TERM, actual);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_type_error2,
+			   PL_CHARS, expected,
+			   PL_TERM, actual);
       break;
     }
     case ERR_AR_TYPE:			/* arithmetic type error */
@@ -128,29 +168,29 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       Number num      = va_arg(args, Number);
       term_t actual   = PL_new_term_ref();
 
-      _PL_put_number(actual, num);
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_type_error2,
-		      PL_ATOM, expected,
-		      PL_TERM, actual);
+      rc = (_PL_put_number(actual, num) &&
+	    PL_unify_term(formal,
+			  PL_FUNCTOR, FUNCTOR_type_error2,
+			    PL_ATOM, expected,
+			    PL_TERM, actual));
       break;
     }
     case ERR_AR_UNDEF:
-    { PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_evaluation_error1,
-		      PL_ATOM, ATOM_undefined);
+    { rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_evaluation_error1,
+			   PL_ATOM, ATOM_undefined);
       break;
     }
     case ERR_AR_OVERFLOW:
-    { PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_evaluation_error1,
-		      PL_ATOM, ATOM_float_overflow);
+    { rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_evaluation_error1,
+			   PL_ATOM, ATOM_float_overflow);
       break;
     }
     case ERR_AR_UNDERFLOW:
-    { PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_evaluation_error1,
-		      PL_ATOM, ATOM_float_underflow);
+    { rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_evaluation_error1,
+			   PL_ATOM, ATOM_float_underflow);
       break;
     }
     case ERR_DOMAIN:			/*  ERR_INSTANTIATION if var(arg) */
@@ -160,30 +200,44 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       if ( PL_is_variable(arg) )
 	goto err_instantiation;
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_domain_error2,
-		      PL_ATOM, domain,
-		      PL_TERM, arg);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_domain_error2,
+			   PL_ATOM, domain,
+			   PL_TERM, arg);
       break;
     }
     case ERR_REPRESENTATION:
     { atom_t what = va_arg(args, atom_t);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_representation_error1,
-		      PL_ATOM, what);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_representation_error1,
+			   PL_ATOM, what);
       break;
     }
     case ERR_MODIFY_STATIC_PROC:
     { Procedure proc = va_arg(args, Procedure);
       term_t pred = PL_new_term_ref();
 
-      unify_definition(pred, proc->definition, 0, GP_NAMEARITY|GP_HIDESYSTEM);
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_permission_error3,
-		      PL_ATOM, ATOM_modify,
-		      PL_ATOM, ATOM_static_procedure,
-		      PL_TERM, pred);
+      rc = (unify_definition(MODULE_user, pred, proc->definition, 0,
+			     GP_NAMEARITY|GP_HIDESYSTEM) &&
+	    PL_unify_term(formal,
+			  PL_FUNCTOR, FUNCTOR_permission_error3,
+			    PL_ATOM, ATOM_modify,
+			    PL_ATOM, ATOM_static_procedure,
+			    PL_TERM, pred));
+      break;
+    }
+    case ERR_MODIFY_THREAD_LOCAL_PROC:
+    { Procedure proc = va_arg(args, Procedure);
+      term_t pred = PL_new_term_ref();
+
+      rc = (unify_definition(MODULE_user, pred, proc->definition, 0,
+			     GP_NAMEARITY|GP_HIDESYSTEM) &&
+	    PL_unify_term(formal,
+			  PL_FUNCTOR, FUNCTOR_permission_error3,
+			    PL_ATOM, ATOM_modify,
+			    PL_ATOM, ATOM_thread_local_procedure,
+			    PL_TERM, pred));
       break;
     }
     case ERR_UNDEFINED_PROC:
@@ -194,72 +248,71 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       if ( clr )
 	caller = clr;
 
-      unify_definition(pred, def, 0, GP_NAMEARITY);
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_existence_error2,
-		      PL_ATOM, ATOM_procedure,
-		      PL_TERM, pred);
+      rc = (unify_definition(MODULE_user, pred, def, 0, GP_NAMEARITY) &&
+	    PL_unify_term(formal,
+			  PL_FUNCTOR, FUNCTOR_existence_error2,
+			    PL_ATOM, ATOM_procedure,
+			    PL_TERM, pred));
       break;
     }
     case ERR_PERMISSION_PROC:
     { atom_t op = va_arg(args, atom_t);
       atom_t type = va_arg(args, atom_t);
-      Definition def = va_arg(args, Definition);
-      term_t pred = PL_new_term_ref();
+      predicate_t pred = va_arg(args, predicate_t);
+      term_t pi = PL_new_term_ref();
 
-      unify_definition(pred, def, 0, GP_NAMEARITY|GP_HIDESYSTEM);
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_permission_error3,
-		    PL_ATOM, op,
-		    PL_ATOM, type,
-		    PL_TERM, pred);
+      rc = ( PL_unify_predicate(pi, pred, GP_NAMEARITY|GP_HIDESYSTEM) &&
+	     PL_unify_term(formal,
+			   PL_FUNCTOR, FUNCTOR_permission_error3,
+			     PL_ATOM, op,
+			     PL_ATOM, type,
+			     PL_TERM, pi));
       break;
     }
     case ERR_NOT_IMPLEMENTED_PROC:
     { const char *name = va_arg(args, const char *);
       int arity = va_arg(args, int);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_not_implemented2,
-		      PL_ATOM, ATOM_procedure,
-		      PL_FUNCTOR, FUNCTOR_divide2,
-		        PL_CHARS, name,
-		        PL_INT, arity);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_not_implemented2,
+			   PL_ATOM, ATOM_procedure,
+			   PL_FUNCTOR, FUNCTOR_divide2,
+			     PL_CHARS, name,
+			     PL_INT, arity);
       break;
     }
     case ERR_FAILED:
     { term_t goal = va_arg(args, term_t);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_failure_error1,
-		      PL_TERM, goal);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_failure_error1,
+			   PL_TERM, goal);
 
       break;
     }
     case ERR_EVALUATION:
     { atom_t what = va_arg(args, atom_t);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_evaluation_error1,
-		      PL_ATOM, what);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_evaluation_error1,
+			   PL_ATOM, what);
       break;
     }
     case ERR_NOT_EVALUABLE:
     { functor_t f = va_arg(args, functor_t);
       term_t actual = PL_new_term_ref();
 
-      put_name_arity(actual, f);
-      
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_type_error2,
-		      PL_ATOM, ATOM_evaluable,
-		      PL_TERM, actual);
+      rc = (put_name_arity(actual, f) &&
+	    PL_unify_term(formal,
+			  PL_FUNCTOR, FUNCTOR_type_error2,
+			    PL_ATOM, ATOM_evaluable,
+			    PL_TERM, actual));
       break;
     }
     case ERR_DIV_BY_ZERO:
-    { PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_evaluation_error1,
-		      PL_ATOM, ATOM_zero_divisor);
+    { rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_evaluation_error1,
+			   PL_ATOM, ATOM_zero_divisor);
       break;
     }
     case ERR_PERMISSION:
@@ -267,22 +320,22 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       atom_t op   = va_arg(args, atom_t);
       term_t obj  = va_arg(args, term_t);
 
-      PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_permission_error3,
-			  PL_ATOM, type,
-			  PL_ATOM, op,
-			  PL_TERM, obj);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_permission_error3,
+			   PL_ATOM, type,
+			   PL_ATOM, op,
+			   PL_TERM, obj);
 
       break;
     }
     case ERR_OCCURS_CHECK:
     { Word p1  = va_arg(args, Word);
       Word p2  = va_arg(args, Word);
-      
-      PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_occurs_check2,
-			  PL_TERM, wordToTermRef(p1),
-			  PL_TERM, wordToTermRef(p2));
+
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_occurs_check2,
+			   PL_TERM, wordToTermRef(p1),
+			   PL_TERM, wordToTermRef(p2));
 
       break;
     }
@@ -290,10 +343,10 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     { atom_t op   = va_arg(args, atom_t);
       term_t obj  = va_arg(args, term_t);
 
-      PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_timeout_error2,
-			  PL_ATOM, op,
-			  PL_TERM, obj);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_timeout_error2,
+			   PL_ATOM, op,
+			   PL_TERM, obj);
 
       break;
     }
@@ -301,10 +354,10 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     { atom_t type = va_arg(args, atom_t);
       term_t obj  = va_arg(args, term_t);
 
-      PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_existence_error2,
-			  PL_ATOM, type,
-			  PL_TERM, obj);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_existence_error2,
+			   PL_ATOM, type,
+			   PL_TERM, obj);
 
       break;
     }
@@ -314,18 +367,21 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       term_t file   = va_arg(args, term_t);
 
       switch(errno)
-      { case EACCES:
-	  PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_permission_error3,
-			  PL_ATOM, action,
-			  PL_ATOM, type,
-			  PL_TERM, file);
+      { case EAGAIN:
+	  action = ATOM_lock;		/* Hack for file-locking*/
+	  /*FALLTHROUGH*/
+	case EACCES:
+	  rc = PL_unify_term(formal,
+			     PL_FUNCTOR, FUNCTOR_permission_error3,
+			       PL_ATOM, action,
+			       PL_ATOM, type,
+			       PL_TERM, file);
 	  break;
 	case EMFILE:
 	case ENFILE:
-	  PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_resource_error1,
-			  PL_ATOM, ATOM_max_files);
+	  rc = PL_unify_term(formal,
+			     PL_FUNCTOR, FUNCTOR_resource_error1,
+			       PL_ATOM, ATOM_max_files);
 	  break;
 #ifdef EPIPE
 	case EPIPE:
@@ -334,10 +390,10 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
 	  /*FALLTHROUGH*/
 #endif
 	default:			/* what about the other cases? */
-	  PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_existence_error2,
-			  PL_ATOM, type,
-			  PL_TERM, file);
+	  rc = PL_unify_term(formal,
+			     PL_FUNCTOR, FUNCTOR_existence_error2,
+			       PL_ATOM, type,
+			       PL_TERM, file);
 	  break;
       }
 
@@ -346,62 +402,62 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     case ERR_STREAM_OP:
     { atom_t action = va_arg(args, atom_t);
       term_t stream = va_arg(args, term_t);
-      
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_io_error2,
-		      PL_ATOM, action,
-		      PL_TERM, stream);
+
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_io_error2,
+			   PL_ATOM, action,
+			   PL_TERM, stream);
       break;
     }
     case ERR_DDE_OP:
     { const char *op  = va_arg(args, const char *);
       const char *err = va_arg(args, const char *);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_dde_error2,
-		      PL_CHARS, op,
-		      PL_CHARS, err);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_dde_error2,
+			   PL_CHARS, op,
+			   PL_CHARS, err);
       break;
     }
     case ERR_SHARED_OBJECT_OP:
     { atom_t action = va_arg(args, atom_t);
       const char *err = va_arg(args, const char *);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_shared_object2,
-		      PL_ATOM,  action,
-		      PL_CHARS, err);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_shared_object2,
+			   PL_ATOM,  action,
+			   PL_CHARS, err);
       break;
     }
-    case ERR_NOT_IMPLEMENTED_FEATURE:		/* non-ISO */
+    case ERR_NOT_IMPLEMENTED:		/* non-ISO */
     { const char *what = va_arg(args, const char *);
 
-      PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_not_implemented2,
-		          PL_ATOM, ATOM_feature,
-			  PL_CHARS, what);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_not_implemented2,
+			   PL_ATOM, ATOM_feature,
+			   PL_CHARS, what);
       break;
     }
     case ERR_RESOURCE:
     { atom_t what = va_arg(args, atom_t);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_resource_error1,
-		      PL_ATOM, what);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_resource_error1,
+			   PL_ATOM, what);
       break;
     }
     case ERR_SYNTAX:
     { const char *what = va_arg(args, const char *);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_syntax_error1,
-		      PL_CHARS, what);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_syntax_error1,
+			   PL_CHARS, what);
       break;
     }
     case ERR_NOMEM:
-    { PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_resource_error1,
-		      PL_ATOM, ATOM_no_memory);
+    { rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_resource_error1,
+			   PL_ATOM, ATOM_no_memory);
 
       break;
     }
@@ -413,12 +469,12 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
 
       switch(errno)
       { case ENOMEM:
-	  PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_resource_error1,
-			  PL_ATOM, ATOM_no_memory);
+	  rc = PL_unify_term(formal,
+			     PL_FUNCTOR, FUNCTOR_resource_error1,
+			       PL_ATOM, ATOM_no_memory);
 	  break;
 	default:
-	  PL_unify_atom(formal, ATOM_system_error);
+	  rc = PL_unify_atom(formal, ATOM_system_error);
 	  break;
       }
 
@@ -427,40 +483,40 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     case ERR_SHELL_FAILED:
     { term_t cmd = va_arg(args, term_t);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_shell2,
-		      PL_ATOM, ATOM_execute,
-		      PL_TERM, cmd);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_shell2,
+			   PL_ATOM, ATOM_execute,
+			   PL_TERM, cmd);
       break;
     }
     case ERR_SHELL_SIGNALLED:
     { term_t cmd = va_arg(args, term_t);
       int sig = va_arg(args, int);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_shell2,
-		      PL_FUNCTOR, FUNCTOR_signal1,
-		        PL_INT, sig,
-		      PL_TERM, cmd);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_shell2,
+			   PL_FUNCTOR, FUNCTOR_signal1,
+			     PL_INT, sig,
+			 PL_TERM, cmd);
       break;
     }
     case ERR_SIGNALLED:
     { int   sig     = va_arg(args, int);
       char *signame = va_arg(args, char *);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_signal2,
-			PL_CHARS,   signame,
-		        PL_INT, sig);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_signal2,
+			   PL_CHARS, signame,
+			   PL_INT, sig);
       break;
     }
     case ERR_CLOSED_STREAM:
     { IOSTREAM *s = va_arg(args, IOSTREAM *);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_existence_error2,
-		    PL_ATOM, ATOM_stream,
-		    PL_POINTER, s);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_existence_error2,
+			   PL_ATOM, ATOM_stream,
+			   PL_POINTER, s);
       do_throw = TRUE;
       break;
     }
@@ -468,25 +524,25 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     { atom_t type  = va_arg(args, atom_t);
       term_t mutex = va_arg(args, term_t);
 
-      PL_unify_term(formal, PL_FUNCTOR, FUNCTOR_busy2, type, mutex);
+      rc = PL_unify_term(formal, PL_FUNCTOR, FUNCTOR_busy2, type, mutex);
       break;
     }
     case ERR_FORMAT:
     { const char *s = va_arg(args, const char*);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR_CHARS, "format", 1,
-		      PL_CHARS, s);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR_CHARS, "format", 1,
+			   PL_CHARS, s);
       break;
     }
     case ERR_FORMAT_ARG:
     { const char *s = va_arg(args, const char*);
       term_t arg = va_arg(args, term_t);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR_CHARS, "format_argument_type", 2,
-		      PL_CHARS, s,
-		      PL_TERM, arg);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR_CHARS, "format_argument_type", 2,
+			   PL_CHARS, s,
+			   PL_TERM, arg);
       break;
     }
     default:
@@ -495,34 +551,40 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
   va_end(args);
 
 					/* build SWI-Prolog context term */
-  if ( pred || msg || caller )
+  if ( rc && (pred || msg || caller) )
   { term_t predterm = PL_new_term_ref();
     term_t msgterm  = PL_new_term_ref();
 
     if ( pred )
-    { PL_unify_term(predterm,
-		    PL_FUNCTOR, FUNCTOR_divide2,
-		      PL_CHARS, pred,
-		      PL_INT, arity);
+    { rc = PL_unify_term(predterm,
+			 PL_FUNCTOR, FUNCTOR_divide2,
+			   PL_CHARS, pred,
+			   PL_INT, arity);
     } else if ( caller )
-    { unify_definition(predterm, caller, 0, GP_NAMEARITY);
+    { rc = unify_definition(MODULE_user, predterm, caller, 0, GP_NAMEARITY);
     }
 
-    if ( msg )
-    { PL_put_atom_chars(msgterm, msg);
+    if ( rc && msg )
+    { rc = PL_put_atom_chars(msgterm, msg);
     }
 
-    PL_unify_term(swi,
-		  PL_FUNCTOR, FUNCTOR_context2,
-		    PL_TERM, predterm,
-		    PL_TERM, msgterm);
+    if ( rc )
+      rc = PL_unify_term(swi,
+			 PL_FUNCTOR, FUNCTOR_context2,
+			   PL_TERM, predterm,
+			   PL_TERM, msgterm);
   }
 
-  PL_unify_term(except,
-		PL_FUNCTOR, FUNCTOR_error2,
-		  PL_TERM, formal,
-		  PL_TERM, swi);
+  if ( rc )
+    rc = PL_unify_term(except,
+		       PL_FUNCTOR, FUNCTOR_error2,
+		         PL_TERM, formal,
+		         PL_TERM, swi);
 
+  if ( !rc )
+  { nomem:
+    fatalError("Cannot report error: no memory");
+  }
 
   if ( do_throw )
     rc = PL_throw(except);
@@ -542,7 +604,7 @@ tostr(char *buf, const char *fmt, ...)
   va_start(args, fmt);
   Svsprintf(buf, fmt, args);
   va_end(args);
-  
+
   return buf;
 }
 
@@ -559,44 +621,47 @@ PL_unify_term(). This predicate saves possible   pending  exceptions and
 restores them to make the call from B_THROW possible.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-void
+#define OK_RECURSIVE 10
+
+int
 printMessage(atom_t severity, ...)
-{ fid_t fid;
-  term_t ex, av;
+{ GET_LD
+  wakeup_state wstate;
+  term_t av;
   predicate_t pred = PROCEDURE_print_message2;
   va_list args;
+  int rc;
+
+  if ( ++LD->in_print_message >= OK_RECURSIVE*3 )
+    fatalError("printMessage(): recursive call\n");
+  if ( !saveWakeup(&wstate, TRUE PASS_LD) )
+  { LD->in_print_message--;
+    return FALSE;
+  }
 
-  blockGC(PASS_LD1);			/* sometimes called from dangerous */
-					/* places */
-  fid = PL_open_foreign_frame();
   av = PL_new_term_refs(2);
-
-  if ( exception_term )
-  { ex = PL_copy_term_ref(exception_term);
-    exception_term = 0;
-  } else
-    ex = 0;
-
   va_start(args, severity);
   PL_put_atom(av+0, severity);
-  PL_unify_termv(av+1, args);
+  rc = PL_unify_termv(av+1, args);
   va_end(args);
 
-  if ( isDefinedProcedure(pred) )
-    PL_call_predicate(NULL, PL_Q_NODEBUG|PL_Q_CATCH_EXCEPTION, pred, av);
-  else
-  { Sfprintf(Serror, "Message: ");
-    PL_write_term(Serror, av+1, 1200, 0);
-    Sfprintf(Serror, "\n");
+  if ( rc )
+  { if ( isDefinedProcedure(pred) && LD->in_print_message <= OK_RECURSIVE )
+    { rc = PL_call_predicate(NULL, PL_Q_NODEBUG|PL_Q_CATCH_EXCEPTION,
+			     pred, av);
+    } else if ( LD->in_print_message <= OK_RECURSIVE*2 )
+    { Sfprintf(Serror, "Message: ");
+      rc = PL_write_term(Serror, av+1, 1200, 0);
+      Sfprintf(Serror, "\n");
+    } else				/* in_print_message == 2 */
+    { Sfprintf(Serror, "printMessage(): recursive call\n");
+    }
   }
 
-  if ( ex )
-  { PL_put_term(exception_bin, ex);
-    exception_term = exception_bin;
-  }
+  restoreWakeup(&wstate PASS_LD);
+  LD->in_print_message--;
 
-  PL_discard_foreign_frame(fid);
-  unblockGC(PASS_LD1);
+  return rc;
 }
 
 
@@ -617,7 +682,7 @@ PL_get_chars_ex(term_t t, char **s, unsigned int flags)
 
 
 int
-PL_get_atom_ex(term_t t, atom_t *a)
+PL_get_atom_ex__LD(term_t t, atom_t *a ARG_LD)
 { if ( PL_get_atom(t, a) )
     succeed;
 
@@ -627,7 +692,9 @@ PL_get_atom_ex(term_t t, atom_t *a)
 
 int
 PL_get_integer_ex(term_t t, int *i)
-{ if ( PL_get_integer(t, i) )
+{ GET_LD
+
+  if ( PL_get_integer(t, i) )
     succeed;
 
   if ( PL_is_integer(t) )
@@ -639,7 +706,9 @@ PL_get_integer_ex(term_t t, int *i)
 
 int
 PL_get_long_ex(term_t t, long *i)
-{ if ( PL_get_long(t, i) )
+{ GET_LD
+
+  if ( PL_get_long(t, i) )
     succeed;
 
   if ( PL_is_integer(t) )
@@ -651,7 +720,9 @@ PL_get_long_ex(term_t t, long *i)
 
 int
 PL_get_int64_ex(term_t t, int64_t *i)
-{ if ( PL_get_int64(t, i) )
+{ GET_LD
+
+  if ( PL_get_int64(t, i) )
     succeed;
 
   if ( PL_is_integer(t) )
@@ -672,6 +743,28 @@ PL_get_intptr_ex(term_t t, intptr_t *i)
 }
 
 
+int
+PL_get_size_ex(term_t t, size_t *i)
+{ int64_t val;
+
+  if ( !PL_get_int64_ex(t, &val) )
+    fail;
+  if ( val < 0 )
+    return PL_error(NULL, 0, NULL, ERR_DOMAIN,
+		    ATOM_not_less_than_zero, t);
+#if SIZEOF_VOIDP < 8
+#if SIZEOF_LONG == SIZEOF_VOIDP
+  if ( val > (int64_t)ULONG_MAX )
+    return PL_error(NULL, 0, NULL, ERR_REPRESENTATION, ATOM_size_t);
+#endif
+#endif
+
+  *i = (size_t)val;
+
+  return TRUE;
+}
+
+
 int
 PL_get_bool_ex(term_t t, int *i)
 { if ( PL_get_bool(t, i) )
@@ -699,14 +792,26 @@ PL_get_char_ex(term_t t, int *p, int eof)
 }
 
 
+int
+PL_get_pointer_ex(term_t t, void **addrp)
+{ GET_LD
+  if ( PL_get_pointer(t, addrp) )
+    succeed;
+
+  return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_address, t);
+}
+
+
 int
 PL_unify_list_ex(term_t l, term_t h, term_t t)
-{ if ( PL_unify_list(l, h, t) )
+{ GET_LD
+
+  if ( PL_unify_list(l, h, t) )
     succeed;
 
   if ( PL_get_nil(l) )
     fail;
-  
+
   return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_list, l);
 }
 
@@ -725,12 +830,14 @@ PL_unify_nil_ex(term_t l)
 
 int
 PL_get_list_ex(term_t l, term_t h, term_t t)
-{ if ( PL_get_list(l, h, t) )
+{ GET_LD
+
+  if ( PL_get_list(l, h, t) )
     succeed;
 
   if ( PL_get_nil(l) )
     fail;
-  
+
   return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_list, l);
 }
 
@@ -747,7 +854,8 @@ PL_get_nil_ex(term_t l)
 
 int
 PL_unify_bool_ex(term_t t, bool val)
-{ bool v;
+{ GET_LD
+  bool v;
 
   if ( PL_is_variable(t) )
     return PL_unify_atom(t, val ? ATOM_true : ATOM_false);
@@ -763,9 +871,11 @@ PL_unify_bool_ex(term_t t, bool val)
 
 int
 PL_get_arg_ex(int n, term_t term, term_t arg)
-{ if ( PL_get_arg(n, term, arg) )
-    succeed;
-  else
+{ GET_LD
+
+  if ( PL_get_arg(n, term, arg) )
+  { succeed;
+  } else
   { term_t a = PL_new_term_ref();
 
     PL_put_integer(a, n);
