@@ -1,4 +1,4 @@
-/*  $Id: pl-error.h,v 1.13 2004/09/21 14:22:25 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -35,7 +35,7 @@
 #define ERR_FAILED	       10	/* predicate_t proc */
 #define ERR_FILE_OPERATION     11	/* atom_t action, atom_t type, term_t */
 #define ERR_PERMISSION	       12	/* atom_t type, atom_t op, term_t obj*/
-#define ERR_NOT_IMPLEMENTED_FEATURE 13	/* const char *what */
+#define ERR_NOT_IMPLEMENTED 13	/* const char *what */
 #define ERR_EXISTENCE	       14	/* atom_t type, term_t obj */
 #define ERR_STREAM_OP	       15	/* atom_t action, term_t obj */
 #define ERR_RESOURCE	       16	/* atom_t resource */
@@ -58,6 +58,10 @@
 #define ERR_NOT_IMPLEMENTED_PROC 33	/* name, arity */
 #define ERR_FORMAT	       34	/* message */
 #define ERR_FORMAT_ARG	       35	/* seq, term */
+#define ERR_OCCURS_CHECK       36	/* Word, Word */
+#define ERR_CHARS_TYPE	       37	/* char *, term */
+#define ERR_MUST_BE_VAR	       38	/* int argn, term_t term */
+#define ERR_MODIFY_THREAD_LOCAL_PROC 39	/* Procedure proc */
 
 #define MSG_ERRNO		((char *)(-1))
 
