@@ -66,8 +66,9 @@ handling times must be cleaned, but that not only holds for this module.
 #undef LD				/* fetch LD once per function */
 #define LD LOCAL_LD
 
-static int	bad_encoding(atom_t name);
+static int	bad_encoding(const char *msg, atom_t name);
 static int	noprotocol(void);
+static PL_blob_t stream_blob;
 
 const atom_t standardStreams[] =
 { ATOM_user_input,			/* 0 */
@@ -254,9 +255,13 @@ freeStream(IOSTREAM *s)
   if ( (symb=lookupHTable(streamContext, s)) )
   { stream_context *ctx = symb->value;
 
-    if ( ctx->filename == source_file_name )
-    { source_file_name = NULL_ATOM;	/* TBD: pop? */
-      source_line_no = -1;
+    if ( ctx->filename != NULL_ATOM )
+    { PL_unregister_atom(ctx->filename);
+
+      if ( ctx->filename == source_file_name )
+      { source_file_name = NULL_ATOM;	/* TBD: pop? */
+	source_line_no = -1;
+      }
     }
 
     freeHeap(ctx, sizeof(*ctx));
@@ -282,10 +287,18 @@ freeStream(IOSTREAM *s)
 
 
 /* MT: locked by caller (openStream()) */
+/* name must be registered by the caller */
 
 static void
 setFileNameStream(IOSTREAM *s, atom_t name)
-{ getStreamContext(s)->filename = name;
+{ stream_context *ctx = getStreamContext(s);
+
+  if ( ctx->filename )
+  { PL_unregister_atom(ctx->filename);
+    ctx->filename = NULL_ATOM;
+  }
+  if ( name != NULL_ATOM )
+    ctx->filename = name;
 }
 
 
@@ -309,6 +322,7 @@ initIO()
 
   streamAliases = newHTable(16);
   streamContext = newHTable(16);
+  PL_register_blob_type(&stream_blob);
 #ifdef __unix__
 { int fd;
 
@@ -346,6 +360,7 @@ initIO()
   GD->io_initialised = TRUE;
 }
 
+
 		 /*******************************
 		 *	     GET HANDLES	*
 		 *******************************/
@@ -354,10 +369,8 @@ initIO()
 
 static inline IOSTREAM *
 getStream(IOSTREAM *s)
-{ if ( s && s->magic == SIO_MAGIC )	/* TBD: ensure visibility? */
-  { Slock(s);
+{ if ( s && s->magic == SIO_MAGIC && Slock(s) == 0 )
     return s;
-  }
 
   return NULL;
 }
@@ -394,56 +407,166 @@ PL_release_stream(IOSTREAM *s)
 }
 
 
+		 /*******************************
+		 *	      ERRORS		*
+		 *******************************/
+
+static int
+no_stream(term_t t)
+{ return PL_error(NULL, 0, NULL, ERR_EXISTENCE, ATOM_stream, t);
+}
+
+static int
+not_a_stream(term_t t)
+{ return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_stream_or_alias, t);
+}
+
+static int
+symbol_no_stream(atom_t symbol)
+{ GET_LD
+  term_t t = PL_new_term_ref();
+  PL_put_atom(t, symbol);
+  return no_stream(t);
+}
+
+static int
+symbol_not_a_stream(atom_t symbol)
+{ GET_LD
+  term_t t = PL_new_term_ref();
+  PL_put_atom(t, symbol);
+  return not_a_stream(t);
+}
+
+
+
+		 /*******************************
+		 *	  PROLOG HANDLES	*
+		 *******************************/
+
+typedef struct stream_ref
+{ IOSTREAM *read;
+  IOSTREAM *write;
+} stream_ref;
+
+
+static int
+write_stream_ref(IOSTREAM *s, atom_t aref, int flags)
+{ stream_ref *ref = PL_blob_data(aref, NULL, NULL);
+
+  if ( ref->read && ref->write )
+    Sfprintf(s, "<stream>(%p,%p)", ref->read, ref->write);
+  else if ( ref->read )
+    Sfprintf(s, "<stream>(%p)", ref->read);
+  else
+    Sfprintf(s, "<stream>(%p)", ref->write);
+
+  return TRUE;
+}
+
+
+static void
+acquire_stream_ref(atom_t aref)
+{ stream_ref *ref = PL_blob_data(aref, NULL, NULL);
+
+  if ( ref->read )
+    ref->read->references++;
+  if ( ref->write )
+    ref->write->references++;
+}
+
+
+static int
+release_stream_ref(atom_t aref)
+{ stream_ref *ref = PL_blob_data(aref, NULL, NULL);
+
+  if ( ref->read )
+  { if ( --ref->read->references == 0 && ref->read->erased )
+      unallocStream(ref->read);
+  }
+  if ( ref->write )
+  { if ( --ref->write->references == 0 && ref->write->erased )
+      unallocStream(ref->write);
+  }
+
+  return TRUE;
+}
+
+
+static int
+save_stream_ref(atom_t aref, IOSTREAM *fd)
+{ stream_ref *ref = PL_blob_data(aref, NULL, NULL);
+
+  return PL_warning("Cannot save reference to <stream>(%p,%p)",
+		    ref->read, ref->write);
+}
+
+
+static atom_t
+load_stream_ref(IOSTREAM *fd)
+{ return PL_new_atom("<saved-stream-ref>");
+}
+
+
+static PL_blob_t stream_blob =
+{ PL_BLOB_MAGIC,
+  PL_BLOB_UNIQUE,
+  "stream_ref",
+  release_stream_ref,
+  NULL,
+  write_stream_ref,
+  acquire_stream_ref,
+  save_stream_ref,
+  load_stream_ref
+};
+
+
 #define SH_ERRORS   0x01		/* generate errors */
 #define SH_ALIAS    0x02		/* allow alias */
 #define SH_UNLOCKED 0x04		/* don't lock the stream */
-#define SH_SAFE	    0x08		/* Lookup in table */
+#define SH_OUTPUT   0x08		/* We want an output stream */
+#define SH_INPUT    0x10		/* We want an input stream */
 
 static int
-get_stream_handle__LD(term_t t, IOSTREAM **s, int flags ARG_LD)
-{ atom_t alias;
+get_stream_handle__LD(atom_t a, IOSTREAM **sp, int flags ARG_LD)
+{ stream_ref *ref;
+  PL_blob_t *type;
+  IOSTREAM *s;
 
-  if ( PL_is_functor(t, FUNCTOR_dstream1) )
-  { void *p;
-    term_t a = PL_new_term_ref();
+  ref = PL_blob_data(a, NULL, &type);
+  if ( type == &stream_blob )
+  { if ( ref->read )
+    { if ( ref->write && (flags&SH_OUTPUT) )
+	s = ref->write;
+      else
+	s = ref->read;
+    } else
+      s = ref->write;
 
-    _PL_get_arg(1, t, a);
-    if ( PL_get_pointer(a, &p) )
-    { if ( flags & SH_SAFE )
-      { Symbol symb;
+    if ( s->erased )
+       goto noent;
 
-	LOCK();
-	symb = lookupHTable(streamContext, p);
-	UNLOCK();
+    assert(s->magic == SIO_MAGIC);
 
-	if ( !symb )
-	  goto noent;
-      }
-
-      if ( flags & SH_UNLOCKED )
-      { if ( ((IOSTREAM *)p)->magic == SIO_MAGIC )
-	{ *s = p;
-	  return TRUE;
-	}
-	goto noent;
-      }
-
-      if ( (*s = getStream(p)) )
-	return TRUE;
-
-      goto noent;
+    if ( flags & SH_UNLOCKED )
+    { *sp = s;
+      return TRUE;
+    } else if ( (s=getStream(s)) )
+    { *sp = s;
+      return TRUE;
     }
-  } else if ( PL_get_atom(t, &alias) )
+
+    return symbol_no_stream(a);
+  } else
   { Symbol symb;
 
     if ( !(flags & SH_UNLOCKED) )
       LOCK();
-    if ( (symb=lookupHTable(streamAliases, (void *)alias)) )
+    if ( (symb=lookupHTable(streamAliases, (void *)a)) )
     { IOSTREAM *stream;
       uintptr_t n = (uintptr_t)symb->value;
 
       if ( n < 6 )			/* standard stream! */
-      { stream = LD->IO.streams[n];
+      { stream = LD->IO.streams[n];	/* TBD: No need to lock for std-streams */
       } else
 	stream = symb->value;
 
@@ -453,10 +576,10 @@ get_stream_handle__LD(term_t t, IOSTREAM **s, int flags ARG_LD)
       if ( stream )
       { if ( (flags & SH_UNLOCKED) )
 	{ if ( stream->magic == SIO_MAGIC )
-	  { *s = stream;
+	  { *sp = stream;
 	    return TRUE;
 	  }
-	} else if ( (*s = getStream(stream)) )
+	} else if ( (*sp = getStream(stream)) )
 	  return TRUE;
 	goto noent;
       }
@@ -468,24 +591,58 @@ get_stream_handle__LD(term_t t, IOSTREAM **s, int flags ARG_LD)
   }
 
   if ( flags & SH_ERRORS )
-    return PL_error(NULL, 0, NULL, ERR_DOMAIN,
-		    (flags&SH_ALIAS) ? ATOM_stream_or_alias : ATOM_stream, t);
+    symbol_not_a_stream(a);
 
   return FALSE;
 
 noent:
   if ( flags & SH_ERRORS )
-    PL_error(NULL, 0, NULL, ERR_EXISTENCE, ATOM_stream, t);
+    symbol_no_stream(a);
+
   return FALSE;
 }
 
 #define get_stream_handle(t, sp, flags) \
 	get_stream_handle__LD(t, sp, flags PASS_LD)
 
+
+static int
+term_stream_handle(term_t t, IOSTREAM **s, int flags ARG_LD)
+{ atom_t a;
+
+  if ( !PL_get_atom(t, &a) )
+    return not_a_stream(t);
+
+  return get_stream_handle(a, s, SH_ERRORS|SH_ALIAS);
+}
+
+
 int
 PL_get_stream_handle(term_t t, IOSTREAM **s)
 { GET_LD
-  return get_stream_handle(t, s, SH_ERRORS|SH_ALIAS);
+
+  return term_stream_handle(t, s, SH_ERRORS|SH_ALIAS PASS_LD);
+}
+
+
+static int
+unify_stream_ref(term_t t, IOSTREAM *s)
+{ GET_LD
+  stream_ref ref;
+  int rval;
+
+  memset(&ref, 0, sizeof(ref));
+  if ( s->flags & SIO_INPUT )
+    ref.read = s;
+  else
+    ref.write = s;
+
+  rval = PL_unify_blob(t, &ref, sizeof(ref), &stream_blob);
+
+  if ( !rval && !PL_is_variable(t) )
+    return PL_error(NULL, 0, "stream-argument", ERR_MUST_BE_VAR, 0);
+
+  return rval;
 }
 
 
@@ -502,40 +659,25 @@ PL_unify_stream_or_alias(term_t t, IOSTREAM *s)
   LOCK();
   ctx = getStreamContext(s);
   if ( ctx->alias_head )
-  { rval = PL_unify_atom(t, ctx->alias_head->name);
-  } else
-  { term_t a = PL_new_term_ref();
-
-    PL_put_pointer(a, s);
-    PL_cons_functor(a, FUNCTOR_dstream1, a);
-
-    rval = PL_unify(t, a);
-  }
+    rval = PL_unify_atom(t, ctx->alias_head->name);
+  else
+    rval = unify_stream_ref(t, s);
   UNLOCK();
 
+
   return rval;
 }
 
 
 int
 PL_unify_stream(term_t t, IOSTREAM *s)
-{ GET_LD
-  stream_context *ctx;
-  term_t a = PL_new_term_ref();
+{ stream_context *ctx;
 
   LOCK();
   ctx = getStreamContext(s);
   UNLOCK();
 
-  PL_put_pointer(a, s);
-  PL_cons_functor(a, FUNCTOR_dstream1, a);
-
-  if ( PL_unify(t, a) )
-    return TRUE;
-  if ( PL_is_functor(t, FUNCTOR_dstream1) )
-    return FALSE;
-
-  return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_stream, t);
+  return unify_stream_ref(t, s);
 }
 
 
@@ -566,19 +708,24 @@ getOutputStream(term_t t, IOSTREAM **stream)
   IOSTREAM *s;
 
   if ( t == 0 )
-  { *stream = getStream(Scurout);
-    return TRUE;
-  } else if ( PL_get_atom(t, &a) && a == ATOM_user )
-  { *stream = getStream(Suser_output);
-    return TRUE;
-  } else
-  { *stream = NULL;			/* make compiler happy */
+  { if ( (*stream = getStream(Scurout)) )
+      return TRUE;
+    return no_stream(t);
   }
 
-  if ( !PL_get_stream_handle(t, &s) )
+  if ( !PL_get_atom(t, &a) )
+    return not_a_stream(t);
+
+  if ( a == ATOM_user )
+  { if ( (*stream = getStream(Suser_output)) )
+      return TRUE;
+    return no_stream(t);
+  }
+
+  if ( !get_stream_handle(a, &s, SH_ERRORS|SH_ALIAS|SH_OUTPUT) )
     return FALSE;
 
-  if ( !(s->flags &SIO_OUTPUT) )
+  if ( !(s->flags&SIO_OUTPUT) )
   { releaseStream(s);
     return PL_error(NULL, 0, NULL, ERR_PERMISSION,
 		    ATOM_output, ATOM_stream, t);
@@ -595,16 +742,21 @@ getInputStream__LD(term_t t, IOSTREAM **stream ARG_LD)
   IOSTREAM *s;
 
   if ( t == 0 )
-  { *stream = getStream(Scurin);
-    return TRUE;
-  } else if ( PL_get_atom(t, &a) && a == ATOM_user )
-  { *stream = getStream(Suser_input);
-    return TRUE;
-  } else
-  { *stream = NULL;			/* make compiler happy */
+  { if ( (*stream = getStream(Scurin)) )
+      return TRUE;
+    return no_stream(t);
   }
 
-  if ( !get_stream_handle(t, &s, SH_ERRORS|SH_ALIAS) )
+  if ( !PL_get_atom(t, &a) )
+    return not_a_stream(t);
+
+  if ( a == ATOM_user )
+  { if ( (*stream = getStream(Suser_input)) )
+      return TRUE;
+    return no_stream(t);
+  }
+
+  if ( !get_stream_handle(a, &s, SH_ERRORS|SH_ALIAS|SH_INPUT) )
     return FALSE;
 
   if ( !(s->flags &SIO_INPUT) )
@@ -618,6 +770,51 @@ getInputStream__LD(term_t t, IOSTREAM **stream ARG_LD)
 }
 
 
+/** stream_pairs(+Pair, -Read, -Write)
+    stream_pairs(-Pair, +Read, +Write)
+*/
+
+static
+PRED_IMPL("stream_pair", 3, stream_pair, 0)
+{ PRED_LD
+  IOSTREAM *in = NULL, *out = NULL;
+  int rc = FALSE;
+
+  if ( !PL_is_variable(A1) )
+  { stream_ref *ref;
+    atom_t a;
+    PL_blob_t *type;
+
+    if ( !PL_get_atom(A1, &a) ||
+	 !(ref=PL_blob_data(a, NULL, &type)) ||
+	 type != &stream_blob ||
+	 !ref->read ||
+	 !ref->write )
+      return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_stream_pair, A1);
+
+    return ( PL_unify_stream_or_alias(A2, ref->read) &&
+	     PL_unify_stream_or_alias(A3, ref->write) );
+  }
+
+  if ( getInputStream(A2, &in) &&
+       getOutputStream(A3, &out) )
+  { stream_ref ref;
+
+    ref.read = in;
+    ref.write = out;
+
+    rc = PL_unify_blob(A1, &ref, sizeof(ref), &stream_blob);
+  }
+
+  if ( in )
+    releaseStream(in);
+  if ( out )
+    releaseStream(out);
+
+  return rc;
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 In windows GUI applications, the IO-streams  are   not  bound. We do not
 wish to generate an error on the  stream   errors  that may be caused by
@@ -652,14 +849,21 @@ reportStreamError(IOSTREAM *s)
 
     if ( (s->flags & SIO_FERR) )
     { if ( s->exception )
-      { fid_t fid = PL_open_foreign_frame();
-	term_t ex = PL_new_term_ref();
-	PL_recorded(s->exception, ex);
+      { fid_t fid;
+	term_t ex;
+	int rc;
+
+	LD->exception.processing = TRUE;	/* allow using spare stack */
+	if ( !(fid = PL_open_foreign_frame()) )
+	  return FALSE;
+	ex = PL_new_term_ref();
+	rc = PL_recorded(s->exception, ex);
 	PL_erase(s->exception);
 	s->exception = NULL;
-	PL_raise_exception(ex);
+	if ( rc )
+	  rc = PL_raise_exception(ex);
 	PL_close_foreign_frame(fid);
-	return FALSE;
+	return rc;
       }
 
       if ( s->flags & SIO_INPUT )
@@ -749,7 +953,7 @@ dieIO()
 { if ( GD->io_initialised )
   { noprotocol();
     closeFiles(TRUE);
-    PopTty(Sinput, &ttytab);
+    PopTty(Sinput, &ttytab, TRUE);
   }
 }
 
@@ -839,7 +1043,7 @@ protocol(const char *str, size_t n)
 { GET_LD
   IOSTREAM *s;
 
-  if ( LD && (s = getStream(Sprotocol)) )
+  if ( LD && Sprotocol && (s = getStream(Sprotocol)) )
   { while( n-- > 0 )
       Sputcode(*str++&0xff, s);
     Sflush(s);
@@ -939,19 +1143,24 @@ setupOutputRedirect(term_t to, redir_context *ctx, int redir)
   ctx->redirected = redir;
 
   if ( to == 0 )
-  { ctx->stream = getStream(Scurout);
+  { if ( !(ctx->stream = getStream(Scurout)) )
+      return no_stream(to);
     ctx->is_stream = TRUE;
-  } else if ( PL_get_atom(to, &a) && a == ATOM_user )
-  { ctx->stream = getStream(Suser_output);
-    ctx->is_stream = TRUE;
-  } else if ( get_stream_handle(to, &ctx->stream, SH_SAFE) )
-  { if ( !(ctx->stream->flags &SIO_OUTPUT) )
-    { releaseStream(ctx->stream);
-      return PL_error(NULL, 0, NULL, ERR_PERMISSION,
-		      ATOM_output, ATOM_stream, to);
-    }
+  } else if ( PL_get_atom(to, &a) )
+  { if ( a == ATOM_user )
+    { if ( !(ctx->stream = getStream(Suser_output)) )
+	return no_stream(to);
+      ctx->is_stream = TRUE;
+    } else if ( get_stream_handle(a, &ctx->stream, SH_OUTPUT) )
+    { if ( !(ctx->stream->flags &SIO_OUTPUT) )
+      { releaseStream(ctx->stream);
+	return PL_error(NULL, 0, NULL, ERR_PERMISSION,
+			ATOM_output, ATOM_stream, to);
+      }
 
-    ctx->is_stream = TRUE;
+      ctx->is_stream = TRUE;
+    } else
+      return FALSE;
   } else
   { if ( PL_is_functor(to, FUNCTOR_codes2) )
     { ctx->out_format = PL_CODE_LIST;
@@ -1024,11 +1233,11 @@ closeOutputRedirect(redir_context *ctx)
     rval = PL_unify_wchars_diff(out, tail, ctx->out_format,
 				ctx->size/sizeof(wchar_t),
 				(wchar_t*)ctx->data);
-    if ( tail )
+    if ( rval && tail )
       rval = PL_unify(tail, diff);
 
     if ( ctx->data != ctx->buffer )
-      free(ctx->data);
+      Sfree(ctx->data);
   }
 
   return rval;
@@ -1050,7 +1259,7 @@ discardOutputRedirect(redir_context *ctx)
   } else
   { closeStream(ctx->stream);
     if ( ctx->data != ctx->buffer )
-      free(ctx->data);
+      Sfree(ctx->data);
   }
 }
 
@@ -1158,7 +1367,7 @@ getSingleChar(IOSTREAM *stream, int signals)
   if ( c == 4 || c == 26 )		/* should ask the terminal! */
     c = -1;
 
-  PopTty(stream, &buf);
+  PopTty(stream, &buf, TRUE);
   debugstatus.suspendTrace--;
   Sunlock(stream);
 
@@ -1194,7 +1403,7 @@ readLine(IOSTREAM *in, IOSTREAM *out, char *buffer)
       case '\r':
       case EOF:
         *buf++ = EOS;
-        PopTty(in, &tbuf);
+        PopTty(in, &tbuf, TRUE);
 	Sunlock(in);
 	Sunlock(out);
 
@@ -1258,7 +1467,7 @@ noprotocol(void)
 { GET_LD
   IOSTREAM *s;
 
-  if ( (s = getStream(Sprotocol)) )
+  if ( Sprotocol && (s = getStream(Sprotocol)) )
   { TableEnum e;
     Symbol symb;
 
@@ -1426,6 +1635,7 @@ PRED_IMPL("set_stream", 2, set_stream, 0)
 	if ( !PL_get_atom_ex(a, &fn) )
 	  goto error;
 
+	PL_register_atom(fn);
 	LOCK();
 	setFileNameStream(s, fn);
 	UNLOCK();
@@ -1465,7 +1675,7 @@ PRED_IMPL("set_stream", 2, set_stream, 0)
 	if ( !PL_get_atom_ex(a, &val) )
 	  goto error;
 	if ( (enc = atom_to_encoding(val)) == ENC_UNKNOWN )
-	{ bad_encoding(val);
+	{ bad_encoding(NULL, val);
 	  goto error;
 	}
 
@@ -1530,6 +1740,41 @@ error:
   return FALSE;
 }
 
+#ifdef __WINDOWS__			/* defined in pl-nt.c */
+extern int ftruncate(int fileno, int64_t length);
+#define HAVE_FTRUNCATE
+#endif
+
+static
+PRED_IMPL("set_end_of_stream", 1, set_end_of_stream, 0)
+{ IOSTREAM *s;
+  int rc;
+
+  if ( (rc=PL_get_stream_handle(A1, &s)) )
+  {
+#ifdef HAVE_FTRUNCATE
+    int fileno = Sfileno(s);
+
+    if ( fileno >= 0 )
+    { if ( ftruncate(fileno, Stell64(s)) != 0 )
+	rc = PL_error(NULL, 0, MSG_ERRNO, ERR_FILE_OPERATION,
+		      ATOM_set_end_of_stream, ATOM_stream,
+		      A1);
+    } else
+    { rc = PL_error(NULL, 0, "not a file", ERR_PERMISSION,
+		    ATOM_set_end_of_stream);
+    }
+#else
+    rc = notImplemented("set_end_of_stream", 1);
+#endif
+
+    releaseStream(s);
+  }
+
+  return rc;
+}
+
+
 
 		/********************************
 		*          STRING I/O           *
@@ -2186,8 +2431,12 @@ static
 PRED_IMPL("get_single_char", 1, get_single_char, 0)
 { GET_LD
   IOSTREAM *s = getStream(Suser_input);
-  int c = getSingleChar(s, TRUE);
+  int c;
 
+  if ( !s )
+    return symbol_no_stream(ATOM_user_input);
+
+  c = getSingleChar(s, TRUE);
   if ( c == EOF )
   { if ( PL_exception(0) )
     { releaseStream(s);
@@ -2315,9 +2564,13 @@ PRED_IMPL("ttyflush", 0, ttyflush, 0)
 { PRED_LD
   IOSTREAM *s = getStream(Suser_output);
 
-  Sflush(s);
+  if ( s )
+  { Sflush(s);
 
-  return streamStatus(s);
+    return streamStatus(s);
+  }
+
+  return symbol_no_stream(ATOM_user_output);
 }
 
 
@@ -2492,12 +2745,12 @@ encoding_to_atom(IOENC enc)
 
 
 static int
-bad_encoding(atom_t name)
+bad_encoding(const char *msg, atom_t name)
 { GET_LD
   term_t t = PL_new_term_ref();
 
   PL_put_atom(t, name);
-  return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_encoding, t);
+  return PL_error(NULL, 0, msg, ERR_DOMAIN, ATOM_encoding, t);
 }
 
 
@@ -2539,6 +2792,7 @@ static const opt_spec open4_options[] =
   { ATOM_close_on_abort, OPT_BOOL },
   { ATOM_buffer,	 OPT_ATOM },
   { ATOM_lock,		 OPT_ATOM },
+  { ATOM_wait,		 OPT_BOOL },
   { ATOM_encoding,	 OPT_ATOM },
   { ATOM_bom,	 	 OPT_BOOL },
   { NULL_ATOM,	         0 }
@@ -2555,6 +2809,7 @@ openStream(term_t file, term_t mode, term_t options)
   atom_t eof_action     = ATOM_eof_code;
   atom_t buffer         = ATOM_full;
   atom_t lock		= ATOM_none;
+  int	 wait		= TRUE;
   atom_t encoding	= NULL_ATOM;
   int    close_on_abort = TRUE;
   int	 bom		= -1;
@@ -2567,7 +2822,7 @@ openStream(term_t file, term_t mode, term_t options)
   if ( options )
   { if ( !scan_options(options, 0, ATOM_stream_option, open4_options,
 		       &type, &reposition, &alias, &eof_action,
-		       &close_on_abort, &buffer, &lock, &encoding, &bom) )
+		       &close_on_abort, &buffer, &lock, &wait, &encoding, &bom) )
       return FALSE;
   }
 
@@ -2596,10 +2851,23 @@ openStream(term_t file, term_t mode, term_t options)
   if ( encoding != NULL_ATOM )
   { enc = atom_to_encoding(encoding);
     if ( enc == ENC_UNKNOWN )
-    { bad_encoding(encoding);
-
+    { bad_encoding(NULL, encoding);
       return NULL;
     }
+    if ( type == ATOM_binary && enc != ENC_OCTET )
+    { bad_encoding("type(binary) implies encoding(octet)", encoding);
+      return NULL;
+    }
+    switch(enc)				/* explicitely specified: do not */
+    { case ENC_OCTET:			/* switch to Unicode.  For implicit */
+      case ENC_ASCII:			/* and unicode types we must detect */
+      case ENC_ISO_LATIN_1:		/* and skip the BOM */
+      case ENC_WCHAR:
+	bom = FALSE;
+        break;
+      default:
+	;
+    }
   } else if ( type == ATOM_binary )
   { enc = ENC_OCTET;
     bom = FALSE;
@@ -2614,7 +2882,7 @@ openStream(term_t file, term_t mode, term_t options)
 
 					/* LOCK */
   if ( lock != ATOM_none )
-  { *h++ = 'l';
+  { *h++ = (wait ? 'l' : 'L');
     if ( lock == ATOM_read || lock == ATOM_shared )
       *h++ = 'r';
     else if ( lock == ATOM_write || lock == ATOM_exclusive )
@@ -2641,10 +2909,12 @@ openStream(term_t file, term_t mode, term_t options)
   }
 #ifdef HAVE_POPEN
   else if ( PL_is_functor(file, FUNCTOR_pipe1) )
-  { term_t a = PL_new_term_ref();
+  { term_t a;
     char *cmd;
 
-    PL_get_arg(1, file, a);
+    PL_clear_exception();
+    a = PL_new_term_ref();
+    _PL_get_arg(1, file, a);
     if ( !PL_get_chars(a, &cmd, CVT_ATOM|CVT_STRING|REP_FN) )
     { PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_atom, a);
       return NULL;
@@ -2759,13 +3029,15 @@ pl_see(term_t f)
   atom_t a;
   term_t mode;
 
+  if ( !PL_get_atom_ex(f, &a) )
+    return FALSE;
+
   LOCK();
-  if ( get_stream_handle(f, &s, SH_ALIAS|SH_UNLOCKED) )
+  if ( get_stream_handle(a, &s, SH_ALIAS|SH_UNLOCKED) )
   { Scurin = s;
     goto ok;
   }
-
-  if ( PL_get_atom(f, &a) && a == ATOM_user )
+  if ( a == ATOM_user )
   { Scurin = Suser_input;
     goto ok;
   }
@@ -2798,10 +3070,13 @@ pl_seen()
 
   pop_input_context();
 
-  if ( s->flags & SIO_NOFEOF )
+  if ( s && s->flags & SIO_NOFEOF )
     return TRUE;
 
-  return closeStream(s);
+  if ( s )
+    return closeStream(s);
+
+  return symbol_no_stream(ATOM_current_input);
 }
 
 static
@@ -2837,17 +3112,18 @@ do_tell(term_t f, atom_t m)
   atom_t a;
   term_t mode;
 
+  if ( !PL_get_atom_ex(f, &a) )
+    return FALSE;
+
   LOCK();
-  if ( get_stream_handle(f, &s, SH_UNLOCKED) )
+  if ( get_stream_handle(a, &s, SH_UNLOCKED) )
   { Scurout = s;
     goto ok;
   }
-
-  if ( PL_get_atom(f, &a) && a == ATOM_user )
+  if ( a == ATOM_user )
   { Scurout = Suser_output;
     goto ok;
   }
-
   if ( (s = findStreamFromFile(a, IO_TELL)) )
   { Scurout = s;
     goto ok;
@@ -2896,10 +3172,13 @@ PRED_IMPL("told", 0, told, 0)
 
   popOutputContext();
 
-  if ( s->flags & SIO_NOFEOF )
+  if ( s && s->flags & SIO_NOFEOF )
     return TRUE;
 
-  return closeStream(s);
+  if ( s )
+    return closeStream(s);
+
+  return symbol_no_stream(ATOM_current_output);
 }
 
 		 /*******************************
@@ -2959,17 +3238,67 @@ PRED_IMPL("open_null_stream", 1, open_null_stream, 0)
 }
 
 
-static
-PRED_IMPL("close", 1, close, PL_FA_ISO)
-{ IOSTREAM *s;
+static int
+do_close(IOSTREAM *s, int force)
+{ if ( force )
+  { if ( !s )
+      return TRUE;
+    if ( s == Sinput )
+      Sclearerr(s);
+    else if ( s == Soutput || s == Serror )
+    { Sflush(s);
+      Sclearerr(s);
+    } else
+    { Sflush(s);
+      Sclose(s);
+    }
 
-  if ( PL_get_stream_handle(A1, &s) )
-    return closeStream(s);
+    return TRUE;
+  } else if ( s )
+  { return closeStream(s);
+  } else
+  { return FALSE;
+  }
+}
+
+
+static int
+pl_close(term_t stream, int force ARG_LD)
+{ IOSTREAM *s;
+  atom_t a;
+  stream_ref *ref;
+  PL_blob_t *type;
+
+  if ( !PL_get_atom(stream, &a) )
+    return not_a_stream(stream);
+
+  ref = PL_blob_data(a, NULL, &type);
+  if ( type == &stream_blob )
+  { int rc = TRUE;
+
+    if ( ref->read )
+      rc = do_close(getStream(ref->read), force);
+    if ( ref->write )
+      rc = rc && do_close(getStream(ref->write), force);
+
+    return rc;
+  }
+
+  if ( get_stream_handle(a, &s, SH_ERRORS|SH_ALIAS) )
+    return do_close(s, force);
 
   return FALSE;
 }
 
 
+static
+PRED_IMPL("close", 1, close, PL_FA_ISO)
+{ PRED_LD
+
+  return pl_close(A1, FALSE PASS_LD);
+}
+
+
 static const opt_spec close2_options[] =
 { { ATOM_force,		 OPT_BOOL },
   { NULL_ATOM,		 0 }
@@ -2978,28 +3307,13 @@ static const opt_spec close2_options[] =
 
 static
 PRED_IMPL("close", 2, close2, PL_FA_ISO)
-{ IOSTREAM *s;
+{ PRED_LD
   int force = FALSE;
 
   if ( !scan_options(A2, 0, ATOM_close_option, close2_options, &force) )
     return FALSE;
 
-  if ( !PL_get_stream_handle(A1, &s) )
-    return FALSE;
-  if ( !force )
-    return closeStream(s);
-
-  if ( s == Sinput )
-    Sclearerr(s);
-  else if ( s == Soutput || s == Serror )
-  { Sflush(s);
-    Sclearerr(s);
-  } else
-  { Sflush(s);
-    Sclose(s);
-  }
-
-  return TRUE;
+  return pl_close(A1, force PASS_LD);
 }
 
 
@@ -3342,6 +3656,7 @@ PRED_IMPL("stream_property", 2, stream_property,
   prop_enum *pe;
   fid_t fid;
   term_t a1;
+  atom_t a;
 
   term_t stream = A1;
   term_t property = A2;
@@ -3377,8 +3692,11 @@ PRED_IMPL("stream_property", 2, stream_property,
 	break;
       }
 
+      if ( !PL_get_atom(stream, &a) )
+	return not_a_stream(stream);
+
       LOCK();				/* given stream */
-      if ( get_stream_handle(stream, &s, SH_ERRORS|SH_UNLOCKED) )
+      if ( get_stream_handle(a, &s, SH_ERRORS|SH_UNLOCKED) )
       { functor_t f;
 
 	if ( PL_is_variable(property) )	/* generate properties */
@@ -3450,7 +3768,15 @@ PRED_IMPL("stream_property", 2, stream_property,
   }
 
 
-  fid = PL_open_foreign_frame();
+  if ( !(fid = PL_open_foreign_frame()) )
+  { error:
+
+    if ( pe->e )
+      freeTableEnum(pe->e);
+
+    freeHeap(pe, sizeof(*pe));
+    return FALSE;
+  }
 
   for(;;)
   { if ( pe->s )				/* given stream */
@@ -3461,7 +3787,8 @@ PRED_IMPL("stream_property", 2, stream_property,
 	  goto enum_e;
       }
 
-      fid2 = PL_open_foreign_frame();
+      if ( !(fid2 = PL_open_foreign_frame()) )
+	goto error;
       for( ; pe->p->functor ; pe->p++ )
       { if ( PL_unify_functor(property, pe->p->functor) )
 	{ int rval;
@@ -3489,6 +3816,9 @@ PRED_IMPL("stream_property", 2, stream_property,
 	  }
 	}
 
+	if ( exception_term )
+	  goto error;
+
 	if ( pe->fixed_p )
 	  break;
 	PL_rewind_foreign_frame(fid2);
@@ -3509,6 +3839,8 @@ PRED_IMPL("stream_property", 2, stream_property,
 	    pe->p = sprop_list;
 	  break;
 	}
+	if ( exception_term )
+	  goto error;
       }
     }
 
@@ -3527,8 +3859,10 @@ static
 PRED_IMPL("is_stream", 1, is_stream, 0)
 { GET_LD
   IOSTREAM *s;
+  atom_t a;
 
-  if ( get_stream_handle(A1, &s, SH_SAFE) )
+  if ( PL_get_atom(A1, &a) &&
+       get_stream_handle(a, &s, 0) )
   { releaseStream(s);
     return TRUE;
   }
@@ -3590,8 +3924,12 @@ static int
 getRepositionableStream(term_t stream, IOSTREAM **sp)
 { GET_LD
   IOSTREAM *s;
+  atom_t a;
 
-  if ( get_stream_handle(stream, &s, SH_ERRORS) )
+  if ( !PL_get_atom(stream, &a) )
+    return not_a_stream(stream);
+
+  if ( get_stream_handle(a, &s, SH_ERRORS) )
   { if ( !s->position || !s->functions || !s->functions->seek )
     { PL_error(NULL, 0, NULL, ERR_PERMISSION,
 	       ATOM_reposition, ATOM_stream, stream);
@@ -3610,7 +3948,7 @@ getRepositionableStream(term_t stream, IOSTREAM **sp)
 static
 PRED_IMPL("set_stream_position", 2, set_stream_position, PL_FA_ISO)
 { PRED_LD
-  IOSTREAM *s;
+  IOSTREAM *s = NULL;			/* make compiler happy */
   int64_t charno, byteno;
   long linepos, lineno;
   term_t a = PL_new_term_ref();
@@ -4011,13 +4349,22 @@ PRED_IMPL("set_prolog_IO", 3, set_prolog_IO, 0)
 { PRED_LD
   IOSTREAM *in = NULL, *out = NULL, *error = NULL;
   int rval = FALSE;
+  int wrapin = FALSE;
 
-  if ( !PL_get_stream_handle(A1, &in) ||
-       !PL_get_stream_handle(A2, &out) )
+  if ( !term_stream_handle(A1, &in, SH_ERRORS|SH_ALIAS|SH_UNLOCKED PASS_LD) ||
+       !term_stream_handle(A2, &out, SH_ERRORS|SH_ALIAS PASS_LD) )
     goto out;
 
+  wrapin = (LD->IO.streams[0] != in);
+  if ( wrapin )
+  { if ( !(in = getStream(in)) )	/* lock it */
+      goto out;
+  }
+
   if ( PL_compare(A2, A3) == 0 )	/* == */
   { error = getStream(Snew(out->handle, out->flags, out->functions));
+    if ( !error )
+      goto out;
     error->flags &= ~SIO_ABUF;		/* disable buffering */
     error->flags |= SIO_NBUF;
   } else
@@ -4029,20 +4376,22 @@ PRED_IMPL("set_prolog_IO", 3, set_prolog_IO, 0)
   out->flags &= ~SIO_ABUF;		/* output: line buffered */
   out->flags |= SIO_LBUF;
 
-  LD->IO.streams[0] = in;		/* user_input */
   LD->IO.streams[1] = out;		/* user_output */
   LD->IO.streams[2] = error;		/* user_error */
-  LD->IO.streams[3] = in;		/* current_input */
   LD->IO.streams[4] = out;		/* current_output */
 
-  wrapIO(in, Sread_user, NULL);
-  LD->prompt.next = TRUE;
+  if ( wrapin )
+  { LD->IO.streams[3] = in;		/* current_input */
+    LD->IO.streams[0] = in;		/* user_input */
+    wrapIO(in, Sread_user, NULL);
+    LD->prompt.next = TRUE;
+  }
 
   UNLOCK();
   rval = TRUE;
 
 out:
-  if ( in )
+  if ( wrapin && in )
     releaseStream(in);
   if ( out )
     releaseStream(out);
@@ -4220,6 +4569,8 @@ BeginPredDefs(file)
   PRED_DEF("source_location", 2, source_location, 0)
   PRED_DEF("copy_stream_data", 3, copy_stream_data3, 0)
   PRED_DEF("copy_stream_data", 2, copy_stream_data2, 0)
+  PRED_DEF("stream_pair", 3, stream_pair, 0)
+  PRED_DEF("set_end_of_stream", 1, set_end_of_stream, 0)
 
 					/* SWI internal */
   PRED_DEF("$push_input_context", 0, push_input_context, 0)
