@@ -22,12 +22,6 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
-#if __TOS__
-#include <tos.h>
-#define HIDDEN	0x02
-#define SUBDIR 0x10
-#endif
-
 #include "pl-incl.h"
 #include "pl-ctype.h"
 
@@ -122,7 +116,9 @@ static bool	match_pattern(matchcode *, char *);
 
 static inline void
 setMap(matchcode *map, int c)
-{ if ( !trueFeature(FILE_CASE_FEATURE) )
+{ GET_LD
+
+  if ( !truePrologFlag(PLFLAG_FILE_CASE) )
     c = makeLower(c);
 
   map[(c)/8] |= 1 << ((c) % 8);
@@ -224,7 +220,7 @@ compile_pattern(compiled_pattern *Out, char *p, int curl)
 	      return (char *) NULL;
 	    }
 	  }
-	  
+
 	  continue;
 	}
       case ANY:
@@ -244,10 +240,13 @@ compile_pattern(compiled_pattern *Out, char *p, int curl)
 	}
 	/*FALLTHROUGH*/
       default:
-        if ( !trueFeature(FILE_CASE_FEATURE) )
+      { GET_LD
+
+        if ( !truePrologFlag(PLFLAG_FILE_CASE) )
 	  c = makeLower(c);
 	Output(c);
 	continue;
+      }
     }
 
     Output(EXIT);
@@ -277,9 +276,10 @@ match_pattern(matchcode *p, char *str)
 	  s++;
 	  continue;
       case ANYOF:					/* [...] */
-        { matchcode c2 = *s;
+        { GET_LD
+	  matchcode c2 = *s;
 
-	  if ( !trueFeature(FILE_CASE_FEATURE) )
+	  if ( !truePrologFlag(PLFLAG_FILE_CASE) )
 	    c2 = makeLower(c2);
 
 	  if ( p[c2 / 8] & (1 << (c2 % 8)) )
@@ -302,33 +302,39 @@ match_pattern(matchcode *p, char *str)
 	  if ( match_pattern(p+1, (char *)s) )
 	    succeed;
 	  p += *p;
-	  continue;	  
+	  continue;
       default:						/* character */
+      { GET_LD
+
 	  if ( c == *s ||
-	       (!trueFeature(FILE_CASE_FEATURE) && c == makeLower(*s)) )
+	       (!truePrologFlag(PLFLAG_FILE_CASE) && c == makeLower(*s)) )
 	  { s++;
 	    continue;
 	  }
           fail;
+      }
     }
   }
 }
 
 
-word
-pl_wildcard_match(term_t pattern, term_t string)
+/** wildcard_match(+Pattern, +Name) is semidet.
+*/
+
+static
+PRED_IMPL("wildcard_match", 2, wildcard_match, 0)
 { char *p, *s;
   compiled_pattern buf;
 
-  if ( !PL_get_chars_ex(pattern, &p, CVT_ALL) ||
-       !PL_get_chars_ex(string,  &s, CVT_ALL) )
+  if ( !PL_get_chars_ex(A1, &p, CVT_ALL) ||
+       !PL_get_chars_ex(A2,  &s, CVT_ALL) )
     fail;
 
   if ( compilePattern(p, &buf) )
   { return matchPattern(s, &buf);
   }
 
-  return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_pattern, pattern);
+  return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_pattern, A1);
 }
 
 
@@ -431,17 +437,17 @@ expand(const char *pattern, GlobInfo info)
 
     for(;;)
     { int c;
-  
+
       switch( (c=*s++) )
       { case EOS:
 	  if ( s > pat )		/* something left and expanded */
 	  { un_escape(prefix, pat, s);
-	  
+
 	    end = info->end;
 	    for( ; info->start < end; info->start++ )
 	    { char path[MAXPATHLEN];
 	      size_t plen;
-  
+
 	      strcpy(path, expand_entry(info, info->start));
 	      plen = strlen(path);
 	      if ( prefix[0] && plen > 0 && path[plen-1] != '/' )
@@ -470,10 +476,10 @@ expand(const char *pattern, GlobInfo info)
       }
       break;
     }
-    
+
     for( tail=s; *tail && !IS_DIR_SEPARATOR(*tail); tail++ )
       ;
-  
+
 /*  By now, head points to the start of the path holding meta characters,
     while tail points to the tail:
 
@@ -483,13 +489,13 @@ expand(const char *pattern, GlobInfo info)
 */
     un_escape(prefix, pat, head);
     un_escape(patbuf, head, tail);
-  
+
     if ( !compilePattern(patbuf, &cbuf) )		/* syntax error */
       fail;
     dot = (patbuf[0] == '.');			/* do dots as well */
-  
+
     end = info->end;
-  
+
     for(; info->start < end; info->start++)
     { DIR *d;
       struct dirent *e;
@@ -499,7 +505,7 @@ expand(const char *pattern, GlobInfo info)
 
       strcpy(path, current);
       strcat(path, prefix);
-      
+
       if ( (d=opendir(path[0] ? OsPath(path, tmp) : ".")) )
       { size_t plen = strlen(path);
 
@@ -514,7 +520,7 @@ expand(const char *pattern, GlobInfo info)
 	  if ( (dot || e->d_name[0] != '.') &&
 	       matchPattern(e->d_name, &cbuf) )
 	  { char newp[MAXPATHLEN];
-  
+
 	    strcpy(newp, path);
 	    strcpy(&newp[plen], e->d_name);
 /*	    if ( !tail[0] || ExistsDirectory(newp) )
@@ -536,15 +542,16 @@ expand(const char *pattern, GlobInfo info)
 
 static int
 compareBagEntries(const void *a1, const void *a2)
-{ GlobInfo info = LD->glob_info;
+{ GET_LD
+  GlobInfo info = LD->glob_info;
   int i1 = *(int *)a1;
   int i2 = *(int *)a2;
   const char *s1, *s2;
 
   s1 = expand_str(info, i1);
   s2 = expand_str(info, i2);
-  
-  if ( trueFeature(FILE_CASE_FEATURE) )
+
+  if ( truePrologFlag(PLFLAG_FILE_CASE) )
     return mbscoll(s1, s2);
   else
     return mbscasecoll(s1, s2);
@@ -553,28 +560,30 @@ compareBagEntries(const void *a1, const void *a2)
 
 static void
 sort_expand(GlobInfo info)
-{ int *ip = &fetchBuffer(&info->files, info->start, int);
+{ GET_LD
+  int *ip = &fetchBuffer(&info->files, info->start, int);
   int is = info->end - info->start;
-  
+
   LD->glob_info = info;
   qsort(ip, is, sizeof(int), compareBagEntries);
 }
 
 
-word
-pl_expand_file_name(term_t f, term_t list)
-{ char spec[MAXPATHLEN];
+static
+PRED_IMPL("expand_file_name", 2, expand_file_name, 0)
+{ PRED_LD
+  char spec[MAXPATHLEN];
   char *s;
   glob_info info;
-  term_t l    = PL_copy_term_ref(list);
+  term_t l    = PL_copy_term_ref(A2);
   term_t head = PL_new_term_ref();
   int i;
 
-  if ( !PL_get_chars_ex(f, &s, CVT_ALL|REP_FN) )
+  if ( !PL_get_chars_ex(A1, &s, CVT_ALL|REP_FN) )
     fail;
   if ( strlen(s) > sizeof(spec)-1 )
     return PL_error(NULL, 0, "File name too intptr_t",
-		    ERR_DOMAIN, ATOM_pattern, f);
+		    ERR_DOMAIN, ATOM_pattern, A1);
 
   if ( !expandVars(s, spec, sizeof(spec)) )
     fail;
@@ -599,3 +608,52 @@ pl_expand_file_name(term_t f, term_t list)
   free_expand_info(&info);
   succeed;
 }
+
+
+/** directory_files(+Dir, -Files) is det.
+
+Files is a list of atoms that describe the entries in Dir.
+*/
+
+static
+PRED_IMPL("directory_files", 2, directory_files, 0)
+{ PRED_LD
+  char *dname;
+  DIR *dir;
+
+  if ( !PL_get_file_name(A1, &dname, PL_FILE_READ|PL_FILE_OSPATH) )
+    return FALSE;
+
+  if ( (dir=opendir(dname)) )
+  { struct dirent *e;
+    term_t tail = PL_copy_term_ref(A2);
+    term_t head = PL_new_term_ref();
+
+    for(e=readdir(dir); e; e = readdir(dir))
+    { PL_put_variable(head);
+      if ( PL_handle_signals() < 0 ||
+	   !PL_unify_list(tail, head, tail) ||
+	   !PL_unify_chars(head, PL_ATOM|REP_FN, (size_t)-1, e->d_name) )
+      { closedir(dir);
+	return FALSE;
+      }
+    }
+    closedir(dir);
+
+    return PL_unify_nil(tail);
+  }
+
+  return PL_error(NULL, 0, OsError(), ERR_FILE_OPERATION,
+		  ATOM_open, ATOM_directory, A1);
+}
+
+
+		 /*******************************
+		 *      PUBLISH PREDICATES	*
+		 *******************************/
+
+BeginPredDefs(glob)
+  PRED_DEF("expand_file_name", 2, expand_file_name, 0)
+  PRED_DEF("wildcard_match",   2, wildcard_match,   0)
+  PRED_DEF("directory_files",  2, directory_files,  0)
+EndPredDefs
