@@ -45,7 +45,7 @@
 
 
 		 /*******************************
-		 *	    MESSAGE BOX		*
+		 *	       CONSOLE		*
 		 *******************************/
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -73,6 +73,38 @@ hasConsole(void)
 }
 
 
+int
+PL_wait_for_console_input(void *handle)
+{ BOOL rc;
+  HANDLE hConsole = handle;
+
+  for(;;)
+  { rc = MsgWaitForMultipleObjects(1,
+				   &hConsole,
+				   FALSE,	/* wait for either event */
+				   INFINITE,
+				   QS_ALLINPUT);
+
+    if ( rc == WAIT_OBJECT_0+1 )
+    { MSG msg;
+
+      while( PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) )
+      { TranslateMessage(&msg);
+	DispatchMessage(&msg);
+      }
+    } else if ( rc == WAIT_OBJECT_0 )
+    { return TRUE;
+    } else
+    { Sdprintf("MsgWaitForMultipleObjects(): 0x%x\n", rc);
+    }
+  }
+}
+
+
+		 /*******************************
+		 *	    MESSAGE BOX		*
+		 *******************************/
+
 void
 PlMessage(const char *fm, ...)
 { va_list(args);
@@ -191,6 +223,23 @@ Pause(double t)
   }
 }
 
+
+		 /*******************************
+		 *	  SET FILE SIZE		*
+		 *******************************/
+
+int
+ftruncate(int fileno, int64_t length)
+{ errno_t e;
+
+  if ( (e=_chsize_s(fileno, length)) == 0 )
+    return 0;
+
+  errno = e;
+  return -1;
+}
+
+
 		 /*******************************
 		 *	 QUERY CPU TIME		*
 		 *******************************/
@@ -329,7 +378,8 @@ get_showCmd(term_t show, int *cmd)
 
 static int
 win_exec(size_t len, const wchar_t *cmd, UINT show)
-{ STARTUPINFOW startup;
+{ GET_LD
+  STARTUPINFOW startup;
   PROCESS_INFORMATION info;
   int rval;
   wchar_t *wcmd;
@@ -683,7 +733,8 @@ unify_csidl_path(term_t t, int csidl)
 
 static
 PRED_IMPL("win_folder", 2, win_folder, PL_FA_NONDETERMINISTIC)
-{ int n;
+{ GET_LD
+  int n;
 
   switch( CTX_CNTRL )
   { case FRG_FIRST_CALL:
@@ -798,7 +849,8 @@ reg_open_key(const wchar_t *which, int create)
 
 static
 PRED_IMPL("win_registry_get_value", 3, win_registry_get_value, 0)
-{ DWORD type;
+{ GET_LD
+  DWORD type;
   BYTE  data[MAXREGSTRLEN];
   DWORD len = sizeof(data);
   size_t klen, namlen;
@@ -849,7 +901,6 @@ static struct regdef
 { { "localSize",    &GD->defaults.local },
   { "globalSize",   &GD->defaults.global },
   { "trailSize",    &GD->defaults.trail },
-  { "argumentSize", &GD->defaults.argument },
   { NULL,           NULL }
 };
 
@@ -885,11 +936,6 @@ getDefaultsFromRegistry()
   { setStacksFromKey(key);
     RegCloseKey(key);
   }
-
-#ifdef O_CYCLIC
-  if ( GD->defaults.argument == DEFARGUMENT )
-    GD->defaults.argument = GD->defaults.global/4;
-#endif
 }
 
 		 /*******************************
