@@ -37,6 +37,17 @@
 #include <math.h>		/* avoid abs() problem with msvc++ */
 #include <stdio.h>		/* rename() and remove() prototypes */
 
+#if TIME_WITH_SYS_TIME
+# include <sys/time.h>
+# include <time.h>
+#else
+# if HAVE_SYS_TIME_H
+#  include <sys/time.h>
+# else
+#  include <time.h>
+# endif
+#endif
+
 #if HAVE_SYS_STAT_H
 #include <sys/stat.h>
 #endif
@@ -99,13 +110,9 @@ long clock_wait_ticks;
 This module is a contraction of functions that used to be all  over  the
 place.   together  with  pl-os.h  (included  by  pl-incl.h) this file
 should define a basic  layer  around  the  OS,  on  which  the  rest  of
-SWI-Prolog  is  based.   SWI-Prolog  has  been developed on SUN, running
-SunOs 3.4 and later 4.0.
+SWI-Prolog  is  based.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-Unfortunately some OS's simply do not offer  an  equivalent  to  SUN  os
-features.   In  most  cases part of the functionality of the system will
-have to be dropped. See the header of pl-incl.h for details.
-- - - - - - - - - - -  - - - - - */
 
 		/********************************
 		*         INITIALISATION        *
@@ -119,7 +126,11 @@ have to be dropped. See the header of pl-incl.h for details.
 
 bool
 initOs(void)
-{ DEBUG(1, Sdprintf("OS:initExpand() ...\n"));
+{ GET_LD
+
+  GD->statistics.start_time = WallTime();
+
+  DEBUG(1, Sdprintf("OS:initExpand() ...\n"));
   initExpand();
   DEBUG(1, Sdprintf("OS:initEnviron() ...\n"));
   initEnviron();
@@ -411,7 +422,8 @@ setOSPrologFlags(void)
 
 uintptr_t
 UsedMemory(void)
-{
+{ GET_LD
+
 #if defined(HAVE_GETRUSAGE) && defined(HAVE_RU_IDRSS)
   struct rusage usage;
 
@@ -498,7 +510,9 @@ setRandom(unsigned int *seedp)
 
 uint64_t
 _PL_Random(void)
-{ if ( !LD->os.rand_initialised )
+{ GET_LD
+
+  if ( !LD->os.rand_initialised )
   { setRandom(NULL);
     LD->os.rand_initialised = TRUE;
   }
@@ -539,7 +553,7 @@ available to the Prolog user based on these functions.  These  functions
 are  in  this  module as non-UNIX OS probably don't have getpid() or put
 temporaries on /tmp.
 
-    atom_t TemporaryFile(const char *id)
+    atom_t TemporaryFile(const char *id, int *fdp)
 
     The return value of this call is an atom,  whose  string  represents
     the  path  name of a unique file that can be used as temporary file.
@@ -553,14 +567,6 @@ temporaries on /tmp.
     not be created at all, or might already have been deleted.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-struct tempfile
-{ atom_t	name;
-  TempFile	next;
-};					/* chain of temporary files */
-
-#define tmpfile_head (GD->os._tmpfile_head)
-#define tmpfile_tail (GD->os._tmpfile_tail)
-
 #ifndef DEFTMPDIR
 #ifdef __WINDOWS__
 #define DEFTMPDIR "c:/tmp"
@@ -569,22 +575,64 @@ struct tempfile
 #endif
 #endif
 
+static int
+free_tmp_symbol(Symbol s)
+{ int rc;
+  atom_t tname = (atom_t)s->name;
+  PL_chars_t txt;
+
+  get_atom_text(tname, &txt);
+  PL_mb_text(&txt, REP_FN);
+  rc = RemoveFile(txt.text.t);
+  PL_free_text(&txt);
+
+  PL_unregister_atom(tname);
+  return rc;
+}
+
+
+static void
+void_free_tmp_symbol(Symbol s)
+{ (void)free_tmp_symbol(s);
+}
+
+
+#ifndef O_EXCL
+#define O_EXCL 0
+#endif
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+
 atom_t
-TemporaryFile(const char *id)
+TemporaryFile(const char *id, int *fdp)
 { char temp[MAXPATHLEN];
-  TempFile tf = allocHeap(sizeof(struct tempfile));
-  char envbuf[MAXPATHLEN];
-  char *tmpdir;
+  static char *tmpdir = NULL;
+  atom_t tname;
+  int retries = 0;
 
-  if ( !((tmpdir = Getenv("TEMP", envbuf, sizeof(envbuf))) ||
-	 (tmpdir = Getenv("TMP",  envbuf, sizeof(envbuf)))) )
-    tmpdir = DEFTMPDIR;
+  if ( !tmpdir )
+  { LOCK();
+    if ( !tmpdir )
+    { char envbuf[MAXPATHLEN];
+      char *td;
 
+      if ( (td = Getenv("TEMP", envbuf, sizeof(envbuf))) ||
+	   (td = Getenv("TMP",  envbuf, sizeof(envbuf))) )
+	tmpdir = strdup(td);
+      else
+	tmpdir = DEFTMPDIR;
+    }
+    UNLOCK();
+  }
+
+retry:
 #ifdef __unix__
 { static int MTOK_temp_counter = 0;
+  const char *sep = id[0] ? "_" : "";
 
-  Ssprintf(temp, "%s/pl_%s_%d_%d",
-	   tmpdir, id, (int) getpid(), MTOK_temp_counter++);
+  Ssprintf(temp, "%s/pl_%s%s%d_%d",
+	   tmpdir, id, sep, (int) getpid(), MTOK_temp_counter++);
 }
 #endif
 
@@ -599,80 +647,77 @@ TemporaryFile(const char *id)
 #endif
   { PrologPath(tmp, temp, sizeof(temp));
   } else
-    Ssprintf(temp, "%s/pl_%s_%d", tmpdir, id, temp_counter++);
-}
-#endif
+  { const char *sep = id[0] ? "_" : "";
 
-#if EMX
-  static int temp_counter = 0;
-  char *foo;
-
-  if ( (foo = tempnam(".", (const char *)id)) )
-  { strcpy(temp, foo);
-    free(foo);
-  } else
-    Ssprintf(temp, "pl_%s_%d_%d", id, getpid(), temp_counter++);
-#endif
-
-  tf->name = PL_new_atom(temp);		/* locked: ok! */
-  tf->next = NULL;
-
-  startCritical;
-  if ( !tmpfile_tail )
-  { tmpfile_head = tmpfile_tail = tf;
-  } else
-  { tmpfile_tail->next = tf;
-    tmpfile_tail = tf;
+    Ssprintf(temp, "%s/pl_%s%s%d", tmpdir, id, sep, temp_counter++);
   }
-  endCritical;
-
-  return tf->name;
 }
+#endif
+
+  if ( fdp )
+  { int fd;
+
+    if ( (fd=open(temp, O_CREAT|O_EXCL|O_WRONLY|O_BINARY, 0600)) < 0 )
+    { if ( ++retries < 10000 )
+	goto retry;
+      else
+	return NULL_ATOM;
+    }
+
+    *fdp = fd;
+  }
+
+  tname = PL_new_atom(temp);		/* locked: ok! */
+
+  LOCK();
+  if ( !GD->os.tmp_files )
+  { GD->os.tmp_files = newHTable(4);
+    GD->os.tmp_files->free_symbol = void_free_tmp_symbol;
+  }
+  UNLOCK();
+
+  addHTable(GD->os.tmp_files, (void*)tname, (void*)TRUE);
+
+  return tname;
+}
+
+
+int
+DeleteTemporaryFile(atom_t name)
+{ int rc = FALSE;
+
+  if ( GD->os.tmp_files )
+  { LOCK();
+    if ( GD->os.tmp_files && GD->os.tmp_files->size > 0 )
+    { Symbol s = lookupHTable(GD->os.tmp_files, (void*)name);
+
+      if ( s )
+      { rc = free_tmp_symbol(s);
+	deleteSymbolHTable(GD->os.tmp_files, s);
+      }
+    }
+    UNLOCK();
+  }
+
+  return rc;
+}
+
 
 void
 RemoveTemporaryFiles(void)
-{ TempFile tf, tf2;
+{ LOCK();
+  if ( GD->os.tmp_files )
+  { Table t = GD->os.tmp_files;
 
-  startCritical;
-  for(tf = tmpfile_head; tf; tf = tf2)
-  { RemoveFile(stringAtom(tf->name));
-    tf2 = tf->next;
-    freeHeap(tf, sizeof(struct tempfile));
+    GD->os.tmp_files = NULL;
+    UNLOCK();
+    destroyHTable(t);
+  } else
+  { UNLOCK();
   }
-
-  tmpfile_head = tmpfile_tail = NULL;
-  endCritical;
 }
 
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Size of a VM page of memory.  Most BSD machines have this function.  If not,
-here are several alternatives ...
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-#ifndef HAVE_GETPAGESIZE
-#ifdef _SC_PAGESIZE
-int
-getpagesize()
-{ return sysconf(_SC_PAGESIZE);
-}
-#else /*_SC_PAGESIZE*/
-
-#if hpux
-#include <a.out.h>
-int
-getpagesize()
-{
-#ifdef EXEC_PAGESIZE
-  return EXEC_PAGESIZE;
-#else
-  return 4096;				/* not that important */
-#endif
-}
-#endif /*hpux*/
-#endif /*_SC_PAGESIZE*/
-#endif /*HAVE_GETPAGESIZE*/
-
 #if O_HPFS
 
 /*  Conversion rules Prolog <-> OS/2 (using HPFS)
@@ -743,7 +788,8 @@ OsPath(const char *p, char *buf)
 #if O_XOS
 char *
 PrologPath(const char *p, char *buf, size_t len)
-{ int flags = (truePrologFlag(PLFLAG_FILE_CASE) ? 0 : XOS_DOWNCASE);
+{ GET_LD
+  int flags = (truePrologFlag(PLFLAG_FILE_CASE) ? 0 : XOS_DOWNCASE);
 
   return _xos_canonical_filename(p, buf, len, flags);
 }
@@ -800,7 +846,7 @@ forwards char   *canoniseDir(char *);
 
 static void
 initExpand(void)
-{
+{ GET_LD
 #ifdef O_CANONISE_DIRS
   char *dir;
   char *cpaths;
@@ -910,6 +956,7 @@ verify_entry(CanonicalDir d)
 
     d->inode  = buf.st_ino;
     d->device = buf.st_dev;
+    return TRUE;
   } else
   { DEBUG(1, Sdprintf("%s: no longer exists\n", d->canonical));
 
@@ -926,6 +973,9 @@ verify_entry(CanonicalDir d)
       }
     }
 
+    remove_string(d->name);
+    if ( d->canonical != d->name )
+      remove_string(d->canonical);
     free(d);
   }
 
@@ -1126,7 +1176,9 @@ utf8_strlwr(char *s)
 
 char *
 canonisePath(char *path)
-{ if ( !truePrologFlag(PLFLAG_FILE_CASE) )
+{ GET_LD
+
+  if ( !truePrologFlag(PLFLAG_FILE_CASE) )
     utf8_strlwr(path);
 
   canoniseFileName(path);
@@ -1173,7 +1225,8 @@ takeWord(const char **string, char *wrd, int maxlen)
 
 bool
 expandVars(const char *pattern, char *expanded, int maxlen)
-{ int size = 0;
+{ GET_LD
+  int size = 0;
   char wordbuf[MAXPATHLEN];
 
   if ( *pattern == '~' )
@@ -1325,7 +1378,8 @@ ExpandFile(const char *pattern, char **vector)
 
 char *
 ExpandOneFile(const char *spec, char *file)
-{ char *vector[256];
+{ GET_LD
+  char *vector[256];
   int size;
 
   switch( (size=ExpandFile(spec, vector)) )
@@ -1424,10 +1478,13 @@ IsAbsolutePath(const char *p)
 
 char *
 AbsoluteFile(const char *spec, char *path)
-{ char tmp[MAXPATHLEN];
+{ GET_LD
+  char tmp[MAXPATHLEN];
   char buf[MAXPATHLEN];
   char *file = PrologPath(spec, buf, sizeof(buf));
 
+  if ( !file )
+     return (char *) NULL;
   if ( truePrologFlag(PLFLAG_FILEVARS) )
   { if ( !(file = ExpandOneFile(buf, tmp)) )
       return (char *) NULL;
@@ -1472,7 +1529,9 @@ AbsoluteFile(const char *spec, char *path)
 
 void
 PL_changed_cwd(void)
-{ if ( CWDdir )
+{ GET_LD
+
+  if ( CWDdir )
     remove_string(CWDdir);
   CWDdir = NULL;
   CWDlen = 0;
@@ -1481,7 +1540,9 @@ PL_changed_cwd(void)
 
 const char *
 PL_cwd(void)
-{ if ( CWDlen == 0 )
+{ GET_LD
+
+  if ( CWDlen == 0 )
   { char buf[MAXPATHLEN];
     char *rval;
 
@@ -1570,7 +1631,8 @@ DirName(const char *f, char *dir)
 
 bool
 ChDir(const char *path)
-{ char ospath[MAXPATHLEN];
+{ GET_LD
+  char ospath[MAXPATHLEN];
   char tmp[MAXPATHLEN];
 
   OsPath(path, ospath);
@@ -1644,6 +1706,40 @@ LocalTime(long *t, struct tm *r)
 }
 
 
+		 /*******************************
+		 *	      TERMINAL		*
+		 *******************************/
+
+#ifdef HAVE_TCSETATTR
+#include <termios.h>
+#include <unistd.h>
+#define O_HAVE_TERMIO 1
+#else /*HAVE_TCSETATTR*/
+#ifdef HAVE_SYS_TERMIO_H
+#include <sys/termio.h>
+#define termios termio
+#define O_HAVE_TERMIO 1
+#else
+#ifdef HAVE_SYS_TERMIOS_H
+#include <sys/termios.h>
+#define O_HAVE_TERMIO 1
+#endif
+#endif
+#endif /*HAVE_TCSETATTR*/
+
+typedef struct tty_state
+{
+#if defined(O_HAVE_TERMIO)
+  struct termios tab;
+#elif defined(HAVE_SGTTYB)
+  struct sgttyb tab;
+#else
+  int tab;				/* empty is not allowed */
+#endif
+} tty_state;
+
+#define TTY_STATE(buf) (((tty_state*)(buf->state))->tab)
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 			TERMINAL IO MANIPULATION
 
@@ -1668,14 +1764,17 @@ ResetStdin()
 
 static ssize_t
 Sread_terminal(void *handle, char *buf, size_t size)
-{ intptr_t h = (intptr_t)handle;
+{ GET_LD
+  intptr_t h = (intptr_t)handle;
   int fd = (int)h;
   source_location oldsrc = LD->read_source;
 
-  if ( LD->prompt.next && ttymode != TTY_RAW )
-    PL_write_prompt(TRUE);
-  else
-    Sflush(Suser_output);
+  if ( Soutput && true(Soutput, SIO_ISATTY) )
+  { if ( LD->prompt.next && ttymode != TTY_RAW )
+      PL_write_prompt(TRUE);
+    else
+      Sflush(Suser_output);
+  }
 
   PL_dispatch(fd, PL_DISPATCH_WAIT);
   size = (*GD->os.org_terminal.read)(handle, buf, size);
@@ -1694,8 +1793,9 @@ Sread_terminal(void *handle, char *buf, size_t size)
 }
 
 void
-ResetTty()
-{ startCritical;
+ResetTty(void)
+{ GET_LD
+  startCritical;
   ResetStdin();
 
   if ( !GD->os.iofunctions.read )
@@ -1723,32 +1823,36 @@ ResetTty()
 
 bool
 PushTty(IOSTREAM *s, ttybuf *buf, int mode)
-{ struct termios tio;
+{ GET_LD
+  struct termios tio;
   int fd;
 
-  buf->mode = ttymode;
-  ttymode = mode;
+  buf->mode  = ttymode;
+  buf->state = NULL;
+  ttymode    = mode;
 
   if ( (fd = Sfileno(s)) < 0 || !isatty(fd) )
     succeed;				/* not a terminal */
   if ( !truePrologFlag(PLFLAG_TTY_CONTROL) )
     succeed;
 
+  buf->state = allocHeap(sizeof(tty_state));
+
 #ifdef HAVE_TCSETATTR
-  if ( tcgetattr(fd, &buf->tab) )	/* save the old one */
+  if ( tcgetattr(fd, &TTY_STATE(buf)) )	/* save the old one */
     fail;
 #else
-  if ( ioctl(fd, TIOCGETA, &buf->tab) )	/* save the old one */
+  if ( ioctl(fd, TIOCGETA, &TTY_STATE(buf)) )	/* save the old one */
     fail;
 #endif
 
-  tio = buf->tab;
+  tio = TTY_STATE(buf);
 
   switch( mode )
   { case TTY_RAW:
 #if defined(HAVE_TCSETATTR) && defined(HAVE_CFMAKERAW)
 	cfmakeraw(&tio);
-	tio.c_oflag = buf->tab.c_oflag;	/* donot change output modes */
+	tio.c_oflag = TTY_STATE(buf).c_oflag;	/* donot change output modes */
 	tio.c_lflag |= ISIG;
 #else
 	tio.c_lflag &= ~(ECHO|ICANON);
@@ -1789,25 +1893,33 @@ PushTty(IOSTREAM *s, ttybuf *buf, int mode)
 
 
 bool
-PopTty(IOSTREAM *s, ttybuf *buf)
-{ int fd;
+PopTty(IOSTREAM *s, ttybuf *buf, int do_free)
+{ GET_LD
+
   ttymode = buf->mode;
 
-  if ( (fd = Sfileno(s)) < 0 || !isatty(fd) )
-    succeed;				/* not a terminal */
-  if ( !truePrologFlag(PLFLAG_TTY_CONTROL) )
-    succeed;
+  if ( buf->state )
+  { int fd = Sfileno(s);
 
+    if ( fd >= 0 )
+    {
 #ifdef HAVE_TCSETATTR
-  tcsetattr(fd, TCSANOW, &buf->tab);
+      tcsetattr(fd, TCSANOW, &TTY_STATE(buf));
 #else
 #ifdef TIOCSETA
-  ioctl(fd, TIOCSETA, &buf->tab);
+      ioctl(fd, TIOCSETA, &TTY_STATE(buf));
 #else
-  ioctl(fd, TCSETA, &buf->tab);
-  ioctl(fd, TCXONC, (void *)1);
+      ioctl(fd, TCSETA, &TTY_STATE(buf));
+      ioctl(fd, TCXONC, (void *)1);
 #endif
 #endif
+    }
+
+    if ( do_free )
+    { freeHeap(buf->state, sizeof(tty_state));
+      buf->state = NULL;
+    }
+  }
 
   succeed;
 }
@@ -1822,6 +1934,7 @@ PushTty(IOSTREAM *s, ttybuf *buf, int mode)
   int fd;
 
   buf->mode = ttymode;
+  buf->state = NULL;
   ttymode = mode;
 
   if ( (fd = Sfileno(s)) < 0 || !isatty(fd) )
@@ -1829,25 +1942,26 @@ PushTty(IOSTREAM *s, ttybuf *buf, int mode)
   if ( !truePrologFlag(PLFLAG_TTY_CONTROL) )
     succeed;
 
-  if ( ioctl(fd, TIOCGETP, &buf->tab) )  /* save the old one */
+  buf->state = allocHeap(sizeof(tty_state));
+
+  if ( ioctl(fd, TIOCGETP, &TTY_STATE(buf)) )  /* save the old one */
     fail;
-  tio = buf->tab;
+  tio = TTY_STATE(buf);
 
   switch( mode )
-    { case TTY_RAW:
-	tio.sg_flags |= CBREAK;
-	tio.sg_flags &= ~ECHO;
-	break;
-      case TTY_OUTPUT:
-	tio.sg_flags |= (CRMOD);
-	break;
-      case TTY_SAVE:
-	succeed;
-      default:
-	sysError("Unknown PushTty() mode: %d", mode);
-	/*NOTREACHED*/
-      }
-
+  { case TTY_RAW:
+      tio.sg_flags |= CBREAK;
+      tio.sg_flags &= ~ECHO;
+      break;
+    case TTY_OUTPUT:
+      tio.sg_flags |= (CRMOD);
+      break;
+    case TTY_SAVE:
+      succeed;
+    default:
+      sysError("Unknown PushTty() mode: %d", mode);
+      /*NOTREACHED*/
+  }
 
   ioctl(fd, TIOCSETP,  &tio);
   ioctl(fd, TIOCSTART, NULL);
@@ -1857,17 +1971,22 @@ PushTty(IOSTREAM *s, ttybuf *buf, int mode)
 
 
 bool
-PopTty(IOSTREAM *s, ttybuf *buf)
+PopTty(IOSTREAM *s, ttybuf *buf, int do_free)
 { ttymode = buf->mode;
-  int fd;
 
-  if ( (fd = Sfileno(s)) < 0 || !isatty(fd) )
-    succeed;				/* not a terminal */
-  if ( !truePrologFlag(PLFLAG_TTY_CONTROL) )
-    succeed;
+  if ( buf->state )
+  { int fd = Sfileno(s);
 
-  ioctl(fd, TIOCSETP,  &buf->tab);
-  ioctl(fd, TIOCSTART, NULL);
+    if ( fd >= 0 )
+    { ioctl(fd, TIOCSETP,  &buf->tab);
+      ioctl(fd, TIOCSTART, NULL);
+    }
+
+    if ( do_free )
+    { freeHeap(buf->state, sizeof(tty_state));
+      buf->state = NULL;
+    }
+  }
 
   succeed;
 }
@@ -1884,8 +2003,9 @@ PushTty(IOSTREAM *s, ttybuf *buf, int mode)
 
 
 bool
-PopTty(IOSTREAM *s, ttybuf *buf)
-{ ttymode = buf->mode;
+PopTty(IOSTREAM *s, ttybuf *buf, int do_free)
+{ GET_LD
+  ttymode = buf->mode;
   if ( ttymode != TTY_RAW )
     LD->prompt.next = TRUE;
 
@@ -2191,7 +2311,8 @@ argument to wait()
 
 int
 System(char *cmd)
-{ int pid;
+{ GET_LD
+  int pid;
   char *shell = "/bin/sh";
   int rval;
   void (*old_int)();
