@@ -301,6 +301,11 @@ int
 Slock(IOSTREAM *s)
 { SLOCK(s);
 
+  if ( s->erased )
+  { SUNLOCK(s);
+    return -1;
+  }
+
 #ifdef DEBUG_IO_LOCKS
   if ( s->locks > 2 )
   { printf("  Lock [%d]: %s: %d locks", PL_thread_self(), Sname(s), s->locks+1);
@@ -322,6 +327,11 @@ StryLock(IOSTREAM *s)
 { if ( !STRYLOCK(s) )
     return -1;
 
+  if ( s->erased )
+  { SUNLOCK(s);
+    return -1;
+  }
+
   if ( !s->locks++ )
   { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) )
       return S__setbuf(s, NULL, TMPBUFSIZE) == (size_t)-1 ? -1 : 0;
@@ -1700,6 +1710,22 @@ Stell(IOSTREAM *s)
 		 *	      CLOSE		*
 		 *******************************/
 
+void
+unallocStream(IOSTREAM *s)
+{
+#ifdef O_PLMT
+  if ( s->mutex )
+  { recursiveMutexDelete(s->mutex);
+    free(s->mutex);
+    s->mutex = NULL;
+  }
+#endif
+
+  if ( !(s->flags & SIO_STATIC) )
+    free(s);
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 (*) Sclose() can be called recursively. For example if an XPCE object is
 only referenced from an open stream,  the close-function will delete the
@@ -1757,19 +1783,13 @@ Sclose(IOSTREAM *s)
   run_close_hooks(s);			/* deletes Prolog registration */
   SUNLOCK(s);
 
-#ifdef O_PLMT
-  if ( s->mutex )
-  { recursiveMutexDelete(s->mutex);
-    free(s->mutex);
-    s->mutex = NULL;
-  }
-#endif
-
   s->magic = SIO_CMAGIC;
   if ( s->message )
     free(s->message);
-  if ( !(s->flags & SIO_STATIC) )
-    free(s);
+  if ( s->references == 0 )
+    unallocStream(s);
+  else
+    s->erased = TRUE;
 
   return rval;
 }
@@ -1887,11 +1907,34 @@ Svprintf(const char *fm, va_list args)
 #define A_LEFT	0			/* left-aligned field */
 #define A_RIGHT 1			/* right-aligned field */
 
+#define SNPRINTF3(fm, a1) \
+	{ size_t __r; \
+	  assert(fs == fbuf); \
+	  __r = snprintf(fs, sizeof(fbuf), fm, a1); \
+	  if ( __r >= sizeof(fbuf) ) \
+	  { if ( (fs_malloced = fs = malloc(__r+1)) == NULL ) goto error; \
+	    __r = snprintf(fs, __r+1, fm, a1); \
+	  } \
+	  fe = fs+__r; \
+	}
+#define SNPRINTF4(fm, a1, a2) \
+	{ size_t __r; \
+	  assert(fs == fbuf); \
+	  __r = snprintf(fs, sizeof(fbuf), fm, a1, a2); \
+	  if ( __r >= sizeof(fbuf) ) \
+	  { if ( (fs_malloced = fs = malloc(__r+1)) == NULL ) goto error; \
+	    __r = snprintf(fs, __r+1, fm, a1, a2); \
+	  } \
+	  fe = fs+__r; \
+	}
+
+
 int
 Svfprintf(IOSTREAM *s, const char *fm, va_list args)
-{ intptr_t printed = 0;
+{ int printed = 0;
   char buf[TMPBUFSIZE];
   int tmpbuf;
+  char *fs_malloced = NULL;
 
   SLOCK(s);
 
@@ -1979,8 +2022,7 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	      *fp++ = '#';
 	    *fp++ = 'p';
 	    *fp   = '\0';
-	    sprintf(fs, fmbuf, ptr);
-	    fe = &fs[strlen(fs)];
+	    SNPRINTF3(fmbuf, ptr);
 
 	    break;
 	  }
@@ -2013,7 +2055,7 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	    if ( islong < 2 )
 	    { *fp++ = *fm;
 	      *fp   = '\0';
-	      sprintf(fs, fmbuf, v);
+	      SNPRINTF3(fmbuf, v);
 	    } else
 	    {
 #ifdef __WINDOWS__
@@ -2024,9 +2066,8 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 #endif
 	      *fp++ = *fm;
 	      *fp   = '\0';
-	      sprintf(fs, fmbuf, vl);
+	      SNPRINTF3(fmbuf, vl);
 	    }
-	    fe = &fs[strlen(fs)];
 
 	    break;
 	  }
@@ -2041,16 +2082,16 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	    *fp++ = '%';
 	    if ( modified )
 	      *fp++ = '#';
-	    if ( has_arg2 )		/* specified percission */
+	    if ( has_arg2 )		/* specified precission */
 	    { *fp++ = '.';
 	      *fp++ = '*';
 	      *fp++ = *fm;
 	      *fp   = '\0';
-	      sprintf(fs, fmbuf, arg2, v);
+	      SNPRINTF4(fmbuf, arg2, v);
 	    } else
 	    { *fp++ = *fm;
 	      *fp   = '\0';
-	      sprintf(fs, fmbuf, v);
+	      SNPRINTF3(fmbuf, v);
 	    }
 	    fe = &fs[strlen(fs)];
 
@@ -2116,6 +2157,10 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	  }
 	}
 	fm++;
+	if ( fs_malloced )
+	{ fs_malloced = NULL;
+	  free(fs_malloced);
+	}
       }
     } else if ( *fm == '\\' && fm[1] )
     { OUTCHR(s, fm[1]);
@@ -2135,6 +2180,9 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
   return (int)printed;
 
 error:
+  if ( fs_malloced )
+    free(fs_malloced);
+
   SUNLOCK(s);
   return -1;
 }
@@ -2175,7 +2223,7 @@ Svsprintf(char *buf, const char *fm, va_list args)
 int
 Svdprintf(const char *fm, va_list args)
 { int rval;
-  IOSTREAM *s = Soutput;
+  IOSTREAM *s = Serror;
 
   Slock(s);
   rval = Svfprintf(s, fm, args);
@@ -2668,7 +2716,6 @@ is of no value.
 IOSTREAM *
 Snew(void *handle, int flags, IOFUNCTIONS *functions)
 { IOSTREAM *s;
-  int fd;
 
   if ( !(s = malloc(sizeof(IOSTREAM))) )
   { errno = ENOMEM;
@@ -2699,6 +2746,7 @@ Snew(void *handle, int flags, IOFUNCTIONS *functions)
 #endif
 
 #ifndef __WINDOWS__			/* (*) */
+{ int fd;
   if ( (fd = Sfileno(s)) >= 0 )
   { if ( isatty(fd) )
       s->flags |= SIO_ISATTY;
@@ -2706,6 +2754,7 @@ Snew(void *handle, int flags, IOFUNCTIONS *functions)
     fcntl(fd, F_SETFD, FD_CLOEXEC);
 #endif
   }
+}
 #endif
 
   return s;
@@ -2737,6 +2786,7 @@ Sopen_file(const char *path, const char *how)
   enum {lnone=0,lread,lwrite} lock = lnone;
   IOSTREAM *s;
   IOENC enc = ENC_UNKNOWN;
+  int wait = TRUE;
 
   for( ; *how; how++)
   { switch(*how)
@@ -2747,6 +2797,9 @@ Sopen_file(const char *path, const char *how)
       case 'r':				/* no record */
 	flags &= ~SIO_RECORDPOS;
         break;
+      case 'L':				/* lock r: read, w: write */
+	wait = FALSE;
+        /*FALLTHROUGH*/
       case 'l':				/* lock r: read, w: write */
 	if ( *++how == 'r' )
 	  lock = lread;
@@ -2802,7 +2855,7 @@ Sopen_file(const char *path, const char *how)
     memset(&buf, 0, sizeof(buf));
     buf.l_type = (lock == lread ? F_RDLCK : F_WRLCK);
 
-    if ( fcntl(fd, F_SETLKW, &buf) < 0 )
+    if ( fcntl(fd, wait ? F_SETLKW : F_SETLK, &buf) < 0 )
     { int save = errno;
       close(fd);
       errno = save;
@@ -2812,14 +2865,20 @@ Sopen_file(const char *path, const char *how)
 #if __WINDOWS__
     HANDLE h = (HANDLE)_get_osfhandle(fd);
     OVERLAPPED ov;
+    int flags = 0;
+
+    if ( lock == lwrite )
+      flags |= LOCKFILE_EXCLUSIVE_LOCK;
+    if ( !wait )
+      flags |= LOCKFILE_FAIL_IMMEDIATELY;
 
     memset(&ov, 0, sizeof(ov));
-    if ( !LockFileEx(h, (lock == lread ? 0 : LOCKFILE_EXCLUSIVE_LOCK),
+    if ( !LockFileEx(h, flags,
 		     0,
 		     0, 0xfffffff,
 		     &ov) )
     { close(fd);
-      errno = EACCES;			/* TBD: proper error */
+      errno = (wait ? EACCES : EAGAIN);	/* TBD: proper error */
       return NULL;
     }
 #else
@@ -2843,8 +2902,8 @@ Sopen_file(const char *path, const char *how)
 
 IOSTREAM *
 Sfdopen(int fd, const char *type)
-{ int flags;
-  intptr_t lfd;
+{ intptr_t lfd;
+  int flags = SIO_FILE|SIO_RECORDPOS|SIO_FBUF;
 
   if ( fd < 0 )
   { errno = EINVAL;
@@ -2856,9 +2915,15 @@ Sfdopen(int fd, const char *type)
 #endif
 
   if ( *type == 'r' )
-    flags = SIO_FILE|SIO_INPUT|SIO_RECORDPOS|SIO_FBUF;
-  else
-    flags = SIO_FILE|SIO_OUTPUT|SIO_RECORDPOS|SIO_FBUF;
+  { flags |= SIO_INPUT;
+  } else if ( *type == 'w' )
+  { flags |= SIO_OUTPUT;
+  } else
+  { errno = EINVAL;
+    return NULL;
+  }
+  if ( type[1] != 'b' )
+    flags |= SIO_TEXT;
 
   lfd = (intptr_t)fd;
 
@@ -3017,7 +3082,7 @@ Swrite_memfile(void *handle, char *buf, size_t size)
 { memfile *mf = handle;
 
   if ( mf->here + size + 1 >= mf->allocated )
-  { intptr_t ns = S__memfile_nextsize(mf->here + size + 1);
+  { size_t ns = S__memfile_nextsize(mf->here + size + 1);
     char *nb;
 
     if ( mf->allocated == 0 || !mf->malloced )
