@@ -197,7 +197,8 @@ get_taia(term_t t, struct taia *taia, double *seconds)
 
 static int
 get_tz_arg(int i, term_t t, term_t a, atom_t *tz)
-{ atom_t name;
+{ GET_LD
+  atom_t name;
 
   _PL_get_arg(i, t, a);
   if ( !PL_get_atom_ex(a, &name) )
@@ -211,7 +212,9 @@ get_tz_arg(int i, term_t t, term_t a, atom_t *tz)
 
 static int
 get_int_arg(int i, term_t t, term_t a, int *val)
-{ _PL_get_arg(i, t, a);
+{ GET_LD
+
+  _PL_get_arg(i, t, a);
 
   return PL_get_integer_ex(a, val);
 }
@@ -219,7 +222,9 @@ get_int_arg(int i, term_t t, term_t a, int *val)
 
 static int
 get_float_arg(int i, term_t t, term_t a, double *val)
-{ _PL_get_arg(i, t, a);
+{ GET_LD
+
+  _PL_get_arg(i, t, a);
 
   return PL_get_float_ex(a, val);
 }
@@ -227,7 +232,8 @@ get_float_arg(int i, term_t t, term_t a, double *val)
 
 static int
 get_bool_arg(int i, term_t t, term_t a, int *val)
-{ atom_t name;
+{ GET_LD
+  atom_t name;
 
   _PL_get_arg(i, t, a);
   if ( PL_get_atom(a, &name) )
@@ -246,7 +252,9 @@ get_bool_arg(int i, term_t t, term_t a, int *val)
 
 static int
 get_ftm(term_t t, ftm *ftm)
-{ if ( PL_is_functor(t, FUNCTOR_date9) )
+{ GET_LD
+
+  if ( PL_is_functor(t, FUNCTOR_date9) )
   { term_t tmp = PL_new_term_ref();
 
     memset(ftm, 0, sizeof(*ftm));
@@ -262,6 +270,7 @@ get_ftm(term_t t, ftm *ftm)
 	 get_bool_arg (9, t, tmp, &ftm->isdst) )
     { double fp, ip;
 
+    fixup:
       fp = modf(ftm->sec, &ip);
       if ( fp < 0.0 )
       { fp += 1.0;
@@ -274,6 +283,15 @@ get_ftm(term_t t, ftm *ftm)
 
       succeed;
     }
+  } else if ( PL_is_functor(t, FUNCTOR_date3) )
+  { term_t tmp = PL_new_term_ref();
+
+    memset(ftm, 0, sizeof(*ftm));
+
+    if ( get_int_arg  (1, t, tmp, &ftm->tm.tm_year) &&
+	 get_int_arg  (2, t, tmp, &ftm->tm.tm_mon)  &&
+	 get_int_arg  (3, t, tmp, &ftm->tm.tm_mday) )
+      goto fixup;
   }
 
   fail;
@@ -316,7 +334,8 @@ cal_ftm(ftm *ftm, int required)
 
 static
 PRED_IMPL("stamp_date_time", 3, stamp_date_time, 0)
-{ struct taia taia;
+{ PRED_LD
+  struct taia taia;
   term_t compound = A2;
   double argsec;
 
@@ -448,7 +467,8 @@ iso_week_days(int yday, int wday)
 
 static int
 fmt_domain_error(const char *key, int value)
-{ term_t t = PL_new_term_ref();
+{ GET_LD
+  term_t t = PL_new_term_ref();
 
   PL_put_integer(t, value);
 
@@ -457,7 +477,8 @@ fmt_domain_error(const char *key, int value)
 
 static int
 fmt_not_implemented(const char *key)
-{ term_t t = PL_new_term_ref();
+{ GET_LD
+  term_t t = PL_new_term_ref();
 
   PL_put_atom_chars(t, key);
 
@@ -756,7 +777,7 @@ format_time(IOSTREAM *fd, const wchar_t *format, ftm *ftm, int posix)
 	  case 'z':			/* Time-zone as offset */
 	  { int min = -ftm->utcoff/60;
 
-	    if ( min > 0 )
+	    if ( min >= 0 )
 	    { OUTCHR(fd, '+');
 	    } else
 	    { min = -min;
@@ -861,7 +882,7 @@ pl_format_time(term_t out, term_t format, term_t time, int posix)
       tb.utcoff     = 0;
     }
   } else if ( !get_ftm(time, &tb) )
-  { fail;
+  { return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_time, time);
   }
 
   if ( !setupOutputRedirect(out, &ctx, FALSE) )
@@ -880,7 +901,8 @@ PRED_IMPL("format_time", 3, format_time3, 0)
 
 static
 PRED_IMPL("format_time", 4, format_time4, 0)
-{ int posix = FALSE;
+{ PRED_LD
+  int posix = FALSE;
   atom_t locale;
 
   if ( !PL_get_atom_ex(A4, &locale) )
