@@ -94,6 +94,31 @@ PL_save_text(PL_chars_t *text, int flags)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+PL_from_stack_text() moves a string from  the   stack,  so  it won't get
+corrupted if GC/shift comes along.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static void
+PL_from_stack_text(PL_chars_t *text)
+{ if ( text->storage == PL_CHARS_STACK )
+  { size_t bl = bufsize_text(text, text->length+1);
+
+    if ( bl < sizeof(text->buf) )
+    { memcpy(text->buf, text->text.t, bl);
+      text->text.t = text->buf;
+      text->storage = PL_CHARS_LOCAL;
+    } else
+    { Buffer b = findBuffer(BUF_RING);
+
+      addMultipleBuffer(b, text->text.t, bl, char);
+      text->text.t = baseBuffer(b, char);
+      text->storage = PL_CHARS_RING;
+    }
+  }
+}
+
+
 int
 PL_get_text__LD(term_t l, PL_chars_t *text, int flags ARG_LD)
 { word w = valHandle(l);
@@ -104,6 +129,7 @@ PL_get_text__LD(term_t l, PL_chars_t *text, int flags ARG_LD)
   } else if ( (flags & CVT_STRING) && isString(w) )
   { if ( !get_string_text(w, text PASS_LD) )
       goto maybe_write;
+    PL_from_stack_text(text);
   } else if ( (flags & CVT_INTEGER) && isInteger(w) )
   { number n;
 
@@ -167,16 +193,21 @@ PL_get_text__LD(term_t l, PL_chars_t *text, int flags ARG_LD)
     text->encoding = ENC_ISO_LATIN_1;
     text->storage  = PL_CHARS_LOCAL;
     text->canonical = TRUE;
-  } else if ( (flags & CVT_WRITE) )
+  } else if ( (flags & (CVT_WRITE|CVT_WRITE_CANONICAL)) )
   { IOENC encodings[3];
     IOENC *enc;
     char *r;
+    int wflags;
 
   case_write:
     encodings[0] = ENC_ISO_LATIN_1;
     encodings[1] = ENC_WCHAR;
     encodings[2] = ENC_UNKNOWN;
 
+    wflags = ((flags&CVT_WRITE_CANONICAL)
+		? PL_WRT_QUOTED|PL_WRT_IGNOREOPS|PL_WRT_NUMBERVARS
+		: PL_WRT_NUMBERVARS);
+
     for(enc = encodings; *enc != ENC_UNKNOWN; enc++)
     { size_t size;
       IOSTREAM *fd;
@@ -186,7 +217,7 @@ PL_get_text__LD(term_t l, PL_chars_t *text, int flags ARG_LD)
       fd = Sopenmem(&r, &size, "w");
       fd->encoding = *enc;
       if ( PL_write_term(fd, l, 1200, 0) &&
-	   Sputcode(EOS, fd) >= 0 &&
+	   Sputcode(EOS, fd) >= wflags &&
 	   Sflush(fd) >= 0 )
       { text->encoding = *enc;
 	text->storage = (r == text->buf ? PL_CHARS_LOCAL : PL_CHARS_MALLOC);
@@ -218,11 +249,11 @@ PL_get_text__LD(term_t l, PL_chars_t *text, int flags ARG_LD)
   succeed;
 
 maybe_write:
-  if ( (flags & CVT_WRITE) )
+  if ( (flags & (CVT_WRITE|CVT_WRITE_CANONICAL)) )
     goto case_write;
 
 error:
-  if ( isVar(w) && (flags & CVT_VARNOFAIL) )
+  if ( canBind(w) && (flags & CVT_VARNOFAIL) )
     return 2;
 
   if ( (flags & CVT_EXCEPTION) )
@@ -244,7 +275,8 @@ error:
 
 atom_t
 textToAtom(PL_chars_t *text)
-{ PL_canonise_text(text);
+{ if ( !PL_canonise_text(text) )
+    return 0;
 
   if ( text->encoding == ENC_ISO_LATIN_1 )
   { return lookupAtom(text->text.t, text->length);
@@ -256,7 +288,8 @@ textToAtom(PL_chars_t *text)
 
 word
 textToString(PL_chars_t *text)
-{ PL_canonise_text(text);
+{ if ( !PL_canonise_text(text) )
+    return 0;
 
   if ( text->encoding == ENC_ISO_LATIN_1 )
   { return globalString(text->length, text->text.t);
@@ -271,15 +304,22 @@ PL_unify_text(term_t term, term_t tail, PL_chars_t *text, int type)
 { switch(type)
   { case PL_ATOM:
     { atom_t a = textToAtom(text);
-      int rval = _PL_unify_atomic(term, a);
 
-      PL_unregister_atom(a);
-      return rval;
+      if ( a )
+      { int rval = _PL_unify_atomic(term, a);
+
+	PL_unregister_atom(a);
+	return rval;
+      }
+      return FALSE;
     }
     case PL_STRING:
     { word w = textToString(text);
 
-      return _PL_unify_atomic(term, w);
+      if ( w )
+	return _PL_unify_atomic(term, w);
+      else
+	return FALSE;
     }
     case PL_CODE_LIST:
     case PL_CHAR_LIST:
@@ -301,7 +341,9 @@ PL_unify_text(term_t term, term_t tail, PL_chars_t *text, int type)
 	  { const unsigned char *s = (const unsigned char *)text->text.t;
 	    const unsigned char *e = &s[text->length];
 
-	    p0 = p = allocGlobal(text->length*3);
+	    if ( !(p0 = p = allocGlobal(text->length*3)) )
+	      return FALSE;
+
 	    for( ; s < e; s++)
 	    { *p++ = FUNCTOR_dot2;
 	      if ( type == PL_CODE_LIST )
@@ -317,7 +359,9 @@ PL_unify_text(term_t term, term_t tail, PL_chars_t *text, int type)
 	  { const pl_wchar_t *s = (const pl_wchar_t *)text->text.t;
 	    const pl_wchar_t *e = &s[text->length];
 
-	    p0 = p = allocGlobal(text->length*3);
+	    if ( !(p0 = p = allocGlobal(text->length*3)) )
+	      return FALSE;
+
 	    for( ; s < e; s++)
 	    { *p++ = FUNCTOR_dot2;
 	      if ( type == PL_CODE_LIST )
@@ -334,7 +378,9 @@ PL_unify_text(term_t term, term_t tail, PL_chars_t *text, int type)
 	    const char *e = &s[text->length];
 	    size_t len = utf8_strlen(s, text->length);
 
-	    p0 = p = allocGlobal(len*3);
+	    if ( !(p0 = p = allocGlobal(len*3)) )
+	      return FALSE;
+
 	    while(s<e)
 	    { int chr;
 
@@ -357,16 +403,22 @@ PL_unify_text(term_t term, term_t tail, PL_chars_t *text, int type)
 	    wchar_t wc;
 
 	    memset(&mbs, 0, sizeof(mbs));
-	    while( n > 0 && (rc=mbrtowc(&wc, s, n, &mbs)) != (size_t)-1 )
-	    { len++;
+	    while( n > 0 )
+	    { if ( (rc=mbrtowc(&wc, s, n, &mbs)) == (size_t)-1 || rc == 0 )
+		return PL_error(NULL, 0, "cannot represent text in current locale",
+				ERR_REPRESENTATION, ATOM_encoding);
+
+	      len++;
 	      n -= rc;
 	      s += rc;
 	    }
 
-	    p0 = p = allocGlobal(len*3);
-	    memset(&mbs, 0, sizeof(mbs));
-	    n = text->length;
+	    if ( !(p0 = p = allocGlobal(len*3)) )
+	      return FALSE;
 
+	    n = text->length;
+	    s = text->text.t;
+	    memset(&mbs, 0, sizeof(mbs));
 	    while(n > 0)
 	    { rc = mbrtowc(&wc, s, n, &mbs);
 
@@ -726,7 +778,7 @@ PL_canonise_text(PL_chars_t *text)
 
 	for(; w<e; w++)
 	{ if ( *w > 0xff )
-	    return FALSE;
+	    return TRUE;
 	}
 
 	return PL_demote_text(text);
@@ -795,8 +847,11 @@ PL_canonise_text(PL_chars_t *text)
 	wchar_t wc;
 
 	memset(&mbs, 0, sizeof(mbs));
-	while( n > 0 && (rc=mbrtowc(&wc, s, n, &mbs)) != (size_t)-1 )
-	{ if ( wc > 0xff )
+	while( n > 0 )
+	{ if ( (rc=mbrtowc(&wc, s, n, &mbs)) == (size_t)-1 || rc == 0)
+	    return FALSE;		/* encoding error */
+
+	  if ( wc > 0xff )
 	    iso = FALSE;
 	  len++;
 	  n -= rc;
@@ -828,8 +883,10 @@ PL_canonise_text(PL_chars_t *text)
 	    }
 
 	    to = text->text.t;
-	    while( n > 0 && (rc=mbrtowc(&wc, from, n, &mbs)) != (size_t)-1 )
-	    { *to++ = (char)wc;
+	    while( n > 0 )
+	    { rc = mbrtowc(&wc, from, n, &mbs);
+
+	      *to++ = (char)wc;
 	      n -= rc;
 	      from += rc;
 	    }
@@ -851,8 +908,10 @@ PL_canonise_text(PL_chars_t *text)
 	    }
 
 	    to = text->text.w;
-	    while( n > 0 && (rc=mbrtowc(&wc, from, n, &mbs)) != (size_t)-1 )
-	    { *to++ = wc;
+	    while( n > 0 )
+	    { rc = mbrtowc(&wc, from, n, &mbs);
+
+	      *to++ = wc;
 	      n -= rc;
 	      from += rc;
 	    }
